import 'dart:convert';

import '../../interfaces/baas_object.dart';

import '../request/request_utils.dart';
import '../../interfaces/database.dart';
import 'package:saker/framework/log/log.dart';

class AliYunDataBase implements DataBase {
  String _collName;
  Logger logger;

  AliYunDataBase(this._collName, this.logger);

  String get name => _collName;

  /*
  * where: {
        "playerName": "Sean Plott",
      },
      order: "-score,-name",
      limit: 200,
      skip: 0,
  * */

  Future<BaasDBGetObjResult<T>> getObjects<T extends BaasObject>(
      Map<String, dynamic> queryCondition,
      int start,
      int limit,
      String orderBy,
      Function createT,
      {int count: 0,
      Duration timeout: baasDBTimeout}) async {
    if (queryCondition == null || queryCondition.isEmpty) {
      logger?.error("DataBase getObjects queryCondition is null or empty.");
      return null;
    }
    String url = '/mbaas/classes/$_collName';

    Map<String, dynamic> request = {
      "where": json.encode(queryCondition),
      "skip": "$start",
      "limit": "$limit",
      "order": orderBy,
      "count": "$count"
    };

    ResponseResult result = await RequestUtils.sendGetRequest(url, timeout,"getObjects", request);
    logger?.debug("DataBase getObjects: url:$url req:${json.encode(request)} resp: ${result.toString()}");

    BaasDBGetObjResult<T> baasDBGetResult = new BaasDBGetObjResult<T>()
      ..success = result.success
      ..items = []
      ..errMsg = result.errMsg;

    if (result.success) {
      try {
        Map<String, dynamic> map = json.decode(result.responseBody) as Map<String, dynamic>;

        if (map.containsKey("count")) {
          baasDBGetResult.totalCount = map['count'] as int;
        }

        if (map.containsKey("results")) {
          List<Map<String, dynamic>> res = (map['results'] as List)?.cast();
          for (Map<String, dynamic> m in res) {
            T t = createT() as T;
            t.fromJson(m);
            baasDBGetResult.items.add(t);
          }
        }
      } catch (e, s) {
        logger?.error("DataBase getObjects: decode json response failed: $e  $s");
        baasDBGetResult.success = false;
        baasDBGetResult.errMsg = "DataBase getObjects: decode json response failed: $e  $s";
      }
    }

    return baasDBGetResult;
  }

  Future<BaasDBGetResult> get(
      Map<String, dynamic> queryCondition, int start, int limit, String orderBy,
      {int count: 0, Duration timeout: baasDBTimeout}) async {

    if (queryCondition == null || queryCondition.isEmpty) {
      logger?.error("DataBase get queryCondition is null or empty.");
      return null;
    }
    String url = '/mbaas/classes/$_collName';

    Map<String, dynamic> request = {
      "where": json.encode(queryCondition),
      "skip": "$start",
      "limit": "$limit",
      "order": orderBy,
      "count": "$count"
    };



    ResponseResult result = await RequestUtils.sendGetRequest(url, timeout,"get", request);
    logger?.debug("DataBase get: url:$url req:${json.encode(request)} resp: ${result.toString()}");

    BaasDBGetResult baasDBGetResult = new BaasDBGetResult()
      ..success = result.success
      ..items = []
      ..errMsg = result.errMsg;

    if (result.success) {
      try {
        Map<String, dynamic> map = json.decode(result.responseBody) as Map<String, dynamic>;

        if (map.containsKey("count")) {
          baasDBGetResult.totalCount = map['count'] as int;
        }

        if (map.containsKey("results")) {
          List<Map<String, dynamic>> res = (map['results'] as List)?.cast();
          for (Map<String, dynamic> m in res) {
            baasDBGetResult.items.add(m);
          }
        }
      } catch (e, s) {
        logger?.error("DataBase get: decode json response failed: $e  $s");
        baasDBGetResult.success = false;
        baasDBGetResult.errMsg = "DataBase get: decode json response failed: $e  $s";
      }
    }

    return baasDBGetResult;
  }

  Future<BaasDBAddResult> addObject(BaasObject baasObjet,
      {Duration timeout: baasDBTimeout}) async {
    if (baasObjet == null) {
      logger?.error("DataBase addObject: baasObject is null.");
      return null;
    }
    return add(baasObjet.toJson(), timeout: timeout);
  }

  Future<BaasDBAddResult> add(Map<String, dynamic> paramJson, {Duration timeout: baasDBTimeout}) async {

    if (paramJson == null || paramJson.isEmpty) {
      logger?.error("DataBase add :paramJson null or empty object.");
      return null;
    }
    String url = '/mbaas/classes/$_collName';
    String request = json.encode(paramJson);


    ResponseResult resp = await RequestUtils.sendPostRequest(url, timeout, "add", paramJson);
    logger?.debug("DataBase add : url:$url req:${request} resp: $resp"); //{"objectId":"Yl7RADAwCO","createdAt":"2019-08-21T03:24:56.002Z"}

    BaasDBAddResult baasDBAddResult = new BaasDBAddResult()
      ..success = resp.success
      ..errMsg = resp.errMsg;

    if (resp.success) {
      try {

        Map<String, dynamic> map = json.decode(resp.responseBody) as Map<String, dynamic>;
        baasDBAddResult.objectId = map['objectId'] as String;
        baasDBAddResult.createAt = map['createdAt'] as String;
      } catch (e, s) {
        logger?.error("DataBase add :decode json response failed: $e  $s");
        baasDBAddResult.success = false;
        baasDBAddResult.errMsg = "DataBase add :decode json response failed: $e  $s";

      }
    }
    return baasDBAddResult;
  }

  Future<List<BaasDBAddResult>> addObjects(List<BaasObject> items,
      {Duration timeout: baasDBTimeout}) {
    if (items == null || items.isEmpty) {
      logger?.error("DataBase addObjects: null or empty list");
      return null;
    }
    List list = [];
    items.forEach((elem) {
      list.add(elem.toJson());
    });
    return addMany(list, timeout: timeout);
  }

  Future<List<BaasDBAddResult>> addMany(List items,
      {Duration timeout: baasDBTimeout}) async {
    if (items == null || items.isEmpty) {
      logger?.error("DataBase addMany : null or empty List.");
      return null;
    }

    List paramJson = [];
    items.forEach((e) {
      Map obj = {};
      obj["method"] = "POST";
      obj["path"] = '/mbaas/classes/${this._collName}';
      obj['body'] = e;
      paramJson.add(obj);
    });

    String url = '/mbaas/batch';

    /*
    *
  [
    {
        "success": {
            "objectId": "0ZZ2lJEDVF",
            "createdAt": "2019-08-21T03:24:56.126Z"
        }
    },
    {
        "success": {
            "objectId": "Sj5DR8KHp0",
            "createdAt": "2019-08-21T03:24:56.126Z"
        }
    },
    {
      "error": {
        "code": 101,
        "error": "object not found for delete"
      }
    }
  ]
*/
    String request = json.encode({"requests": paramJson});

    ResponseResult resp = await RequestUtils.sendPostRequest(url, timeout, "addMany", {"requests": paramJson});
    logger?.debug("DataBase addMany: url:$url  req: $request resp: $resp");
    List<BaasDBAddResult> result = [];

    if (resp.success) {
      try {
        List<Map<String, dynamic>> res = (json.decode(resp.responseBody) as List)?.cast();
        for (Map m in res) {
          BaasDBAddResult baasDBAddResult = new BaasDBAddResult();

          if (m.containsKey("success")) {
            baasDBAddResult.success = true;
            Map<String, dynamic> succMap = m["success"] as Map<String, dynamic>;
            baasDBAddResult.objectId = succMap['objectId'] as String;
            baasDBAddResult.createAt = succMap['createdAt'] as String;
          } else if (m.containsKey("error")) {
            baasDBAddResult.success = false;
            baasDBAddResult.errMsg = json.encode("m");
          }
          result.add(baasDBAddResult);
        }
      } catch (e, s) {
        logger?.error("DataBase addMany:decode json response failed: $e  $s");
      }
    }
    return result;
  }

  Future<BaasDBUpdateResult> updateObjectById(BaasObject baasObjet,
      {Duration timeout: baasDBTimeout}) {
    return updateById(baasObjet.objectId, baasObjet.toJson(), timeout: timeout);
  }

  Future<BaasDBUpdateResult> updateById(
      String id, Map<String, dynamic> paramJson,
      {Duration timeout: baasDBTimeout}) async {
    if (id == null || id.isEmpty) {
      logger?.error("DataBase update: error: invaild id");
      return null;
    }

    if (paramJson == null || paramJson.isEmpty) {
      logger?.error("DataBase update: error: invalid value");
      return null;
    }

    String url = '/mbaas/classes/$_collName/$id';
    String request = json.encode(paramJson);

    ResponseResult resp = await RequestUtils.sendPutRequest(url, timeout,"updateById", paramJson);

    logger?.debug("DataBase update: url:$url  req:$request resp: $resp");
    BaasDBUpdateResult baasDBUpdateResult = new BaasDBUpdateResult()
      ..success = resp.success
      ..errMsg = resp.errMsg;

    if (resp.success) {
      try {
        Map<String, dynamic> m = json.decode(resp.responseBody) as Map<String, dynamic>;
        baasDBUpdateResult.objectId = m['objectId'] as String ?? "";
        baasDBUpdateResult.updateAt = m['updatedAt'] as String ?? "";
      } catch (e, s) {
        logger?.error("DataBase update:decode json response failed: $e  $s");
        baasDBUpdateResult.success = false;
        baasDBUpdateResult.errMsg = "DataBase update: decode json response failed: $e  $s";
      }
    }
    return baasDBUpdateResult;
  }



  Future<BaasDBUpsertResult> upsert(Map<String,dynamic> queryParamJson, Map<String,dynamic> upsertParamJson, {Duration timeout : baasDBTimeout}) async {
    if (queryParamJson == null || queryParamJson.isEmpty) {
      logger?.error("DataBase upsert: error: invalid queryParamJson");
      return null;
    }

    if (upsertParamJson == null || upsertParamJson.isEmpty) {
      logger?.error("DataBase upsert: error: invalid upsertParamJson");
      return null;
    }

    String url = '/mbaas/classes/$_collName';

    Map<String,dynamic> realParamJson =<String, dynamic>{};
//    realParamJson["where"] = queryParamJson;
    realParamJson.putIfAbsent("where", ()=>queryParamJson);
    realParamJson.putIfAbsent("update", ()=>upsertParamJson);

    String request = json.encode(realParamJson);

    ResponseResult resp = await RequestUtils.sendPutRequest(url, timeout,"upsert", realParamJson);

    logger?.debug("DataBase upsert: url:$url  req:$request resp: $resp");
    BaasDBUpsertResult baasDBUpsertResult = new BaasDBUpsertResult()
      ..success = resp.success
      ..errMsg = resp.errMsg;

    if (resp.success) {
      try {
        Map<String, dynamic> m = json.decode(resp.responseBody) as Map<String, dynamic>;
        baasDBUpsertResult.updateAt = m['updatedAt'] as String ?? "";
      } catch (e, s) {
        logger?.error("DataBase upsert:decode json response failed: $e  $s");
        baasDBUpsertResult.success = false;
        baasDBUpsertResult.errMsg = "DataBase upsert: decode json response failed: $e  $s";
      }
    }
    return baasDBUpsertResult;
  }




  Future<BaasDBDelResult> deleteById(String id,
      {Duration timeout: baasDBTimeout}) async {
    if (id == null || id.isEmpty) {
      logger?.error("DataBase deleteById: failed: invaild id");
      return null;
    }

    String url = '/mbaas/classes/$_collName/$id';


    ResponseResult resp = await RequestUtils.sendDeleteRequest(url, timeout,"deleteById");
    logger?.debug("DataBase delete: url:$url resp: $resp");
    BaasDBDelResult baasDBDelResult = new BaasDBDelResult()
      ..success = resp.success
      ..errMsg = resp.errMsg;

    return baasDBDelResult;
  }

  Future<BaasDBCountResult> count(Map<String, dynamic> queryCondition,
      {Duration timeout: baasDBTimeout}) async {
    if (queryCondition == null || queryCondition.isEmpty) {
      logger?.error("DataBase count: queryCondition null or empty.");
      return null;
    }
    String url = '/mbaas/classes/$_collName';

    Map<String, dynamic> request = {
      "where": json.encode(queryCondition),
      "count": "1",
      "limit": "0"
    };


    ResponseResult result =
        await RequestUtils.sendGetRequest(url, timeout,"count", request);
    logger?.debug(
        "DataBase count: url:$url req:${json.encode(request)} resp: ${result.toString()}");

    BaasDBCountResult baasDBCountResult = new BaasDBCountResult()
      ..success = result.success
      ..errMsg = result.errMsg
      ..totalCount = 0;

    if (result.success) {
      try {
        Map<String, dynamic> map = json.decode(result.responseBody) as Map<String, dynamic>;
        if (map.containsKey("count")) {
          baasDBCountResult.totalCount = map['count'] as int;
        }
      } catch (e, s) {
        logger?.error("DataBase count: decode json response failed: $e  $s");
        baasDBCountResult.success = false;
        baasDBCountResult.errMsg =
            "DataBase count: decode json response failed: $e  $s";
      }
    }

    return baasDBCountResult;
  }

  Future<List<BaasDBDelResult>> deleteManybyIds(List<String> itemObjectIDs,
      {Duration timeout: baasDBTimeout}) async {
    if (itemObjectIDs == null || itemObjectIDs.isEmpty) {
      logger?.error("DataBase deleteManybyIds : null or empty param list.");
      return null;
    }

    List paramJson = [];
    itemObjectIDs.forEach((e) {
      Map obj = {};
      obj["method"] = "DELETE";
      obj["path"] = '/mbaas/classes/${this._collName}/$e';
      paramJson.add(obj);
    });

    String url = '/mbaas/batch';

    String request = json.encode({"requests": paramJson});


    ResponseResult resp = await RequestUtils.sendPostRequest(url, timeout,"deleteManybyIds", {"requests": paramJson});
    logger?.debug("DataBase deleteMany: url:$url  req: $request resp: $resp");
    List<BaasDBDelResult> result = [];

    if (resp.success) {
      try {
        List<Map<String, dynamic>> res = (json.decode(resp.responseBody) as List)?.cast();
        for (Map m in res) {
          BaasDBDelResult baasDBDelResult = new BaasDBDelResult();

          if (m.containsKey("success")) {
            baasDBDelResult.success = true;
          } else if (m.containsKey("error")) {
            baasDBDelResult.success = false;
            baasDBDelResult.errMsg = json.encode("m");
          }
          result.add(baasDBDelResult);
        }
      } catch (e, s) {
        logger?.error("DataBase deleteMany:decode json response failed: $e  $s");
      }
    }
    return result;
  }

  Future<BaasDBDelResult> delete(Map<String, dynamic> queryCondition,
      {int count: 0, Duration timeout: baasDBTimeout}) async {
    if (queryCondition == null || queryCondition.isEmpty) {
      logger?.error("DataBase delete: deleteCondition null or empty.");
      return null;
    }
    String url = '/mbaas/classes/$_collName';

    Map<String, dynamic> request = {"where": json.encode(queryCondition)};

    ResponseResult result = await RequestUtils.sendDeleteRequest(url, timeout,"delete", paramJSON: request);
    logger?.debug("DataBase delete: url:$url req:${json.encode(request)} resp: ${result.toString()}");

    BaasDBDelResult baasDBDelResult = new BaasDBDelResult()
      ..success = result.success
      ..errMsg = result.errMsg;

    return baasDBDelResult;
  }
}

