
import 'dart:convert';
import 'dart:io';

import 'package:saker/common/net/ip_address.dart';
import 'package:saker/framework/log/log_type.dart';
import 'package:saker/framework/stat/service_monitor.dart';
import 'package:saker/framework/trace/trace_info.dart';

import '../config/baas_config.dart';
import './http_conn_pool.dart';


class ResponseResult{
  bool success;
  String responseBody;
  String errMsg;

  String toString(){
    return "ResponseResult : success? $success, responseBody:$responseBody, errMsg:$errMsg";
  }
}

class RequestUtils{

  static final String getMethod = "GET";
  static final String postMethod = "POST";
  static final String putMethod = "PUT";
  static final String delMethod = "DELETE";

  static Future<ResponseResult> sendHttpRequest(String method, String urlPath, Duration timeout, String methodName, {Map<String,dynamic> paramJSON = null, bool keepAlive = true} )async{
    String host = baasSDKConfigAliyun.host;
    int port = baasSDKConfigAliyun.port;

    String request = json.encode(paramJSON);
    ResponseResult result = new ResponseResult();

    ServiceMonitor providerMonitor = new ServiceMonitor()
      ..traceId = TraceInfo.getEagleeyeTraceId()
      ..rpcId = ""
      ..reqMethod = methodName
      ..req = "$urlPath $request"
      ..localIP = IPAddress.getHostIp()
      ..ttid = TraceInfo.getttid()
      ..success = true
      ..remoteIP = ""
      ..mtopPath = TraceInfo.getMtopPath()
      ..rt = new DateTime.now().millisecondsSinceEpoch
      ..respSize = 0
      ..logType = LogType.BaasRpc;

    HttpClient httpClient = httpConnPool.getConnection(host);
    Uri uri = null;
    if(method == getMethod && paramJSON != null) {
      uri = new Uri(scheme: baasSDKConfigAliyun.schema, host: baasSDKConfigAliyun.host, path: urlPath, queryParameters: paramJSON);
    }else{
      uri = new Uri(scheme: baasSDKConfigAliyun.schema, host: baasSDKConfigAliyun.host, path: urlPath);
    }

    HttpClientRequest httpClientRequest = await httpClient.openUrl(method, uri)
        .timeout(timeout,onTimeout: (){
          providerMonitor.success = false;
          providerMonitor.errCode = "httpclient openUrl timeout";
          result.success = false;
          result.errMsg = "httpclient openUrl timeout";
          return null;
    }).catchError((e,s) {
      baasSDKConfigAliyun.logger?.error('send${method}Request make Request failed, $host  $port $urlPath  ${e}, $s');
      providerMonitor.success = false;
      providerMonitor.errCode = e.toString();
      result.success = false;
      result.errMsg = e.toString();
    });

    if (httpClientRequest == null) {
      providerMonitor.success = false;
      result.success = false;
      if(providerMonitor.errCode == null){
        providerMonitor.errCode = "httpClientRequest null";
        result.errMsg = "httpClientRequest null";
      }
      providerMonitor.flush();

      if(!keepAlive){
        httpClient.close();
      }
      return result;
    }

    httpClientRequest.headers.add('Content-Type', 'application/fastjson');
    httpClientRequest.headers.add('accept', 'application/fastjson');
    httpClientRequest.headers.add('x-mbaas-application-id', baasSDKConfigAliyun.appID);
    httpClientRequest.headers.add('X-mBaas-Master-Key', baasSDKConfigAliyun.masterKey);
    httpClientRequest.headers.add('x-mbaas-env-id', baasSDKConfigAliyun.envId);

    if(paramJSON != null && (method == putMethod || method == postMethod || method == delMethod)){
      httpClientRequest.write(json.encode(paramJSON));
    }

    HttpClientResponse response = await httpClientRequest.close().timeout(timeout, onTimeout: () {
      providerMonitor.success = false;
      providerMonitor.errCode = "get Response timeout";
      result.success = false;
      result.errMsg = "get Response timeout";
      return null;
    }).catchError((e,s) {
      baasSDKConfigAliyun.logger?.error('send${method}Request HTTP RES ERROR:$host  $port $urlPath  ${e}, $s');
      providerMonitor.success = false;
      providerMonitor.errCode = e.toString();
      result.success = false;
      result.errMsg = e.toString();
    });

    if (response == null) {
      providerMonitor.flush();

      if(!keepAlive){
        httpClient.close();
      }
      return result;
    }

    result.success = response.statusCode >= 200 && response.statusCode < 300;

    result.responseBody = await response.transform(utf8.decoder).join()
        .catchError((e,s){
      baasSDKConfigAliyun.logger?.error('send${method}Request HTTP JOIN ERROR:$host  $port $urlPath ${e},$s');
      result.success = false;
      result.errMsg = 'send${method}Request HTTP JOIN ERROR:$host  $port $urlPath ${e}';

      providerMonitor.success = false;
      providerMonitor.errCode = e.toString();
    });

    if(!result.success && (result.errMsg == null || result.errMsg.isEmpty)){
      result.errMsg = result.responseBody;
    }

    providerMonitor.success = result.success;
    providerMonitor.respSize = result.responseBody?.length;
    providerMonitor.res = result.responseBody;
    providerMonitor.errCode = result.errMsg;
    providerMonitor.flush();

    if(!keepAlive){
      httpClient.close();
    }
    return result;
  }


  static Future<ResponseResult> sendGetRequest( String urlPath, Duration timeout,String methodName, Map<String,dynamic> paramJSON)async{
    return sendHttpRequest(getMethod, urlPath, timeout,methodName, paramJSON: paramJSON, keepAlive: false);
  }


  static Future<ResponseResult> sendPostRequest( String urlPath, Duration timeout, String methodName, Map<String,dynamic> paramJSON)async{
    return sendHttpRequest(postMethod, urlPath, timeout,methodName, paramJSON: paramJSON, keepAlive: false);
  }


  static Future<ResponseResult> sendPutRequest( String urlPath, Duration timeout, String methodName,  Map<String,dynamic>  paramJSON)async{
    return sendHttpRequest(putMethod, urlPath, timeout,methodName, paramJSON: paramJSON, keepAlive: false);
  }

  static Future<ResponseResult> sendDeleteRequest( String urlPath, Duration timeout, String methodName, { Map<String,dynamic>  paramJSON=null})async{
    return sendHttpRequest(delMethod, urlPath, timeout,methodName, paramJSON: paramJSON , keepAlive: false);
  }
}