import 'dart:convert';
import 'dart:io';

import 'dart:typed_data';
import 'package:convert/convert.dart';
import 'package:dcache/dcache.dart';
import 'package:saker/common/utils/type_cast.dart';
import 'package:saker/framework/hessian/hessian2_const.dart';

class HessianInputException implements IOException {
  final String message;
  final String cause;

  const HessianInputException(String msg, String reason): message = msg ,cause = reason;

  String toString() {
    var b = new StringBuffer()..write('HessianInputException: ')..write(message);
    if (cause != null) {
      b.write(', cause = $cause');
    }
    return b.toString();
  }
}

Uint8List staticBuffer = null;

class Hessian2Input {

  Hessian2Input(Uint8List buffer,int skipBytes, String remoteIP){
    _buffer = buffer;
    _offset = 0;
    _length = buffer.lengthInBytes;
    _byteData = new ByteData.view(_buffer.buffer);
    _types = [];
    _refs = [];
    _classDefs = [];
    _remoteIP = remoteIP;

    for(int idx =0;idx<skipBytes;idx++){
      read();
    }
  }

  /**
   * Reads a null
   *
   * <pre>
   * N
   * </pre>
   */
  void readNull() {
    int tag = read();

    switch (tag) {
      case BC_NULL: return;

      default:
        throw new HessianInputException("readNull", "$tag");
    }
  }

  /**
   * Reads a boolean
   *
   * <pre>
   * T
   * F
   * </pre>
   */
  bool readBoolean(){
    int tag = _offset < _length ? (_buffer[_offset++] & 0xff) : read();

    switch (tag) {
      case BC_NULL://'N':
        return false;
      case BC_TRUE: return true;
      case BC_FALSE: return false;

    // direct integer
      case 0x80: case 0x81: case 0x82: case 0x83:
      case 0x84: case 0x85: case 0x86: case 0x87:
      case 0x88: case 0x89: case 0x8a: case 0x8b:
      case 0x8c: case 0x8d: case 0x8e: case 0x8f:

      case 0x90: case 0x91: case 0x92: case 0x93:
      case 0x94: case 0x95: case 0x96: case 0x97:
      case 0x98: case 0x99: case 0x9a: case 0x9b:
      case 0x9c: case 0x9d: case 0x9e: case 0x9f:

      case 0xa0: case 0xa1: case 0xa2: case 0xa3:
      case 0xa4: case 0xa5: case 0xa6: case 0xa7:
      case 0xa8: case 0xa9: case 0xaa: case 0xab:
      case 0xac: case 0xad: case 0xae: case 0xaf:

      case 0xb0: case 0xb1: case 0xb2: case 0xb3:
      case 0xb4: case 0xb5: case 0xb6: case 0xb7:
      case 0xb8: case 0xb9: case 0xba: case 0xbb:
      case 0xbc: case 0xbd: case 0xbe: case 0xbf:
      return tag != BC_INT_ZERO;

    // INT_BYTE = 0
      case 0xc8:
        return read() != 0;

    // INT_BYTE != 0
      case 0xc0: case 0xc1: case 0xc2: case 0xc3:
      case 0xc4: case 0xc5: case 0xc6: case 0xc7:
      case 0xc9: case 0xca: case 0xcb:
      case 0xcc: case 0xcd: case 0xce: case 0xcf:
//      read();
      ++_offset;
      return true;

    // INT_SHORT = 0
      case 0xd4:
        return (256 * read() + read()) != 0;

    // INT_SHORT != 0
      case 0xd0: case 0xd1: case 0xd2: case 0xd3:
      case 0xd5: case 0xd6: case 0xd7:
//      read();
//      read();
      _offset +=2;
      return true;

      case BC_INT:
        return _parseInt() != 0;

      case 0xd8: case 0xd9: case 0xda: case 0xdb:
      case 0xdc: case 0xdd: case 0xde: case 0xdf:

      case 0xe0: case 0xe1: case 0xe2: case 0xe3:
      case 0xe4: case 0xe5: case 0xe6: case 0xe7:
      case 0xe8: case 0xe9: case 0xea: case 0xeb:
      case 0xec: case 0xed: case 0xee: case 0xef:
      return tag != BC_LONG_ZERO;

    // LONG_BYTE = 0
      case 0xf8:
        return read() != 0;

    // LONG_BYTE != 0
      case 0xf0: case 0xf1: case 0xf2: case 0xf3:
      case 0xf4: case 0xf5: case 0xf6: case 0xf7:
      case 0xf9: case 0xfa: case 0xfb:
      case 0xfc: case 0xfd: case 0xfe: case 0xff:
//      read();
      ++_offset;
      return true;

    // INT_SHORT = 0
      case 0x3c:
        return (256 * read() + read()) != 0;

    // INT_SHORT != 0
      case 0x38: case 0x39: case 0x3a: case 0x3b:
      case 0x3d: case 0x3e: case 0x3f:
//      read();
//      read();
      _offset +=2;
      return true;

      case BC_LONG_INT:
        return ((0x1000000 * read() + 0x10000 * read() + 0x100 * read() + read()) != 0);

      case BC_LONG://'L'
        return _parseLong() != 0;

      case BC_DOUBLE_ZERO:
        return false;

      case BC_DOUBLE_ONE:
        return true;

      case BC_DOUBLE_BYTE:
        return read() != 0;

      case BC_DOUBLE_SHORT:
        return (0x100 * read() + read()) != 0;

      case BC_DOUBLE_MILL:{
        int mills = _parseInt();
        return mills != 0;
      }

      case BC_DOUBLE://'D':
        return _parseDouble() != 0.0;

      default:
        throw HessianInputException("readBoolean:boolean needed", "$tag");
    }
  }

  /**
   * Reads a short
   *
   * <pre>
   * I b32 b24 b16 b8
   * </pre>
   */
  int readShort() {
    return readInt();
  }

  /**
   * Reads an integer
   *
   * <pre>
   * I b32 b24 b16 b8
   * </pre>
   */
  int readInt(){
    //int tag = _offset < _length ? (_buffer[_offset++] & 0xff) : read();
    int tag = read();

    switch (tag) {
      case BC_NULL://'N':
        return 0;

    // direct integer
      case 0x80: case 0x81: case 0x82: case 0x83:
      case 0x84: case 0x85: case 0x86: case 0x87:
      case 0x88: case 0x89: case 0x8a: case 0x8b:
      case 0x8c: case 0x8d: case 0x8e: case 0x8f:

      case 0x90: case 0x91: case 0x92: case 0x93:
      case 0x94: case 0x95: case 0x96: case 0x97:
      case 0x98: case 0x99: case 0x9a: case 0x9b:
      case 0x9c: case 0x9d: case 0x9e: case 0x9f:

      case 0xa0: case 0xa1: case 0xa2: case 0xa3:
      case 0xa4: case 0xa5: case 0xa6: case 0xa7:
      case 0xa8: case 0xa9: case 0xaa: case 0xab:
      case 0xac: case 0xad: case 0xae: case 0xaf:

      case 0xb0: case 0xb1: case 0xb2: case 0xb3:
      case 0xb4: case 0xb5: case 0xb6: case 0xb7:
      case 0xb8: case 0xb9: case 0xba: case 0xbb:
      case 0xbc: case 0xbd: case 0xbe: case 0xbf:
      return tag - BC_INT_ZERO;

    /* byte int */
      case 0xc0: case 0xc1: case 0xc2: case 0xc3:
      case 0xc4: case 0xc5: case 0xc6: case 0xc7:
      case 0xc8: case 0xc9: case 0xca: case 0xcb:
      case 0xcc: case 0xcd: case 0xce: case 0xcf:
      return ((tag - BC_INT_BYTE_ZERO) << 8) + read();

    /* short int */
      case 0xd0: case 0xd1: case 0xd2: case 0xd3:
      case 0xd4: case 0xd5: case 0xd6: case 0xd7:
      return ((tag - BC_INT_SHORT_ZERO) << 16) + 256 * read() + read();

      case BC_INT://'I':
      case BC_LONG_INT:
        return ((read() << 24) + (read() << 16) + (read() << 8) + read());

    // direct long
      case 0xd8: case 0xd9: case 0xda: case 0xdb:
      case 0xdc: case 0xdd: case 0xde: case 0xdf:

      case 0xe0: case 0xe1: case 0xe2: case 0xe3:
      case 0xe4: case 0xe5: case 0xe6: case 0xe7:
      case 0xe8: case 0xe9: case 0xea: case 0xeb:
      case 0xec: case 0xed: case 0xee: case 0xef:
      return tag - BC_LONG_ZERO;

    /* byte long */
      case 0xf0: case 0xf1: case 0xf2: case 0xf3:
      case 0xf4: case 0xf5: case 0xf6: case 0xf7:
      case 0xf8: case 0xf9: case 0xfa: case 0xfb:
      case 0xfc: case 0xfd: case 0xfe: case 0xff:
      return ((tag - BC_LONG_BYTE_ZERO) << 8) + read();

    /* short long */
      case 0x38: case 0x39: case 0x3a: case 0x3b:
      case 0x3c: case 0x3d: case 0x3e: case 0x3f:
      return ((tag - BC_LONG_SHORT_ZERO) << 16) + 256 * read() + read();

      case BC_LONG://'L':
        return _parseLong();

      case BC_DOUBLE_ZERO:
        return 0;

      case BC_DOUBLE_ONE:
        return 1;

    //case LONG_BYTE:
      case BC_DOUBLE_BYTE:
        return (_offset < _length ? _buffer[_offset++] : read());

    //case INT_SHORT:
    //case LONG_SHORT:
      case BC_DOUBLE_SHORT:
        return  (256 * read() + read());

      case BC_DOUBLE_MILL: {
        int mills = _parseInt();
        return (0.001 * mills).toInt();
      }

      case BC_DOUBLE://'D':
        return _parseDouble().toInt();

      case BC_FALSE://'F':
        return 0;

      case BC_TRUE://'T':
        return 1;

      default:
        throw HessianInputException("readInt:integer needed", "$tag");
    }
  }

  /**
   * Reads a long
   *
   * <pre>
   * L b64 b56 b48 b40 b32 b24 b16 b8
   * </pre>
   */
  int readLong() {
    int tag = read();

    switch (tag) {
      case BC_NULL://'N':
        return 0;

    // direct integer
      case 0x80: case 0x81: case 0x82: case 0x83:
      case 0x84: case 0x85: case 0x86: case 0x87:
      case 0x88: case 0x89: case 0x8a: case 0x8b:
      case 0x8c: case 0x8d: case 0x8e: case 0x8f:

      case 0x90: case 0x91: case 0x92: case 0x93:
      case 0x94: case 0x95: case 0x96: case 0x97:
      case 0x98: case 0x99: case 0x9a: case 0x9b:
      case 0x9c: case 0x9d: case 0x9e: case 0x9f:

      case 0xa0: case 0xa1: case 0xa2: case 0xa3:
      case 0xa4: case 0xa5: case 0xa6: case 0xa7:
      case 0xa8: case 0xa9: case 0xaa: case 0xab:
      case 0xac: case 0xad: case 0xae: case 0xaf:

      case 0xb0: case 0xb1: case 0xb2: case 0xb3:
      case 0xb4: case 0xb5: case 0xb6: case 0xb7:
      case 0xb8: case 0xb9: case 0xba: case 0xbb:
      case 0xbc: case 0xbd: case 0xbe: case 0xbf:
      return tag - BC_INT_ZERO;

    /* byte int */
      case 0xc0: case 0xc1: case 0xc2: case 0xc3:
      case 0xc4: case 0xc5: case 0xc6: case 0xc7:
      case 0xc8: case 0xc9: case 0xca: case 0xcb:
      case 0xcc: case 0xcd: case 0xce: case 0xcf:
      return ((tag - BC_INT_BYTE_ZERO) << 8) + read();

    /* short int */
      case 0xd0: case 0xd1: case 0xd2: case 0xd3:
      case 0xd4: case 0xd5: case 0xd6: case 0xd7:
      return ((tag - BC_INT_SHORT_ZERO) << 16) + 256 * read() + read();

    //case LONG_BYTE:
      case BC_DOUBLE_BYTE:
        return (_offset < _length ? _buffer[_offset++] : read());

    //case INT_SHORT:
    //case LONG_SHORT:
      case BC_DOUBLE_SHORT:
        return  (256 * read() + read());

      case BC_INT://'I':
      case BC_LONG_INT:
        return _parseInt();

    // direct long
      case 0xd8: case 0xd9: case 0xda: case 0xdb:
      case 0xdc: case 0xdd: case 0xde: case 0xdf:

      case 0xe0: case 0xe1: case 0xe2: case 0xe3:
      case 0xe4: case 0xe5: case 0xe6: case 0xe7:
      case 0xe8: case 0xe9: case 0xea: case 0xeb:
      case 0xec: case 0xed: case 0xee: case 0xef:
      return tag - BC_LONG_ZERO;

    /* byte long */
      case 0xf0: case 0xf1: case 0xf2: case 0xf3:
      case 0xf4: case 0xf5: case 0xf6: case 0xf7:
      case 0xf8: case 0xf9: case 0xfa: case 0xfb:
      case 0xfc: case 0xfd: case 0xfe: case 0xff:
      return ((tag - BC_LONG_BYTE_ZERO) << 8) + read();

    /* short long */
      case 0x38: case 0x39: case 0x3a: case 0x3b:
      case 0x3c: case 0x3d: case 0x3e: case 0x3f:
      return ((tag - BC_LONG_SHORT_ZERO) << 16) + 256 * read() + read();

      case BC_LONG://'L':
        return _parseLong();

      case BC_DOUBLE_ZERO:
        return 0;

      case BC_DOUBLE_ONE:
        return 1;

      case BC_DOUBLE_MILL:{
        int mills = _parseInt();
        return (0.001 * mills).toInt();
      }

      case BC_DOUBLE://'D':
        return _parseDouble().toInt();

      case BC_FALSE://'F':
        return 0;

      case BC_TRUE:// 'T':
        return 1;
      default:
        throw HessianInputException("readLong:long needed", "$tag");
    }
  }

  /**
   * Reads a float
   *
   * <pre>
   * D b64 b56 b48 b40 b32 b24 b16 b8
   * </pre>
   */
  double readFloat() {
    return readDouble();
  }

  /**
   * Reads a double
   *
   * <pre>
   * D b64 b56 b48 b40 b32 b24 b16 b8
   * </pre>
   */
  double readDouble(){
    int tag = read();

    switch (tag) {
      case BC_NULL://'N':
        return 0.0;

      case BC_DOUBLE_ZERO:
        return 0.0;

      case BC_DOUBLE_ONE:
        return 1.0;

      case BC_DOUBLE_BYTE:
        return  (_offset < _length ? _buffer[_offset++] : read()).toDouble();

      case BC_DOUBLE_SHORT:
        return (256 * read() + read()).toDouble();

      case BC_DOUBLE_MILL: {
        int mills = _parseInt();
        return 0.001 * mills;
      }

      case BC_DOUBLE://'D':
        return _parseDouble();

    // direct integer
      case 0x80: case 0x81: case 0x82: case 0x83:
      case 0x84: case 0x85: case 0x86: case 0x87:
      case 0x88: case 0x89: case 0x8a: case 0x8b:
      case 0x8c: case 0x8d: case 0x8e: case 0x8f:

      case 0x90: case 0x91: case 0x92: case 0x93:
      case 0x94: case 0x95: case 0x96: case 0x97:
      case 0x98: case 0x99: case 0x9a: case 0x9b:
      case 0x9c: case 0x9d: case 0x9e: case 0x9f:

      case 0xa0: case 0xa1: case 0xa2: case 0xa3:
      case 0xa4: case 0xa5: case 0xa6: case 0xa7:
      case 0xa8: case 0xa9: case 0xaa: case 0xab:
      case 0xac: case 0xad: case 0xae: case 0xaf:

      case 0xb0: case 0xb1: case 0xb2: case 0xb3:
      case 0xb4: case 0xb5: case 0xb6: case 0xb7:
      case 0xb8: case 0xb9: case 0xba: case 0xbb:
      case 0xbc: case 0xbd: case 0xbe: case 0xbf:
      return (tag - 0x90).toDouble();

    /* byte int */
      case 0xc0: case 0xc1: case 0xc2: case 0xc3:
      case 0xc4: case 0xc5: case 0xc6: case 0xc7:
      case 0xc8: case 0xc9: case 0xca: case 0xcb:
      case 0xcc: case 0xcd: case 0xce: case 0xcf:
      return (((tag - BC_INT_BYTE_ZERO) << 8) + read()).toDouble();

    /* short int */
      case 0xd0: case 0xd1: case 0xd2: case 0xd3:
      case 0xd4: case 0xd5: case 0xd6: case 0xd7:
      return (((tag - BC_INT_SHORT_ZERO) << 16) + 256 * read() + read()).toDouble();

      case BC_INT://'I':
      case BC_LONG_INT:
        return _parseInt().toDouble();

    // direct long
      case 0xd8: case 0xd9: case 0xda: case 0xdb:
      case 0xdc: case 0xdd: case 0xde: case 0xdf:

      case 0xe0: case 0xe1: case 0xe2: case 0xe3:
      case 0xe4: case 0xe5: case 0xe6: case 0xe7:
      case 0xe8: case 0xe9: case 0xea: case 0xeb:
      case 0xec: case 0xed: case 0xee: case 0xef:
      return (tag - BC_LONG_ZERO).toDouble();

    /* byte long */
      case 0xf0: case 0xf1: case 0xf2: case 0xf3:
      case 0xf4: case 0xf5: case 0xf6: case 0xf7:
      case 0xf8: case 0xf9: case 0xfa: case 0xfb:
      case 0xfc: case 0xfd: case 0xfe: case 0xff:
      return (((tag - BC_LONG_BYTE_ZERO) << 8) + read()).toDouble();

    /* short long */
      case 0x38: case 0x39: case 0x3a: case 0x3b:
      case 0x3c: case 0x3d: case 0x3e: case 0x3f:
      return (((tag - BC_LONG_SHORT_ZERO) << 16) + 256 * read() + read()).toDouble();

      case BC_LONG://'L':
        return  _parseLong().toDouble();

      case BC_FALSE://'F':
        return 0.0;

      case BC_TRUE://'T':
        return 1.0;
      default:
        throw HessianInputException("readDouble:double needed", "$tag");
    }
  }

  /**
   * Reads a date.
   *
   * <pre>
   * T b64 b56 b48 b40 b32 b24 b16 b8
   * </pre>
   */
  int readUTCDate(){
    int tag = read();

    if (tag == BC_DATE) {
      return _parseLong();
    }
    else if (tag == BC_DATE_MINUTE) {
      return _parseInt() * 60000;
    }
    else
      throw HessianInputException("readUTCDate: date needed", "$tag");
  }


  /**
   * Reads a string
   *
   * <pre>
   * S b16 b8 string value
   * </pre>
   */
  String readString(){
    int tag = read();

    switch (tag) {
      case BC_STRING://'S':
      case BC_STRING_CHUNK:
        {
          _isLastChunk = tag == 'S';
          _chunkLength = (read() << 8) + read();
          int ch;

          int length = _chunkLength;
          int byteLen = _getStringByteCnt(length);
          if(byteLen == length){
            return _ascii.decode( new Uint8List.view(_buffer.buffer, _buffer.offsetInBytes + _offset - byteLen , byteLen));
          }
          return _utf8.decode( new Uint8List.view(_buffer.buffer, _buffer.offsetInBytes + _offset - byteLen , byteLen));
        }

    // 0-byte string
      case 0x00: case 0x01: case 0x02: case 0x03:
      case 0x04: case 0x05: case 0x06: case 0x07:
      case 0x08: case 0x09: case 0x0a: case 0x0b:
      case 0x0c: case 0x0d: case 0x0e: case 0x0f:

      case 0x10: case 0x11: case 0x12: case 0x13:
      case 0x14: case 0x15: case 0x16: case 0x17:
      case 0x18: case 0x19: case 0x1a: case 0x1b:
      case 0x1c: case 0x1d: case 0x1e: case 0x1f:
      {
        _isLastChunk = true;
        _chunkLength = tag - 0x00;
        int ch = 0;
        int length = _chunkLength;
        int byteLen = _getStringByteCnt(length);
        if(byteLen == length){
          return _ascii.decode( new Uint8List.view(_buffer.buffer, _buffer.offsetInBytes + _offset - byteLen , byteLen));
        }
        return _utf8.decode( new Uint8List.view(_buffer.buffer, _buffer.offsetInBytes + _offset - byteLen , byteLen));
      }

      case 0x30: case 0x31: case 0x32: case 0x33:
      {
        _isLastChunk = true;
        _chunkLength = (tag - 0x30) * 256 + read();
        int ch = 0;
        int length = _chunkLength;
        int byteLen = _getStringByteCnt(length);
        if(byteLen == length){
          return _ascii.decode( new Uint8List.view(_buffer.buffer, _buffer.offsetInBytes + _offset - byteLen , byteLen));
        }
        return _utf8.decode( new Uint8List.view(_buffer.buffer, _buffer.offsetInBytes + _offset - byteLen , byteLen));
      }

      case BC_NULL://'N':
        return null;
      case BC_TRUE://'T':
        return "true";
      case BC_FALSE://'F':
        return "false";

    // direct integer
      case 0x80: case 0x81: case 0x82: case 0x83:
      case 0x84: case 0x85: case 0x86: case 0x87:
      case 0x88: case 0x89: case 0x8a: case 0x8b:
      case 0x8c: case 0x8d: case 0x8e: case 0x8f:

      case 0x90: case 0x91: case 0x92: case 0x93:
      case 0x94: case 0x95: case 0x96: case 0x97:
      case 0x98: case 0x99: case 0x9a: case 0x9b:
      case 0x9c: case 0x9d: case 0x9e: case 0x9f:

      case 0xa0: case 0xa1: case 0xa2: case 0xa3:
      case 0xa4: case 0xa5: case 0xa6: case 0xa7:
      case 0xa8: case 0xa9: case 0xaa: case 0xab:
      case 0xac: case 0xad: case 0xae: case 0xaf:

      case 0xb0: case 0xb1: case 0xb2: case 0xb3:
      case 0xb4: case 0xb5: case 0xb6: case 0xb7:
      case 0xb8: case 0xb9: case 0xba: case 0xbb:
      case 0xbc: case 0xbd: case 0xbe: case 0xbf:
      return "${tag - 0x90}";

    /* byte int */
      case 0xc0: case 0xc1: case 0xc2: case 0xc3:
      case 0xc4: case 0xc5: case 0xc6: case 0xc7:
      case 0xc8: case 0xc9: case 0xca: case 0xcb:
      case 0xcc: case 0xcd: case 0xce: case 0xcf:
      return "${((tag - BC_INT_BYTE_ZERO) << 8) + read()}";

    /* short int */
      case 0xd0: case 0xd1: case 0xd2: case 0xd3:
      case 0xd4: case 0xd5: case 0xd6: case 0xd7:
      return "${((tag - BC_INT_SHORT_ZERO) << 16)+ 256 * read() + read()}";

      case BC_INT://'I':
      case BC_LONG_INT:
        return "${_parseInt()}";

    // direct long
      case 0xd8: case 0xd9: case 0xda: case 0xdb:
      case 0xdc: case 0xdd: case 0xde: case 0xdf:

      case 0xe0: case 0xe1: case 0xe2: case 0xe3:
      case 0xe4: case 0xe5: case 0xe6: case 0xe7:
      case 0xe8: case 0xe9: case 0xea: case 0xeb:
      case 0xec: case 0xed: case 0xee: case 0xef:
      return "${tag - BC_LONG_ZERO}";

    /* byte long */
      case 0xf0: case 0xf1: case 0xf2: case 0xf3:
      case 0xf4: case 0xf5: case 0xf6: case 0xf7:
      case 0xf8: case 0xf9: case 0xfa: case 0xfb:
      case 0xfc: case 0xfd: case 0xfe: case 0xff:
      return "${((tag - BC_LONG_BYTE_ZERO) << 8) + read()}";

    /* short long */
      case 0x38: case 0x39: case 0x3a: case 0x3b:
      case 0x3c: case 0x3d: case 0x3e: case 0x3f:
      return "${((tag - BC_LONG_SHORT_ZERO) << 16) + 256 * read() + read()}";

      case BC_LONG://'L':
        return "${_parseLong()}";

      case BC_DOUBLE_ZERO:
        return "0.0";

      case BC_DOUBLE_ONE:
        return "1.0";

      case BC_DOUBLE_BYTE:
        return "${(_offset < _length ? _buffer[_offset++] : read())}";

      case BC_DOUBLE_SHORT:
        return "${((256 * read() + read()))}";

      case BC_DOUBLE_MILL:
        {
          int mills = _parseInt();
          return "${0.001 * mills}";
        }

      case BC_DOUBLE://'D':
        return "${_parseDouble()}";
      default:
        throw HessianInputException("readString:string needed in readString", "tag");
    }
  }


  /**
   * skip parse a string
   *
   * <pre>
   * S b16 b8 string value
   * </pre>
   */
  String skipReadString(){
    int tag = read();

    switch (tag) {
      case BC_STRING://'S':
      case BC_STRING_CHUNK:
        {
          _isLastChunk = tag == 'S';
          _chunkLength = (read() << 8) + read();
          int ch;

          int length = _chunkLength;
          int byteLen = _getStringByteCnt(length);
          return "";
        }

    // 0-byte string
      case 0x00: case 0x01: case 0x02: case 0x03:
      case 0x04: case 0x05: case 0x06: case 0x07:
      case 0x08: case 0x09: case 0x0a: case 0x0b:
      case 0x0c: case 0x0d: case 0x0e: case 0x0f:

      case 0x10: case 0x11: case 0x12: case 0x13:
      case 0x14: case 0x15: case 0x16: case 0x17:
      case 0x18: case 0x19: case 0x1a: case 0x1b:
      case 0x1c: case 0x1d: case 0x1e: case 0x1f:
      {
        _isLastChunk = true;
        _chunkLength = tag - 0x00;
        int ch = 0;
        int length = _chunkLength;
        int byteLen = _getStringByteCnt(length);

        return "";
      }

      case 0x30: case 0x31: case 0x32: case 0x33:
      {
        _isLastChunk = true;
        _chunkLength = (tag - 0x30) * 256 + read();
        int ch = 0;

        int length = _chunkLength;
        int byteLen = _getStringByteCnt(length);
        return "";
      }
      case BC_NULL://'N':
        return null;
      case BC_TRUE://'T':
        return "true";
      case BC_FALSE://'F':
        return "false";

    // direct integer
      case 0x80: case 0x81: case 0x82: case 0x83:
      case 0x84: case 0x85: case 0x86: case 0x87:
      case 0x88: case 0x89: case 0x8a: case 0x8b:
      case 0x8c: case 0x8d: case 0x8e: case 0x8f:

      case 0x90: case 0x91: case 0x92: case 0x93:
      case 0x94: case 0x95: case 0x96: case 0x97:
      case 0x98: case 0x99: case 0x9a: case 0x9b:
      case 0x9c: case 0x9d: case 0x9e: case 0x9f:

      case 0xa0: case 0xa1: case 0xa2: case 0xa3:
      case 0xa4: case 0xa5: case 0xa6: case 0xa7:
      case 0xa8: case 0xa9: case 0xaa: case 0xab:
      case 0xac: case 0xad: case 0xae: case 0xaf:

      case 0xb0: case 0xb1: case 0xb2: case 0xb3:
      case 0xb4: case 0xb5: case 0xb6: case 0xb7:
      case 0xb8: case 0xb9: case 0xba: case 0xbb:
      case 0xbc: case 0xbd: case 0xbe: case 0xbf:
      return "${tag - 0x90}";

    /* byte int */
      case 0xc0: case 0xc1: case 0xc2: case 0xc3:
      case 0xc4: case 0xc5: case 0xc6: case 0xc7:
      case 0xc8: case 0xc9: case 0xca: case 0xcb:
      case 0xcc: case 0xcd: case 0xce: case 0xcf:
      return "${((tag - BC_INT_BYTE_ZERO) << 8) + read()}";

    /* short int */
      case 0xd0: case 0xd1: case 0xd2: case 0xd3:
      case 0xd4: case 0xd5: case 0xd6: case 0xd7:
      return "${((tag - BC_INT_SHORT_ZERO) << 16)+ 256 * read() + read()}";

      case BC_INT://'I':
      case BC_LONG_INT:
        return "${_parseInt()}";

    // direct long
      case 0xd8: case 0xd9: case 0xda: case 0xdb:
      case 0xdc: case 0xdd: case 0xde: case 0xdf:

      case 0xe0: case 0xe1: case 0xe2: case 0xe3:
      case 0xe4: case 0xe5: case 0xe6: case 0xe7:
      case 0xe8: case 0xe9: case 0xea: case 0xeb:
      case 0xec: case 0xed: case 0xee: case 0xef:
      return "${tag - BC_LONG_ZERO}";

    /* byte long */
      case 0xf0: case 0xf1: case 0xf2: case 0xf3:
      case 0xf4: case 0xf5: case 0xf6: case 0xf7:
      case 0xf8: case 0xf9: case 0xfa: case 0xfb:
      case 0xfc: case 0xfd: case 0xfe: case 0xff:
      return "${((tag - BC_LONG_BYTE_ZERO) << 8) + read()}";

    /* short long */
      case 0x38: case 0x39: case 0x3a: case 0x3b:
      case 0x3c: case 0x3d: case 0x3e: case 0x3f:
      return "${((tag - BC_LONG_SHORT_ZERO) << 16) + 256 * read() + read()}";

      case BC_LONG://'L':
        return "${_parseLong()}";

      case BC_DOUBLE_ZERO:
        return "0.0";

      case BC_DOUBLE_ONE:
        return "1.0";

      case BC_DOUBLE_BYTE:
        return "${(_offset < _length ? _buffer[_offset++] : read())}";

      case BC_DOUBLE_SHORT:
        return "${((256 * read() + read()))}";

      case BC_DOUBLE_MILL:
        {
          int mills = _parseInt();
          return "${0.001 * mills}";
        }

      case BC_DOUBLE://'D':
        return "${_parseDouble()}";
      default:
        throw HessianInputException("readString:string needed in readString", "tag");
    }
  }


  String codeName(int ch)
  {
    if (ch < 0)
      return "end of file";
    else
      return "0x${hex.encode([ch & 0xff])}(char)$ch" ;
  }

  bool _parseChunkLength(){
    if (_isLastChunk)
      return false;

    int code = _offset < _length ? (_buffer[_offset++] & 0xff) : read();

    switch (code) {
      case BC_STRING_CHUNK:
        _isLastChunk = false;
        _chunkLength = (read() << 8) + read();
        break;
      case BC_STRING://'S':
        _isLastChunk = true;
        _chunkLength = (read() << 8) + read();
        break;

      case 0x00: case 0x01: case 0x02: case 0x03:
      case 0x04: case 0x05: case 0x06: case 0x07:
      case 0x08: case 0x09: case 0x0a: case 0x0b:
      case 0x0c: case 0x0d: case 0x0e: case 0x0f:

      case 0x10: case 0x11: case 0x12: case 0x13:
      case 0x14: case 0x15: case 0x16: case 0x17:
      case 0x18: case 0x19: case 0x1a: case 0x1b:
      case 0x1c: case 0x1d: case 0x1e: case 0x1f:
      _isLastChunk = true;
      _chunkLength = code - 0x00;
      break;

      case 0x30: case 0x31: case 0x32: case 0x33:
      _isLastChunk = true;
      _chunkLength = (code - 0x30) * 256 + read();
      break;

      default:
        throw HessianInputException("_parseChunkLength:string needed", "$code" );
    }

    return true;
  }


  /**
   * Reads a byte array
   *
   * <pre>
   * B b16 b8 data value
   * </pre>
   */
  Uint8List readBytes() {
    int tag = read();

    switch (tag) {
      case BC_NULL://'N':
        return null;

      case BC_BINARY:
      case BC_BINARY_CHUNK:
        {
          _isLastChunk = tag == BC_BINARY;
          _chunkLength = (read() << 8) + read();
          List<int> tmpList = [];
          int data;
          while ((data = _parseByte()) >= 0)
            tmpList.add(data);

          return new Uint8List.fromList(tmpList);
        }

      case 0x20: case 0x21: case 0x22: case 0x23:
      case 0x24: case 0x25: case 0x26: case 0x27:
      case 0x28: case 0x29: case 0x2a: case 0x2b:
      case 0x2c: case 0x2d: case 0x2e: case 0x2f:
      {
        _isLastChunk = true;
        _chunkLength = tag - 0x20;
        Uint8List uint8list = new Uint8List(_chunkLength);
        int offset = 0;

        while (offset < _chunkLength) {
          int sublen = readBuffer(uint8list, 0, _chunkLength - offset);
          if (sublen <= 0)
            break;
          offset += sublen;
        }
        return uint8list;
      }

      case 0x34: case 0x35: case 0x36: case 0x37:
      {
        _isLastChunk = true;
        _chunkLength = (tag - 0x34) * 256 + read();
        Uint8List uint8list = new Uint8List(_chunkLength);
        int offset = 0;

        while (offset < _chunkLength) {
          int sublen = readBuffer(uint8list, 0, _chunkLength - offset);
          if (sublen <= 0)
            break;
          offset += sublen;
        }
        return uint8list;
      }
      default:
        throw HessianInputException("readBytes:bytes needed ", "$tag");
    }
  }

  /**
   * Reads bytes from the underlying stream.
   */
  int readBuffer(Uint8List buffer, int offset, int length)  {
    int readLength = 0;

    while (length > 0) {
      while (_chunkLength <= 0) {
        if (_isLastChunk)
          return readLength == 0 ? -1 : readLength;

        int code = read();

        switch (code) {
          case BC_BINARY_CHUNK:
            _isLastChunk = false;
            _chunkLength = (read() << 8) + read();
            break;

          case BC_BINARY:
            _isLastChunk = true;
            _chunkLength = (read() << 8) + read();
            break;

          case 0x20: case 0x21: case 0x22: case 0x23:
          case 0x24: case 0x25: case 0x26: case 0x27:
          case 0x28: case 0x29: case 0x2a: case 0x2b:
          case 0x2c: case 0x2d: case 0x2e: case 0x2f:
          _isLastChunk = true;
          _chunkLength = code - 0x20;
          break;

          case 0x34: case 0x35: case 0x36: case 0x37:
          _isLastChunk = true;
          _chunkLength = (code - 0x34) * 256 + read();
          break;

          default:
            throw HessianInputException("readBuffer:byte[] needed in readBytes", "$code");
        }
      }

      int sublen = _chunkLength;
      if (length < sublen)
        sublen = length;

      if (_length <= _offset )
        return -1;

      if (_length - _offset < sublen)
        sublen = _length - _offset;

//      System.arraycopy(_buffer, _offset, buffer, offset, sublen);
      buffer.setRange(offset, offset+sublen, _buffer,_offset);

      _offset += sublen;

      offset += sublen;
      readLength += sublen;
      length -= sublen;
      _chunkLength -= sublen;
    }

    return readLength;
  }


  /**
   * Reads a byte from the underlying stream.
   */
  int _parseByte() {
    while (_chunkLength <= 0) {
      if (_isLastChunk) {
        return -1;
      }

      int code = read();

      switch (code) {
        case BC_BINARY_CHUNK:
          _isLastChunk = false;
          _chunkLength = (read() << 8) + read();
          break;

        case BC_BINARY://'B':
          _isLastChunk = true;
          _chunkLength = (read() << 8) + read();
          break;

        case 0x20: case 0x21: case 0x22: case 0x23:
        case 0x24: case 0x25: case 0x26: case 0x27:
        case 0x28: case 0x29: case 0x2a: case 0x2b:
        case 0x2c: case 0x2d: case 0x2e: case 0x2f:
        _isLastChunk = true;
        _chunkLength = code - 0x20;
        break;

        case 0x34: case 0x35: case 0x36: case 0x37:
        _isLastChunk = true;
        _chunkLength = (code - 0x34) * 256 + read();
        break;

        default:
          throw HessianInputException("_parseByte byte[] needed", "$code");
      }
    }

    _chunkLength--;
    return read();
  }

  /**
   * Parses a 32-bit integer value from the stream.
   *
   * <pre>
   * b32 b24 b16 b8
   * </pre>
   */
  int _parseInt(){
//    int offset = _offset;
//
//    if (offset + 3 < _length) {
//      byte []buffer = _buffer;
//
//      int b32 = buffer[offset + 0] & 0xff;
//      int b24 = buffer[offset + 1] & 0xff;
//      int b16 = buffer[offset + 2] & 0xff;
//      int b8 = buffer[offset + 3] & 0xff;
//
//      _offset = offset + 4;
//
//      return (b32 << 24) + (b24 << 16) + (b16 << 8) + b8;
//    }
//    else {
//      int val = _byteData.getUint32(_offset);
//      _offset += 4;
//      return val;
    int b32 = read();
    int b24 = read();
    int b16 = read();
    int b8 = read();

    return (b32 << 24) + (b24 << 16) + (b16 << 8) + b8;
  }

  /**
   * Parses a 64-bit long value from the stream.
   *
   * <pre>
   * b64 b56 b48 b40 b32 b24 b16 b8
   * </pre>
   */
  int _parseLong() {

//    int val = _byteData.getUint64(_offset);
//    _offset += 8;
//    return val;

    int b64 = read();
    int b56 = read();
    int b48 = read();
    int b40 = read();
    int b32 = read();
    int b24 = read();
    int b16 = read();
    int b8 = read();

    return ((b64 << 56)
        + (b56 << 48)
        + (b48 << 40)
        + (b40 << 32)
        + (b32 << 24)
        + (b24 << 16)
        + (b16 << 8)
        + b8);
  }



  /**
   * Parses a type from the stream.
   *
   * <pre>
   * type ::= string
   * type ::= int
   * </pre>
   */
  String readType(){
    int code = _offset < _length ? (_buffer[_offset++] & 0xff) : read();
    _offset--;

    switch (code) {
      case 0x00: case 0x01: case 0x02: case 0x03:
      case 0x04: case 0x05: case 0x06: case 0x07:
      case 0x08: case 0x09: case 0x0a: case 0x0b:
      case 0x0c: case 0x0d: case 0x0e: case 0x0f:

      case 0x10: case 0x11: case 0x12: case 0x13:
      case 0x14: case 0x15: case 0x16: case 0x17:
      case 0x18: case 0x19: case 0x1a: case 0x1b:
      case 0x1c: case 0x1d: case 0x1e: case 0x1f:

      case 0x30: case 0x31: case 0x32: case 0x33:
      case BC_STRING_CHUNK:
      case BC_STRING://'S':
        {
          String type = readString();
//        if (_types == null){
//          _types =[];
//        }
          _types.add(type);
          return type;
        }

      default:
        {
          int ref = readInt();
          if (_types.length <= ref)
            throw new HessianInputException("readType:type ref #${ref} is greater than the number of valid types (${_types.length})","$ref");

          return _types[ref];
        }
    }
  }

  int _getStringByteCnt(int chunkLength){
    int length = chunkLength;

    int strLen = 0;

    while (length-- > 0  && _offset < _length) {
      int ch = _buffer[_offset++] & 0xff;

      if (ch < 0x80) {
        ++strLen;
      }
      else if ((ch & 0xe0) == 0xc0) {
        strLen += 2;
        _offset += 1;
      }
      else if ((ch & 0xf0) == 0xe0) {
        strLen += 3;
        _offset += 2;
      }
      else
        throw HessianInputException("getStringByteCnt: parse ch failed.","$ch at $_offset");
    }
    return strLen;
  }

  String _parseString() {
    while (true) {
      if (_chunkLength <= 0) {
        if (! _parseChunkLength())
          return "";
      }

      int length = _chunkLength;
      _chunkLength = 0;

//      while (length-- > 0) {
//        sbuf.write(new String.fromCharCode(parseUTF8Char()));
//      }
//      return sbuf.toString();
      int byteLen = _getStringByteCnt(length);
      if(byteLen == length){
        return _ascii.decode( new Uint8List.view(_buffer.buffer, _buffer.offsetInBytes + _offset - byteLen , byteLen));
      }
      return _utf8.decode( new Uint8List.view(_buffer.buffer, _buffer.offsetInBytes + _offset - byteLen , byteLen));
    }
  }

  /**
   * Reads an arbitrary object from the input stream when the type
   * is unknown.
   */
  dynamic readObject(){
    int tag = _offset < _length ? (_buffer[_offset++] & 0xff) : read();

    switch (tag) {
      case BC_NULL://'N':
        return null;

      case BC_TRUE://'T':
        return true;

      case BC_FALSE://'F':
        return false;

    // direct integer
      case 0x80: case 0x81: case 0x82: case 0x83:
      case 0x84: case 0x85: case 0x86: case 0x87:
      case 0x88: case 0x89: case 0x8a: case 0x8b:
      case 0x8c: case 0x8d: case 0x8e: case 0x8f:

      case 0x90: case 0x91: case 0x92: case 0x93:
      case 0x94: case 0x95: case 0x96: case 0x97:
      case 0x98: case 0x99: case 0x9a: case 0x9b:
      case 0x9c: case 0x9d: case 0x9e: case 0x9f:

      case 0xa0: case 0xa1: case 0xa2: case 0xa3:
      case 0xa4: case 0xa5: case 0xa6: case 0xa7:
      case 0xa8: case 0xa9: case 0xaa: case 0xab:
      case 0xac: case 0xad: case 0xae: case 0xaf:

      case 0xb0: case 0xb1: case 0xb2: case 0xb3:
      case 0xb4: case 0xb5: case 0xb6: case 0xb7:
      case 0xb8: case 0xb9: case 0xba: case 0xbb:
      case 0xbc: case 0xbd: case 0xbe: case 0xbf:
      return (tag - BC_INT_ZERO);

    /* byte int */
      case 0xc0: case 0xc1: case 0xc2: case 0xc3:
      case 0xc4: case 0xc5: case 0xc6: case 0xc7:
      case 0xc8: case 0xc9: case 0xca: case 0xcb:
      case 0xcc: case 0xcd: case 0xce: case 0xcf:
      return (((tag - BC_INT_BYTE_ZERO) << 8) + read());

    /* short int */
      case 0xd0: case 0xd1: case 0xd2: case 0xd3:
      case 0xd4: case 0xd5: case 0xd6: case 0xd7:
      return (((tag - BC_INT_SHORT_ZERO) << 16) + 256 * read() + read());

      case BC_INT://'I':
        return _parseInt();

    // direct long
      case 0xd8: case 0xd9: case 0xda: case 0xdb:
      case 0xdc: case 0xdd: case 0xde: case 0xdf:

      case 0xe0: case 0xe1: case 0xe2: case 0xe3:
      case 0xe4: case 0xe5: case 0xe6: case 0xe7:
      case 0xe8: case 0xe9: case 0xea: case 0xeb:
      case 0xec: case 0xed: case 0xee: case 0xef:
      return(tag - BC_LONG_ZERO);

    /* byte long */
      case 0xf0: case 0xf1: case 0xf2: case 0xf3:
      case 0xf4: case 0xf5: case 0xf6: case 0xf7:
      case 0xf8: case 0xf9: case 0xfa: case 0xfb:
      case 0xfc: case 0xfd: case 0xfe: case 0xff:
      return (((tag - BC_LONG_BYTE_ZERO) << 8) + read());

    /* short long */
      case 0x38: case 0x39: case 0x3a: case 0x3b:
      case 0x3c: case 0x3d: case 0x3e: case 0x3f:
      return (((tag - BC_LONG_SHORT_ZERO) << 16) + 256 * read() + read());

      case BC_LONG_INT:
        return _parseInt();

      case BC_LONG://'L':
        return _parseLong();

      case BC_DOUBLE_ZERO:
        return 0.0;

      case BC_DOUBLE_ONE:
        return 1.0;//Double.valueOf(1);

      case BC_DOUBLE_BYTE:
        return read().toDouble();

      case BC_DOUBLE_SHORT:
        return ((256 * read() + read())).toDouble();

      case BC_DOUBLE_MILL:
        {
          int mills = _parseInt();
          return (0.001 * mills);
        }

      case BC_DOUBLE://'D':
        return _parseDouble();

      case BC_DATE:
        return new DateTime.fromMillisecondsSinceEpoch(_parseLong());

      case BC_DATE_MINUTE:
        return new DateTime.fromMillisecondsSinceEpoch(_parseInt() * 60000);

      case BC_STRING_CHUNK:
      case BC_STRING://'S':
        {
          _isLastChunk = tag == 'S';
          _chunkLength = (read() << 8) + read();
          return _parseString();
        }

      case 0x00: case 0x01: case 0x02: case 0x03:
      case 0x04: case 0x05: case 0x06: case 0x07:
      case 0x08: case 0x09: case 0x0a: case 0x0b:
      case 0x0c: case 0x0d: case 0x0e: case 0x0f:

      case 0x10: case 0x11: case 0x12: case 0x13:
      case 0x14: case 0x15: case 0x16: case 0x17:
      case 0x18: case 0x19: case 0x1a: case 0x1b:
      case 0x1c: case 0x1d: case 0x1e: case 0x1f:
      {
        _isLastChunk = true;
        _chunkLength = tag - 0x00;
        return _parseString();
      }

      case 0x30: case 0x31: case 0x32: case 0x33:
      {
        _isLastChunk = true;
        _chunkLength = (tag - 0x30) * 256 + read();
        return  _parseString();
      }

      case BC_BINARY_CHUNK:
      case BC_BINARY://'B':
        {
          _isLastChunk = tag == 'B';
          _chunkLength = (read() << 8) + read();

          int data;
          List<int> bos = [];

          while ((data = _parseByte()) >= 0)
            bos.add(data);

          return new Uint8List.fromList(bos);
        }

      case 0x20: case 0x21: case 0x22: case 0x23:
      case 0x24: case 0x25: case 0x26: case 0x27:
      case 0x28: case 0x29: case 0x2a: case 0x2b:
      case 0x2c: case 0x2d: case 0x2e: case 0x2f:
      {
        _isLastChunk = true;
        int len = tag - 0x20;
        _chunkLength = 0;

//        byte []data = new byte[len];
        Uint8List data = new Uint8List(len);
        for (int i = 0; i < len; i++){
          data[i] = read();
        }
        return data;
      }

      case 0x34: case 0x35: case 0x36: case 0x37:
      {
        _isLastChunk = true;
        int len = (tag - 0x34) * 256 + read();
        _chunkLength = 0;

        //byte []buffer = new byte[len];
        Uint8List buffer = new Uint8List(len);
        for (int i = 0; i < len; i++) {
          buffer[i] = read();
        }
        return buffer;
      }

      case BC_LIST_VARIABLE:
        {
          // variable length list
          String type = readType();
          return readList(-1);
        }

      case BC_LIST_VARIABLE_UNTYPED:
        {
          return readList(-1);
        }

      case BC_LIST_FIXED:
        {
          // fixed length lists
          String type = readType();
          int length = readInt();

//        Deserializer reader;
//        reader = findSerializerFactory().getListDeserializer(type, null);

          return readLengthList(length);
        }

      case BC_LIST_FIXED_UNTYPED:
        {
          // fixed length lists
          int length = readInt();

//        Deserializer reader;
//        reader = findSerializerFactory().getListDeserializer(null, null);

          return readLengthList(length);
        }

    // compact fixed list
      case 0x70: case 0x71: case 0x72: case 0x73:
      case 0x74: case 0x75: case 0x76: case 0x77:
      {
        // fixed length lists
        String type = readType();
        int length = tag - 0x70;

//        Deserializer reader;
//        reader = findSerializerFactory().getListDeserializer(type, null);

        return readLengthList(length);
      }

    // compact fixed untyped list
      case 0x78: case 0x79: case 0x7a: case 0x7b:
      case 0x7c: case 0x7d: case 0x7e: case 0x7f:
      {
        // fixed length lists
        int length = tag - 0x78;

//        Deserializer reader;
//        reader = findSerializerFactory().getListDeserializer(null, null);

        return readLengthList(length);
      }

      case BC_MAP_UNTYPED ://'H':
        {
          return readMap();
        }

      case BC_MAP://'M':
        {
          String type = readType();

          return readMap();
        }

      case BC_CLASS_DEF://'C':
        {
          ClassDefinition def = readObjectDefinition();

          return readObject();
        }

      case 0x60: case 0x61: case 0x62: case 0x63:
      case 0x64: case 0x65: case 0x66: case 0x67:
      case 0x68: case 0x69: case 0x6a: case 0x6b:
      case 0x6c: case 0x6d: case 0x6e: case 0x6f:
      {
        int ref = tag - 0x60;

        if (_classDefs.length <= ref)
          throw HessianInputException("readObject:No classes defined at reference '${hex.encode([tag])}'", "$tag");

        ClassDefinition def = _classDefs[ref];

        return readObjectInstance( def);
      }

      case BC_OBJECT://'O':
        {
          int ref = readInt();

          if (_classDefs.length <= ref)
            throw HessianInputException("readObject:Illegal object reference #$ref" ,"$ref" );

          ClassDefinition def = _classDefs[ref];

          return readObjectInstance(def);
        }

      case BC_REF:
        {
          int ref = readInt();
          return _refs[ref];
        }

      default:
        if (tag < 0)
          throw new HessianInputException("readObject: unexpected end of file","$tag $_offset");
        else
          throw new HessianInputException("readObject: unknown code ${ codeName(tag)}","$tag  $_offset" );
    }
  }

  Map readObjectInstance( ClassDefinition def){
    //忽略复杂的类定义吧。少年。
    Map obj = new Map();
    addRef(obj);
    for(int idx = 0; idx< def.fieldNames.length;++idx){
      dynamic value = readObject();
      if(value != null)
        obj[def.fieldNames[idx]] = value;
    }

    return obj;
  }


  /**
   * Reads an object definition:
   *
   * <pre>
   * O string <int> (string)* <value>*
   * </pre>
   */
  ClassDefinition readObjectDefinition(){
    String type = readString();
    int len = readInt();

//    ClassDefinition classDefinition = new ClassDefinition()..fieldNames=[];
//    for (int i = 0; i < len; i++) {
//      String name = readString();
//      classDefinition.fieldNames.add(name);
//    }

    ClassDefinition classDefinition = safeCast(cache.get("${_remoteIP}_${type}_$len"));//类定义缓存，key:ip_type_len尽量发现变动重新加载
    if(_remoteIP == "" || classDefinition == null){
      classDefinition = new ClassDefinition()..fieldNames=[];
      for (int i = 0; i < len; i++) {
        String name = readString();
        classDefinition.fieldNames.add(name);
      }
      cache.set("${_remoteIP}_${type}_$len", classDefinition);
    }else{
      //如果类定义已经有了。就跳过类字段的解析，只读取，不解析
      for (int i = 0; i < len; i++) {
        skipReadString();
      }
    }

    _classDefs.add(classDefinition);
    return classDefinition;
  }

  /**
   * Reads a fault.
   */
  Map readFault(){
    Map map = {};

    int code = read();
    for (; code > 0 && code != BC_END/*'Z'*/; code = read()) {
      _offset--;

      Object key = readObject();
      Object value = readObject();

      if (key != null && value != null)
        map[key] = value;
    }

    if (code != BC_END/*'Z'*/)
      throw HessianInputException("readFault: fault not end", "$code");

    return map;
  }

  Map readMap(){
    Map map = new Map();

//    if (_type == null)
//      map = new HashMap();
//    else if (_type.equals(Map.class))
//      map = new HashMap();
//    else if (_type.equals(SortedMap.class))
//      map = new TreeMap();
//    else {
//      try {
//        map = (Map) _ctor.newInstance();
//      } catch (Exception e) {
//        throw new IOExceptionWrapper(e);
//      }
//    }

    addRef(map);

    while (! isEnd()) {
      map[readObject()] = readObject();
    }

    readEnd();
    return map;
  }

  /**
   * Reads the end byte.
   */
  void readEnd(){
    int code = _offset < _length ? (_buffer[_offset++] & 0xff) : read();

    if (code == BC_END)//'Z'
      return;
    else if (code < 0)
      throw HessianInputException("readEnd():unexpected end of file", "$code");
    else
      throw HessianInputException("readEnd():unknown code:${codeName(code)}" ,"" );
  }

  /**
   * Adds a list/map reference.
   */
  int addRef(Object ref)  {
//    if (_refs == null)
//      _refs = [];

    _refs.add(ref);
    return _refs.length - 1;
  }

  /**
   * Adds a list/map reference.
   */
  void setRef(int i, Object ref)
  {
    _refs[i] = ref;
  }

  /**
   * Reads the end byte.
   */
  void readListEnd()  {
    int code = _offset < _length ? (_buffer[_offset++] & 0xff) : read();

    if (code != BC_END/*'Z'*/)
      throw HessianInputException("readListEnd():expected end of list ('Z') at '${codeName(code)}'","$code");
  }


  /**
   * Reads the array.
   */
  List readList( int length)  {
    if (length >= 0) {
      List data = new List(length);

      addRef(data);
//    注释：hessian解析为map类型就可以了。所以不再有类型的拖累
//      if (_componentType != null) {
//        for (int i = 0; i < data.length; i++)
//          data[i] = in.readObject(_componentType);
//      }
//      else {
//        for (int i = 0; i < data.length; i++)
//          data[i] = in.readObject();
//      }
      for (int i = 0; i < length; i++)
        data[i] = readObject();

      readListEnd();
      return data;
    }
    else {
      List list = [];

      addRef(list);
//     hessian解析为json，不再有类型的拖累
//      if (_componentType != null) {
//        while (! in.isEnd())
//          list.add(in.readObject(_componentType));
//      }
//      else {
//        while (! in.isEnd())
//          list.add(in.readObject());
//      }
      while (! isEnd())
        list.add(readObject());

      readListEnd();
      return list;
//      Object []data = createArray(list.size());
//      for (int i = 0; i < data.length; i++)
//        data[i] = list.get(i);
//
//      return data;
    }
  }


  /**
   * Reads the array.
   */
  List readLengthList(int length){
    List data = new List(length);
    addRef(data);

//    if (_componentType != null) {
//      for (int i = 0; i < data.length; i++)
//        data[i] = in.readObject(_componentType);
//    }
//    else {
//      for (int i = 0; i < data.length; i++)
//        data[i] = in.readObject();
//    }
    for (int i = 0; i < length; i++)
      data[i] = readObject();
    return data;
  }

  /**
   * Returns true if this is the end of a list or a map.
   */
  bool isEnd(){
    int code;

    if (_offset < _length)
      code = (_buffer[_offset] & 0xff);
    else {
      code = read();

      if (code >= 0)
        _offset--;
    }

    return (code < 0 || code == BC_END/*'Z'*/);
  }

  /**
   * Parses a 64-bit double value from the stream.
   *
   * <pre>
   * b64 b56 b48 b40 b32 b24 b16 b8
   * </pre>
   */
  double _parseDouble() {
    int sizeInBytes = 8;

    double tmp_val = _byteData.getFloat64(_offset);
    _offset += 8;
    return tmp_val;
//    int bits = _parseLong();
//    return Double.longBitsToDouble(bits);
  }

  /**
   * Normally, shouldn't be called externally, but needed for QA, e.g.
   * ejb/3b01.
   */
  int read() {
    if (_length <= _offset )
      return -1;

    return _buffer[_offset++] & 0xff;
  }

  void reset(){
    _offset = 0;
  }
  int _length;
  int _offset;
  Uint8List _buffer;

  // true if this is the last chunk
  bool _isLastChunk;
  // the chunk length
  int _chunkLength;
  int _stringLength;
//  StringBuffer _sbuf = new StringBuffer();
  List<String> _types = [];
  List<Object> _refs = [];
  List<ClassDefinition> _classDefs=[];
  ByteData _byteData;

  String _remoteIP;

}

Cache cache = new LruCache(storage: new SimpleStorage(size: 1000))
  ..expiration = new Duration(seconds: 60)
  ..syncLoading = true
  ..loader = (key, value){return null;}
;

const _utf8 = const Utf8Codec(allowMalformed: true);
const _ascii = const AsciiCodec(allowInvalid: true);
class ClassDefinition{
  List<String> fieldNames;
}