import 'dart:async';
import 'dart:convert';
import 'dart:typed_data';

import 'package:saker/common/utils/http_util.dart';
import 'package:saker/framework/environment/env_base.dart';
import 'package:saker/framework/event/event_pool.dart';
import 'package:saker/framework/event/event_pool_base.dart';
import 'package:saker/framework/event/notice_type.dart';
import 'package:saker/framework/hsf/hsf_async.dart';
import 'package:saker/framework/log/log.dart';
import 'package:saker/common/net/ip_address.dart';
import 'package:saker/framework/log/log_type.dart';
import 'package:saker/framework/services/mtop_context.dart';
import 'package:saker/framework/services/service_process.dart';
import 'package:saker/framework/services/service_var.dart';
import 'package:saker/framework/stat/service_flow_limit.dart';
import 'package:saker/framework/stat/service_monitor.dart';
import 'package:saker/framework/trace/trace_info.dart';

export 'package:saker/framework/jsonserializer/utils.dart';

abstract class DProviderBase extends ServiceFlowLimit {
  int envValue = 0;

  String serviceRegisterName;
  Map<String,dynamic> serviceRegisterInfo;

  Future<Uint8List> callServiceMethod(String methodName, String paramArray, String argSign);

  Map<String,dynamic> getServiceRegisterInfo() {
    return serviceRegisterInfo;
  }

  String getServiceRegisterName() {
    return serviceRegisterName;
  }
}

class HsfProviderBase {
  static Map<String, DProviderBase> servicesMap = {};
  static Map<String, dynamic> servicesRegister = {};

  static addProvider(DProviderBase base) {
    if (EnvFramework.isFaaS) {
      MyLogger.getDebugLogger()?.debug('HSF Provider discard because is in fass env, ${servicesMap}');
      return;
    }
    //初始化流量统计器
    base.registerStatistic();

    servicesMap[base.getServiceRegisterName()] = base;
    servicesRegister.addAll(base.getServiceRegisterInfo());

    MyLogger.getDebugLogger()?.debug('${servicesMap}');
  }

  static Future<dynamic> dispatch(List<dynamic> element) async {
    String user_data = element[5] as String;
    String hsf_req_props = element[8] as String;

    Map<String, dynamic> userData = {};
    if(user_data != null && user_data.length > 3) {
      Map<String, dynamic> m = (json.decode(user_data) as Map)?.cast();
      userData.addAll(m);
    }
    if (hsf_req_props != null && hsf_req_props.length > 3) {
      Map<String, dynamic> context = (json.decode(hsf_req_props) as Map)?.cast();
      String e = context['eagleeyeUserData'] as String;
      if (e != null && e.isNotEmpty) {
        userData.addAll(Uri.splitQueryString(e));
      }
    }

    String tag = null;
    if (userData.containsKey('dpath_env')) {
      tag = userData['dpath_env'] as String;
    }

    if (userData.isNotEmpty) {
      element[5] = makeQuery(userData);
    }

    if (tag != null && tag.isNotEmpty && tag != 'DPathBaseEnv') {
      if (bizIsolateMap.containsKey(tag)) {
        EventItem eventItem = EventPool.getEvent();
        bizIsolateMap[tag].send([EventType_Notice, [
          NoticeType.BizHsfReq,
          eventItem.sendPort,
          eventItem.id,
          element
        ]
        ]);
        var res = await eventItem.future.timeout(BIZ_REQ_TIMEOUT)
            .catchError((e) => MyLogger.getLogger().error('DispatchHsf error:event error:$e'));
        return res;
      } else {
        MyLogger.getDebugLogger()?.debug('DispatchHsf error:req=${json.encode(element)},tag=${tag}, tag not exist');
      }
    }
    return processHsfDispatch(element, false);
  }

  static Future<dynamic> processHsfDispatch(List<dynamic> element, bool inReloadIsolate) async {
    MyLogger.getDebugLogger()?.debug("ProcessHsfDispatch req raw data:${json.encode(element)},len:${element.length}");

    if (element.length == 9) {
      String svc_name = element[0] as String;
      String method_name = element[1] as String;
      String req = element[2] as String;
      String trace_id = element[3] as String;
      String rpc_id = element[4] as String;
      String user_data = element[5] as String;
      String arg_signs = element[6] as String;
      int context_addr = element[7] as int;
      String hsf_req_props = element[8] as String;

      MyLogger.getDebugLogger()?.debug(
          "ProcessHsfDispatch req: $svc_name $method_name, $req, $trace_id, $rpc_id, $context_addr, userdata:$user_data, arg_sign:$arg_signs  hsf_req_props:$hsf_req_props");

      TraceInfo traceInfo = new TraceInfo();
      traceInfo.eagleeyeIndex = 0;
      traceInfo.eagleeyeRpcid = rpc_id;
      traceInfo.eagleeyeTraceid = trace_id;
      traceInfo.inReloadIsolate = inReloadIsolate;
      traceInfo.mtopPath = "${svc_name}_$method_name";
      if (user_data != null && user_data.isNotEmpty) {
        Map<String,String> user_data_map = Uri.splitQueryString(user_data);
        if (user_data_map.containsKey('t')) {
          traceInfo.pressTestTag = user_data_map['t'];
        }
        traceInfo.eagleeyeContext = user_data_map;
      }
      MtopContext mtopContext = null;
      if (hsf_req_props != null && hsf_req_props.isNotEmpty) {
        Map<String, String> rpcContext = (json.decode(hsf_req_props) as Map)?.map((k, v) => new MapEntry<String, String>(k?.toString(), v?.toString()));
        if (rpcContext != null) {
          mtopContext = new MtopContext(rpcContext, {});
          traceInfo.eagleeyeUserData = rpcContext['eagleeyeUserData'];
        } else {
          mtopContext = MtopContext.EMPTY;
        }
      }

      traceInfo.serviceMonitor = new ServiceMonitor()
        ..traceId = traceInfo.eagleeyeTraceid
        ..rpcId = traceInfo.eagleeyeRpcid
        ..ttid =  ""
        ..req = req
        ..reqMethod = "${svc_name}_$method_name"
        ..mtopPath = "${svc_name}_$method_name"
        ..localIP = IPAddress.getHostIp()
        ..remoteIP = ""
        ..rt = new DateTime.now().millisecondsSinceEpoch
        ..logType = LogType.HsfSvc;


      bool successFlag = true;
      String errCode = "";
      int resp_size = 0;
      Uint8List response = null;
      if (servicesMap.containsKey(svc_name)) {
        DProviderBase providerBase = servicesMap[svc_name];
        if (providerBase.flowLimitBlock()) {
          MyLogger.getLogger().error("HSF Flow limit actived, ${providerBase.getServicePath()}, eagleeyeTraceid=${traceInfo.eagleeyeTraceid},eagleeyeRpcid=${traceInfo.eagleeyeRpcid}.");
          hsf_send_respone(context_addr, "FAIL_SYS_FLOWLIMIT", Uint8List.fromList([]));
          return 0;
        }

        runZoned(() async {
          try {
            response = await providerBase.callServiceMethod(method_name, req, arg_signs)
                .catchError((e, stackTrace) {
              successFlag = false;
              errCode = "ProcessHsfDispatch failed.  $e, \n${stackTrace}";
              MyLogger.getLogger().error("${errCode}");
            });

            if (response != null) {
              MyLogger.getDebugLogger()?.debug("ProcessHsfDispatch resp len: ${response?.length} succ");
              resp_size = response.length;
              errCode = "";
            }
          } catch (e, stackTrace) {
            successFlag = false;
            errCode = "ProcessHsfDispatch failed.  $e, \n${stackTrace}";
            MyLogger.getLogger().error("${errCode}");
          }
        }, zoneValues: {#TraceInfo: traceInfo, #MtopContext:mtopContext},
            onError: (e, stackTrace) {
              successFlag = false;
              errCode = "ProcessHsfDispatch failed.  $e, \n${stackTrace}";
              MyLogger.getLogger().error("${errCode}");
            }
        ).whenComplete((){
          if(traceInfo.serviceMonitor != null ){
            traceInfo.serviceMonitor.success = successFlag;
            traceInfo.serviceMonitor.errCode = errCode;
            if(traceInfo.serviceMonitor.respSize == null){//如果业务侧自己定义了result长度，框架不再修改
              traceInfo.serviceMonitor.respSize = resp_size;
            }
            traceInfo.serviceMonitor.flush();
          }
          if(response == null){
            response = Uint8List.fromList([]);
          }
          hsf_send_respone(context_addr, errCode, response);
        });
      } else {
        MyLogger.getLogger().error("ProcessHsfDispatch req failed to find service in servicesMap: $svc_name, map:${servicesMap.keys}");
      }
    }
    return 0;
  }
}