import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:saker/common/net/ip_address.dart';
import 'package:saker/common/utils/random_utils.dart';
import 'package:saker/common/utils/type_cast.dart';
import 'package:saker/framework/config/frame_work_config_manager.dart';
import 'package:saker/framework/hsf/hsf_config_param.dart';
import 'package:saker/framework/event/event_pool.dart';
import 'package:saker/framework/event/event_pool_base.dart';
import 'package:saker/framework/hsf/hsf_info.dart';
import 'package:saker/framework/hsf/hsf_async.dart';
import 'package:saker/framework/log/log_type.dart';
import 'package:saker/framework/stat/service_monitor.dart';
import 'package:saker/framework/trace/trace_info.dart';
import 'package:saker/framework/log/log.dart';
import 'package:saker/framework/hsf/hsf_filter.dart';
import 'package:saker/framework/traffic/traffic_labour.dart';
import 'package:saker/framework/traffic/traffic_protocol.dart';

const Duration defaultHsfTimeOut = const Duration(seconds: 4);

//调用方式
const int hsfInvokeTypeMask = 0xF;
const int hsfInvokeTypeNative = 0x0;
const int hsfInvokeTypeMesh = 0x1;
const int hsfInvokeTypeHttp = 0x2;
const int hsfInvokeTypeHttp2 = 0x3;
const int hsfInvokeTypeDefault = hsfInvokeTypeNative;

//序列化方式
const int hsfEncodeTypeMask = 0xF0;
const int hsfEncodeTypeHessian = 0x00;
const int hsfEncodeTypeHessian2 = 0x10;
const int hsfEncodeTypeJson = 0x20;
const int hsfEncodeTypeHessianJson = 0x30;
const int hsfEncodeTypeDefault = hsfEncodeTypeHessian;

//透传MtopContext
const int hsfContextTransfer = 0x100;

const int defaultInvokeFlag = hsfEncodeTypeDefault | hsfInvokeTypeDefault;

int getHsfInvokeFlag({int invokeType = hsfInvokeTypeDefault, int encodeType = hsfEncodeTypeHessianJson, int contextTransfer = 0}) {
  return invokeType | encodeType | contextTransfer;
}

class HsfRPCBase {
  HsfFilter hsfFilter = null;
  HsfInfo hsfInfo = null;

  String defaultInvokeName = '';
  String defaultInvokePort = '';
  bool defaultInvokeIsIp = false;

  String servicePath;

  void initHsfConsumeBase(HsfConfigParam config) {
    hsfFilter = config?.filter;
    hsfInfo.beanName = config?.beanName;
    String timeoutMap = json.encode(config?.timeoutConfig??{});
    hsfInfo.serviceAddress = hsf_create_consumer(hsfInfo.consumerName, hsfInfo.version, hsfInfo.group, hsfInfo.aliasName, timeoutMap);
  }

  Future<String> invoke(String rpcMethodName, String dartServiceMethod, String paramStr, List<String> paramTypes, List<Uint8List> params, Duration hsfTimeOut, String invokeIp, int invokeFlag) async {
    int invokeType = invokeFlag & hsfInvokeTypeMask;
    int encodeType = invokeFlag & hsfEncodeTypeMask;
    bool transferContext = (invokeFlag & hsfContextTransfer) > 0;

//    if (hsfFilter != null && hsfFilter.isDegrade('${hsfInfo.consumerName}_${hsfInfo.version}', rpcMethodName)) {
//      MyLogger.getLogger().error('HSF Degrade:$dartServiceMethod MtopPath=' + TraceInfo.getMtopPath());
//      return null;
//    }
    if (hsfFilter != null && hsfFilter.isDegrade(hsfInfo.beanName, rpcMethodName)) {
      MyLogger.getLogger().error('HSF Degrade:$rpcMethodName beanName=${hsfInfo.beanName}');
      return null;
    }
    
    MyLogger.getDebugLogger()?.debug('HSF REQ:$dartServiceMethod param=${paramStr}, rpc type:$invokeType, encodeType:$encodeType');

    bool httpInvoke = hsfInvokeTypeHttp == invokeType || hsfInvokeTypeHttp2 == invokeType;
    if (!httpInvoke && invokeIp.isEmpty && defaultInvokeName.isNotEmpty) {
      //通过默认IP调用
      if (defaultInvokeIsIp) {
        invokeIp = '$defaultInvokeName:$defaultInvokePort';
      } else {
        var address = await InternetAddress.lookup(defaultInvokeName, type: InternetAddressType.IPv4).catchError((e) {
          MyLogger.getLogger().error('HSF LOOKUP HOST Error:$dartServiceMethod e=${e}');
        });
        if (address == null || address.isEmpty) {
          MyLogger.getLogger().error('HSF LOOKUP HOST Error:$dartServiceMethod address is empty');
          return null;
        }
        invokeIp = '${address[0].address}:$defaultInvokePort';
      }
    }

    String traceId = TraceInfo.getEagleeyeTraceId();
    String rpcId = TraceInfo.getEagleeyeRpcId();
    int startTime = new DateTime.now().millisecondsSinceEpoch;
    ServiceMonitor providerMonitor = new ServiceMonitor()
      ..traceId = traceId
      ..rpcId = rpcId
      ..reqMethod = "$dartServiceMethod"
      ..req = ''
      ..localIP = IPAddress.getHostIp()
      ..ttid = TraceInfo.getttid()
      ..success = true
      ..remoteIP = ""
      ..mtopPath = TraceInfo.getMtopPath()
      ..rt = startTime
      ..respSize = 0
      ..logType = LogType.RpcCall;

    String r = null;
    CollectEntity entity = new CollectEntity()
      ..mainInvoke = false
      ..invokeId   = rpcMethodName
      ..params     = paramStr == null?paramStr:params //变成二进制数据可能会有问题
      ..traceId    = traceId
      ..invokeCat  = REQUEST_CAT.HSF;
    var inspect = TrafficLabour.inspect(entity);
    r = safeCast((inspect is Future) ? await inspect : inspect);

    if(r == null) {
      if (!httpInvoke) {
        int retry_idx = 0;
        for (retry_idx = 0; retry_idx < 2; retry_idx++) {
          bool continue_flag = false;
          providerMonitor.success = true;
          providerMonitor.errCode = null;
          EventItem eventItem = EventPool.getEvent();

          MyLogger.getDebugLogger()?.trace('HSF Switch ServiceMeshRate = ${FrameWorkSwitchManager.serviceMeshRate}');
          //如果配置不为0，则按比例切分到mesh，否则关闭按比例切分，invokeType走原来的
          if(FrameWorkSwitchManager.serviceMeshRate != 0 && RandTools.nextInt(101) < FrameWorkSwitchManager.serviceMeshRate){
            invokeType = hsfInvokeTypeMesh;
          }

          //获取zone中的MtopContext
          String rpcContextJson = "";
          if(transferContext || TraceInfo.isInReloadIsolate()) { //TODO:isInReloadIsolate以后不用
            rpcContextJson = TraceInfo.getRpcContextHessianJson();
          }
          int code = hsf_consumer_invoke( hsfInfo.serviceAddress, rpcMethodName, eventItem.id, traceId, rpcId, invokeIp, hsfResultAddress, paramStr,
              paramTypes, params, eagleeyeCtxHessiaanJson: TraceInfo.getEagleeyeCtxJson(), invokeType: invokeType, encodeType: encodeType, rpcContext: rpcContextJson);

          if (code != 0) {
            EventPool.remove(eventItem.id);
            MyLogger.getLogger().error( 'HSF ERROR:$dartServiceMethod RpcID:$rpcId errcode=${code} param=${params} ${paramStr}');

            providerMonitor.success = false;
            providerMonitor.errCode = code?.toString();
          } else {
            r = safeCast(await eventItem.future
                .timeout(hsfTimeOut, onTimeout: () {
                  EventPool.remove(eventItem.id);
                  MyLogger.getLogger().error('HSF TIMEOUT:$dartServiceMethod RpcID:$rpcId EventID=${eventItem.id} Time=${new DateTime.now().millisecondsSinceEpoch -startTime} param=${params} ${paramStr}');
                  providerMonitor.success = false;
                  providerMonitor.errCode = "TIMEOUT";
                })
                .catchError((e) async {
                  int dur = new DateTime.now().millisecondsSinceEpoch - startTime;
                  MyLogger.getLogger().error('HSF ERROR:$dartServiceMethod RpcID:$rpcId EventID=${eventItem.id}, Time=${dur} errmsg=${e}, param=${params} ${paramStr}');

                  if (dur < 10 && e.toString().contains("CONNECTION_FAILURE")) {
                    MyLogger.getLogger().error('HSF RETRY:$dartServiceMethod RpcID:$rpcId EventID=${eventItem.id}, Time=${dur} param=${params} ${paramStr}');
                   //rpcId = TraceInfo.getEagleeyeRpcId();
                    continue_flag = true;
                  }
                  providerMonitor.success = false;
                  providerMonitor.errCode = "HSFERROR";
                })
            );
          }
          if (!continue_flag) {
            break;
          }
        }
      } else {
        r = await hsf_http_invoke( hsfInfo, rpcMethodName, dartServiceMethod, paramStr, hsfTimeOut, traceId, rpcId);
      }
    }
    providerMonitor.respSize = r?.length??0;
    providerMonitor.flush();
    MyLogger.getDebugLogger()?.debug('HSF RES:$dartServiceMethod result=${r}');
    TrafficLabour.collect(entity..returnResult = r);
    return r;
  }

}