import 'dart:async';

import 'package:saker/framework/log/log.dart';
import 'package:saker/framework/services/service_var.dart';
import 'package:saker/framework/jproxy_status/event_controller.dart';
import 'package:saker/framework/jproxy_status/plugin_manager.dart';
import 'package:saker/framework/jproxy_status/plugin_base.dart';

import '../../common/net/ip_address.dart';
import '../log/log_type.dart';
import '../stat/service_monitor.dart';
import 'proto/jproxy.pbgrpc.dart';
import 'package:grpc/grpc.dart';
import 'package:saker/framework/trace/trace_info.dart';

const Duration defaultRpcTimeOut = const Duration(milliseconds: 4000);

class ProxyResult {
  bool success;
  String data;
  String errCode;

  ProxyResult(this.success, this.data, this.errCode);

  @override
  String toString() {
    return 'ProxyResult{success: $success, data: $data, errCode: $errCode}';
  }
}

abstract class AbstractProxySubscriber {
  String get subscribePlugin;
  String get subscribeMethod;
  String get subscribeName;
  String get subscribeBody;
  int    get subscribeTimeout;

  Future<bool> onMessagePush(ProxyPushMsg value);
}

typedef bool onPushFunc(ProxyPushMsg value);

class ProxySubscriber extends AbstractProxySubscriber {
  String subscribePlugin;
  String subscribeMethod;
  String subscribeName;
  String subscribeBody;
  int    subscribeTimeout;

  onPushFunc pushFunc;

  @override
  Future<bool> onMessagePush(ProxyPushMsg value) async {
    return pushFunc(value);
  }

  ProxySubscriber(this.subscribePlugin, this.subscribeMethod, this.subscribeName,
      this.subscribeBody, this.subscribeTimeout, this.pushFunc);
}

class ProxyRpcBase implements ProxyPluginBase {
  ClientChannel _channel;
  ProxyPushClient _proxyPushClient;
  ProxyCallClient _proxyCallClient;
  StreamController<ProxyPushCtl> _streamController;

  final Map<String, StreamController<ProxyPushMsg>> _proxyStreamMap = new Map();
  Map<String, AbstractProxySubscriber> subscribers = {};

  @override
  String getName() => "jproxy";
  @override
  bool getStatus() => _ready;
  bool _ready = false;

  EventController _eventController;

  static ProxyRpcBase _rpcBase;
  static ProxyRpcBase getInstance() {
    if (_rpcBase == null) {
      _rpcBase = new ProxyRpcBase();
    }

    return _rpcBase;
  }


  void init({int port = 50051, Logger logger}) {
    PluginManager.getInstance().register(this);
    _eventController = EventControllerImpl.getInstance();

    _channel = ClientChannel('localhost',
        port: port,
        options: const ChannelOptions(credentials: ChannelCredentials.insecure()));

    _proxyCallClient = ProxyCallClient(_channel,options: new CallOptions(timeout: new Duration(seconds: 4)));

    MyLogger.getLogger().error("INFO: init rpgc connect.");
    final ctrl = StreamController<int>();

    ctrl.stream.listen((c) {
      connect(ctrl, c);
    }).onError((e) => ctrl.close());

    ctrl.add(1);
  }

  //只有_proxyPushClient OK后才能够使_ready值为true
  void connect(StreamController ctrl, int c) {
    MyLogger.getLogger().error('INFO proxy_rpc_base::Try grpc connect:$c');
    _streamController = StreamController<ProxyPushCtl>();

    _proxyPushClient = ProxyPushClient(_channel);
    _proxyPushClient.startStream(_streamController.stream).listen(
            /*void onData(T value)*/
            (v) {
              //[MAIN3] INFO proxy_rpc_base::OnMessage:id:0 data:{"ready":true} plugin:main-service subscriber:main-subscriber method:ready rpcContext: flags:0 context:
              MyLogger.getLogger().error('INFO proxy_rpc_base::OnMessage:id:${v.id} data:${v.data} plugin:${v.plugin} '
                  'subscriber:${v.subscriber} method:${v.method} rpcContext:${v.rpcContext} flags:${v.flags} context:${v.context}');
              if (!_ready) {
                _ready = true;
                _subscribeAll();
              }

              _proxyStreamMap['${v.plugin}::${v.subscriber}']?.add(v);
              MyLogger.getLogger().error('INFO  _proxyStreamMap size when init: ${_proxyStreamMap.length} ${_proxyStreamMap.keys.join("|")}');
            },
          onDone: () => MyLogger.getLogger().error('INFO proxy_rpc_base:: OnDone'),
          cancelOnError: true)
     .onError((e) {
        _ready = false;
        MyLogger.getLogger().error('proxy_rpc_base::Grpc err:$e');
        _streamController.close();
        _streamController = null;

        _eventController?.setOfflineAndChecking();

        new Future.delayed(new Duration(seconds: 1), () => ctrl.add(c+1));
    });
  }

  Future<ProxyResult> call(String plugin, String method, String data, {Duration rpcTimeout: defaultRpcTimeOut, String context, String subscriber}) async {
    final ProxyRequest request = new ProxyRequest()..plugin=plugin..method=method;
    if (data != null) {
      request.data = data;
    }
    if (context != null) {
      request.context = context;
    }
    if (subscriber != null) {
      request.subscriber = subscriber;
    }

    String traceId = TraceInfo.getTraceInfo()?.eagleeyeTraceid;
    String rpcId = TraceInfo.getTraceInfo()?.eagleeyeRpcid;
    if (traceId != null && rpcId != null) {
      request.rpcContext = new ProxyRpcContext()
        ..traceId = traceId
        ..rpcId = rpcId;
    }

    //监控
    int startTime = new DateTime.now().millisecondsSinceEpoch;
    ServiceMonitor providerMonitor = new ServiceMonitor()
      ..traceId = traceId
      ..rpcId = rpcId
      ..reqMethod = "$plugin::$method"
      ..req = ''
      ..localIP = IPAddress.getHostIp()
      ..ttid = TraceInfo.getttid()
      ..success = true
      ..remoteIP = ""
      ..mtopPath = TraceInfo.getMtopPath()
      ..rt = startTime
      ..respSize = 0
      ..logType = LogType.JPluginCall;

    final v = await _proxyCallClient.callMethod(request,options: new CallOptions(timeout: rpcTimeout));

    if(v != null && v.success){ //success
      providerMonitor.success = true;
      providerMonitor.errCode = null;
    }else{
      providerMonitor.success = false;
      providerMonitor.errCode = v?.errCode?.toString();
    }
    providerMonitor.respSize = v?.data?.length ;

    providerMonitor.flush();
    MyLogger.getDebugLogger()?.debug("$plugin::$method req:$data  res:${v?.success}---${v?.data}---${v?.errCode}");
    return new ProxyResult(v != null && v.success, v?.data, v?.errCode);
  }


  AbstractProxySubscriber createSubscriber(String plugin, String method, String subscriber, String data, onPushFunc onPush, {int subscribeTimeout = 0}) {
    return new ProxySubscriber(plugin, method, subscriber, data, subscribeTimeout, onPush);
  }

  Future<ProxyResult> subscribe(AbstractProxySubscriber subscriber) async {
    subscribers[subscriber.subscribeName] = subscriber;
    if (_ready) {
      return _subscribe(subscriber);
    }
    return new ProxyResult(true, "", "WAIT");
  }

  void _subscribeAll() async {
    for (AbstractProxySubscriber s in subscribers.values) {
      _subscribe(s);
    }
  }

  Future<ProxyResult> _subscribe(AbstractProxySubscriber subscriber) async {
    MyLogger.getLogger().error("INFO: proxy_rpc_base::_subscribe  plugin:${subscriber.subscribePlugin}, method:${subscriber.subscribeMethod}"
        "name:${subscriber.subscribeName},body:${subscriber.subscribeBody},timeout:${subscriber.subscribeTimeout}");
    var key = '${subscriber.subscribePlugin}::${subscriber.subscribeName}';
    if (!_proxyStreamMap.containsKey(key)) {
      var controller = StreamController<ProxyPushMsg>();
      controller.stream.listen((val) {
        if (subscriber.subscribeTimeout > 0) {
          subscriber.onMessagePush(val).then((pushRes) {
            var replyMsg = new ProxyPushCtl()
              ..id = val.id
              ..method = pushRes == null || !pushRes ? 'msgAckFail' : 'msgAckOK';
            _streamController.add(replyMsg);
          });
        } else {
          subscriber.onMessagePush(val);
        }
      });
      _proxyStreamMap[key] = controller;
    }

    final v = await _proxyCallClient.callMethod(new ProxyRequest()
      ..plugin = subscriber.subscribePlugin
      ..subscriber = subscriber.subscribeName
      ..method = subscriber.subscribeMethod
      ..data = subscriber.subscribeBody);

    return new ProxyResult(v != null && v.success, v?.data, v?.errCode);
  }
}