import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:isolate';

import 'package:saker/common/net/http_client.dart';
import 'package:saker/common/net/ip_address.dart';
import 'package:saker/common/utils/type_cast.dart';
import 'package:saker/framework/environment/env_base.dart';
import 'package:saker/framework/event/event_pool.dart';
import 'package:saker/framework/event/event_pool_base.dart';
import 'package:saker/framework/event/notice_type.dart';
import 'package:saker/framework/hsf/hsf_async.dart';
import 'package:saker/framework/log/dart_spdlog_extension.dart';
import 'package:saker/framework/reloader/biz_isolate_info.dart';
import 'package:saker/framework/reloader/reload_source.dart';
import 'package:saker/framework/services/service_isolate_entry.dart';
import 'package:saker/framework/services/hsf_isolate_entry.dart';
import 'package:saker/framework/log/log.dart';
import 'package:saker/framework/switch/dart_switch_extension.dart';
import 'package:saker/framework/traffic/traffic_guard.dart';

class ReloadServer {
  final EnvBase envBase;
  List<SendPort> _workerIsolateList = [];
  final Duration FUTURE_CLOSE_WAIT_TIME = new Duration(seconds: 3);
  final Duration FUTURE_CREATE_ISOLATE_WAIT_TIME = new Duration(seconds: 10);

  Map<String, BizIsolateInfo> _bizIsolateMap = {};
  Set<String> _bizDiscardSet = new Set();


  String bizPrefix;
  RegExp regExpOssList;
  RegExp regExpOssTag;

  Map<String, dynamic> hsfProviderMap = {};
  bool _hsfProviderEnable = false;
  Map<String, SendPort> getBizIsolatePorts() {
    return new Map<String, SendPort>.fromIterables(
        _bizIsolateMap.keys, _bizIsolateMap.values.map((e) => e.sendPort));
  }

  ReloadServer(this.envBase) {
    bizPrefix = 'biz';

    regExpOssList = new RegExp(r'(.*)\+0800 CST.*Standard[ ]+(.*)[ ]+'
        + '${EnvFramework.ossBucket}/${EnvFramework.appName}/(.*)');

    regExpOssTag = new RegExp(bizPrefix + r'\.(.*)\.(.*)\.(.*)');
  }

  Future<String> reload(String uTag, {String compileType: 'compile'}) async {
    try {
      int startTime = new DateTime.now().millisecondsSinceEpoch;

      String filename;
      if (compileType == 'compile') {
        filename = '${bizPrefix}.${EnvFramework.envStr}.${uTag}.snapshot';
      } else {
        filename = '${bizPrefix}.${EnvFramework.envStr}.${uTag}.dill';
      }

      List<String> args = ['-c', EnvFramework.ossConf, 'ls', '${EnvFramework.ossBucket}/${EnvFramework.appName}/${filename}'];
      ProcessResult processResult = Process.runSync('ossutil64', args);
      if (processResult == null || processResult.exitCode != 0) {
        MyLogger.getLogger().error('List oss file error:${processResult?.exitCode} stdout:${processResult.stdout} ms:${new DateTime.now().millisecondsSinceEpoch - startTime}');
        return 'Find snapshot package error, uTag=${uTag}';
      }

      String resStr = processResult.stdout as String;
      //2019-06-11 11:33:03 +0800 CST      1641568      Standard   48FDFDB4241218A57631ADF5EFFD11AE      oss://fish-dart/dart-framework/biz.prepub.135460.snapshot
      var regList = regExpOssList.allMatches(resStr);
      if (regList == null || regList.length != 1) {
        return 'Not snapshot package match, uTag=${uTag}';
      }

      var m = regList.first;

      BizIsolateInfo ossFileInfo = new BizIsolateInfo();
      ossFileInfo.lastModifiedTime = DateTime.parse(m.group(1).trim());
      ossFileInfo.eTag = m.group(2).trim();
      ossFileInfo.fileName = m.group(3).trim();
      var tagReg = regExpOssTag.firstMatch(ossFileInfo.fileName);
      if (tagReg != null &&
          tagReg.groupCount >= 3 &&
          tagReg.group(2).isNotEmpty) {
        ossFileInfo.uTag = tagReg.group(2);
      } else {
        return 'Parse snapshot package error, uTag=${uTag}';
      }

      return reloadBiz(ossFileInfo, source: compileType == 'recompile');
    } catch (e) {
      MyLogger.getLogger().error('List oss file error:$e');
    }
    return 'Reload fail, uTag=${uTag}';
  }

  List<BizIsolateInfo> listOssFile(String path) {
    try {
      int startTime = new DateTime.now().millisecondsSinceEpoch;
      List<String> args = ['-c', EnvFramework.ossConf, 'ls', '${EnvFramework.ossBucket}/${EnvFramework.appName}/${path}'];

      ProcessResult processResult = Process.runSync('ossutil64', args);
      if (processResult == null || processResult.exitCode != 0) {
        MyLogger.getLogger().error('List oss file error:${processResult?.exitCode} stdout:${processResult.stdout} ms:${new DateTime.now().millisecondsSinceEpoch - startTime}');
        return null;
      }

      List<BizIsolateInfo> list = [];
      String resStr = processResult.stdout as String;

      var regList = regExpOssList.allMatches(resStr);
      regList.forEach((m) {
        BizIsolateInfo ossFileInfo = new BizIsolateInfo();
        ossFileInfo.lastModifiedTime = DateTime.parse(m.group(1).trim());
        ossFileInfo.eTag = m.group(2).trim();
        ossFileInfo.fileName = m.group(3).trim();
        var tagReg = regExpOssTag.firstMatch(ossFileInfo.fileName);
        if (tagReg != null &&
            tagReg.groupCount >= 2 &&
            tagReg.group(2).isNotEmpty) {
          ossFileInfo.uTag = tagReg.group(2);
          list.add(ossFileInfo);
        }
      });
      return list;
    } catch (e) {
      MyLogger.getLogger().error('List oss file error:$e');
    }

    return null;
  }

  Future<String> reloadBiz(BizIsolateInfo bizInfo, {bool source: false}) async {
    BizIsolateInfo oldBizInfo = _bizIsolateMap[bizInfo.uTag];

    if (oldBizInfo != null && oldBizInfo.eTag == bizInfo.eTag) {
      return 'Isolate is running, uTag=${bizInfo.uTag}, eTag=${bizInfo.eTag}';
    }

    if (_bizDiscardSet.contains(bizInfo.eTag)) {
      return 'Discard the snapshot becasue of repeat loading, uTag=${bizInfo.uTag}, eTag=${bizInfo.eTag}';
    }

    //24小时之前的包不再加载
    if (new DateTime.now().difference(bizInfo.lastModifiedTime).inHours >= 24) {
      _bizDiscardSet.add(bizInfo.eTag);
      MyLogger.getLogger().info(
          'ReloadBiz:${bizInfo.uTag} etag=${bizInfo.eTag} time=${bizInfo.lastModifiedTime}, discard because it is too old.');
      return 'Discard the snapshot becasue of time expire, uTag=${bizInfo.uTag}, eTag=${bizInfo.eTag}';
    }

    MyLogger.getLogger().info(
        'ReloadBiz:${bizInfo.uTag} etag=${bizInfo.eTag} time=${bizInfo.lastModifiedTime}');

    String patch_snapshot = '${EnvFramework.rootPath}/patch/${bizInfo.fileName}';

      if (!downloadOssFile(bizInfo.fileName, patch_snapshot)) {
        MyLogger.getLogger().error('ReloadBiz:${bizInfo.uTag} download ${bizInfo.fileName} error');
        return 'Download snapshot failed, uTag=${bizInfo.uTag}, eTag=${bizInfo.eTag}';
      }

    if (source && oldBizInfo != null && oldBizInfo.isolateId != null) {
      bool a = await reloadSource(oldBizInfo.isolateId, 'file://$patch_snapshot');

      bizInfo.isolateId = oldBizInfo.isolateId;
      MyLogger.getLogger().info("ReloadBiz:${bizInfo.uTag} file=${bizInfo.fileName} ok");

      return 'Reload ok, ReloadSource=${a}, uTag=${bizInfo.uTag}, eTag=${bizInfo.eTag}';
    }

    //如果之前没有自己都isolate，则新建一个isolate。
    EventItem eventItem = EventPool.getEvent();
    var messages = [
      EnvFramework.rootPath, eventItem.sendPort, eventItem.id, bizInfo.uTag];
    var isolate = await Isolate.spawnUri(Uri.parse('file://$patch_snapshot'), null, messages,debugName: bizInfo.uTag)
        .catchError((e) {
              _bizDiscardSet.add(bizInfo.eTag);
              MyLogger.getLogger().error('ReloadBiz:${bizInfo.uTag} spawn error:$e');
        });
    if (isolate == null) {
      return 'Isolate create error, uTag=${bizInfo.uTag}, eTag=${bizInfo.eTag}';
    }

    bizInfo.isolateId = await findIsolateIdByName(bizInfo.uTag);
    bizInfo.sendPort = safeCast(await eventItem.future
        .timeout(FUTURE_CLOSE_WAIT_TIME)
        .catchError((e) {
          _bizDiscardSet.add(bizInfo.eTag);
          MyLogger.getLogger().error('ReloadBiz${bizInfo.uTag} wait error:$e');
        })
    );

    MyLogger.getLogger().warn('ReloadBiz -- ${bizInfo.isolateId} -- ${isolate.controlPort}');

    //通知关闭旧的isolate。
    if (oldBizInfo != null && oldBizInfo.sendPort != null) {
      eventItem = EventPool.getEvent();
      oldBizInfo.sendPort.send([
        EventType_Notice,
        [NoticeType.BizIsolateClose, eventItem.sendPort, eventItem.id, 0]
      ]);
      await eventItem.future.timeout(FUTURE_CLOSE_WAIT_TIME)
          .catchError((e) =>
          MyLogger.getLogger().error('ReloadBiz:${bizInfo.uTag} error while close old:$e'));
    }

    //增量
    if (bizInfo.increment != null && bizInfo.increment.lastModifiedTime.compareTo(bizInfo.lastModifiedTime) > 0 ) {
      String patch_dill = '${EnvFramework.rootPath}/patch/${bizInfo.increment.fileName}';

      if (downloadOssFile(bizInfo.increment.fileName, patch_dill)) {
        bool a = await reloadSource(bizInfo.isolateId, 'file://$patch_dill');
        MyLogger.getLogger().info("ReloadBiz:${bizInfo.uTag} increment file=${bizInfo.fileName} message=${a} ok");
      }
    }

    if (bizInfo.sendPort != null) {
      _bizIsolateMap[bizInfo.uTag] = bizInfo;
      MyLogger.getLogger().info("ReloadBiz:${bizInfo.uTag} ok");

      await broadcastHttpIsolate(
          NoticeType.BizIsolateMapUpdate, getBizIsolatePorts());
      return 'Reload ok, uTag=${bizInfo.uTag}, eTag=${bizInfo.eTag}';
    } else {
      _bizIsolateMap.remove(bizInfo.uTag);
      MyLogger.getLogger().info("ReloadBiz:${bizInfo.uTag} fail");
      return 'Isolate response error, uTag=${bizInfo.uTag}, eTag=${bizInfo.eTag}';
    }
  }

  bool downloadOssFile(String name, String path) {
    try {
      int startTime = new DateTime.now().millisecondsSinceEpoch;
      ProcessResult downRes = Process.runSync('ossutil64', ['-c', EnvFramework.ossConf, 'cp', '-f', '${EnvFramework.ossBucket}/${EnvFramework.appName}/${name}', path]);
      if (downRes == null || downRes.exitCode != 0) {
        MyLogger.getLogger().error(
            "Download oss error:${downRes?.exitCode} stdout:${downRes
                .stdout} ms:${new DateTime.now().millisecondsSinceEpoch -
                startTime}");
        return false;
      }
    } catch (e) {
      MyLogger.getLogger().error('Download oss: error catch: $e');
      return false;
    }

    return true;
  }

  int readLocalCurrentVersion() {
    try {
      File localFile = new File("${EnvFramework.rootPath}/patch/current.version");
      if (localFile.existsSync()) {
        String a = localFile.readAsStringSync();
        return int.tryParse(a)??0;
      }
    } catch (e) {
      MyLogger.getLogger().error('ReadLocalCurrentVersion: error catch: $e');
    }
    return 0;
  }

  void writeLocalCurrentVersion(int version) {
    current_version = version;
  }

  int readRemoteCurrentVersion() {
    String path = "${EnvFramework.rootPath}/patch/${EnvFramework.envStr}.version.1";
    if (!downloadOssFile("patch.${EnvFramework.envStr}.version", path)) {
      MyLogger.getLogger().error("Reload:download patch.${EnvFramework.envStr}.version");
      return 0;
    }
    try {
      File localFile = new File(path);
      if (localFile.existsSync()) {
        String a = localFile.readAsStringSync();
        return int.tryParse(a)??0;
      }
    } catch (e) {
      MyLogger.getLogger().error('ReadRemoteCurrentVersion Catch:$e');
    }
    return 0;
  }

  int current_version = 0;

  @Deprecated('to be replaced by current_isolate_server_index')
  int current_http_index = 0;

  int current_isolate_server_index = 0;

  Future checkBizUpdate() async {
    List<BizIsolateInfo> bizList = listOssFile('${bizPrefix}.${EnvFramework.envStr}.');
    if (bizList != null) {
      Map<String, BizIsolateInfo> bizMap = new Map();

      for (BizIsolateInfo bizInfo in bizList) {
        if (bizInfo.fileName.endsWith('.snapshot')) {
          bizMap[bizInfo.uTag] = bizInfo;
        }
      }

      for (BizIsolateInfo bizInfo in bizList) {
        if (bizInfo.fileName.endsWith('.dill') && bizMap.containsKey(bizInfo.uTag)) {
          bizMap[bizInfo.uTag].increment = bizInfo;
        }
      }

      for (BizIsolateInfo bizInfo in bizMap.values) {
        await reloadBiz(bizInfo);
      }
    }
  }

  checkUpdate() {
    checkBizUpdate();
  }

  //main isolate start。
  start() async {
    envBase.initApp();

    MyLogger.getLogger().error("hsf core init: ${EnvFramework.hsfIOThreadCnt}, ${EnvFramework.hsfWorkerThreadCnt}");
    hsf_core_init(EnvFramework.hsfSnapshotPath, EnvFramework.hsfLogPath,
        EnvFramework.hsfLogLevel, EnvFramework.appName, EnvFramework.hsfIOThreadCnt, EnvFramework.hsfWorkerThreadCnt);

    dart_switch_ext_init(EnvFramework.appName);

    for (int i = current_http_index; i <
        current_http_index + EnvFramework.httpConcurrence; i ++) {
      print('${new DateTime.now().toString()} Server: create isolate:$i');
      EventItem eventItem = EventPool.getEvent();
      var messages = [
        EnvFramework.rootPath,
        eventItem.sendPort,
        eventItem.id,
        i,
        getBizIsolatePorts(),
        EnvBase.getEnv()
      ];

      await Isolate.spawn(worker_isolate_entry, messages).catchError((e) =>
          MyLogger.getLogger().error('Main isolate spawn error:$e'));

      SendPort s = safeCast(await eventItem.future.timeout(
          FUTURE_CREATE_ISOLATE_WAIT_TIME)
          .catchError((e) =>
          MyLogger.getLogger().error('Main isolate wait error:$e')));
      if (s != null) {
        _workerIsolateList.add(s);
      }
    }
    current_http_index += EnvFramework.httpConcurrence;

    _hsfProviderEnable = hsfProviderMap.isNotEmpty;
    if (_hsfProviderEnable) {
      EventItem eventItem = EventPool.getEvent();
      //new isolate create hsf provider.
      var hsfMessages = [
        EnvFramework.rootPath,
        eventItem.sendPort,
        eventItem.id,
        EnvBase.getEnv(),
        json.encode(hsfProviderMap)
      ];
      await Isolate.spawn(hsf_server_isolate_entry, hsfMessages).catchError((e) =>
          MyLogger.getLogger().error('HSF Provider isolate spawn error:hsf_server_isolate_entry $e'));

      await eventItem.future.timeout(FUTURE_CREATE_ISOLATE_WAIT_TIME)
          .catchError((e) =>
          MyLogger.getLogger().error('HSF Provider isolate wait error:$e'));
    }

    //启动HTTP
    if (EnvFramework.httpDelay != null) {
      new Future.delayed(EnvFramework.httpDelay, () {
        checkAllIsolateReady();
      });
    } else {
      checkAllIsolateReady();
    }

    MyLogger.getLogger().warn("CheckUpdate:hotReloadInterval=${EnvFramework.hotReloadInterval}");
    if (EnvFramework.hotReloadInterval > 0) {
      current_version = readLocalCurrentVersion();
      await checkUpdate();
    }

    reportSelf();

//    trafficGuardEntry(1, children: _workerIsolateList);
  }

  void reportSelf() {
    Map<String, String> report = {
      'appName': EnvFramework.appName,
      'fwVer': EnvFramework.version,
      'port': EnvFramework.httpPort?.toString(),
      'env': EnvFramework.envStr,
      'ip': IPAddress.getHostIp(),
      'hostname': Platform.localHostname
    };

    String reportData = json.encode(report);

    httpGet('http://pre-dart-framework.alibaba-inc.com/StatService?data=$reportData')
        .then((v) {
          if (v != null && v.isNotEmpty) {
            MyLogger.getLogger().error('Report framework state: $v');
          }
        }).catchError((e) => MyLogger.getLogger().error('StatService $e'));

    new Future.delayed(new Duration(days: 1), () => reportSelf());
  }

  void hsfOnline() {
    MyLogger.getLogger().info("trye to set hsfOnline......");
    try {
      ProcessResult downRes = Process.runSync("curl", ["http://localhost:12201/hsf/online?k=HSF"]);
      if (downRes == null || downRes.exitCode != 0) {
        MyLogger.getLogger().error("hsfOnline failed. ${downRes?.exitCode} stdout:${downRes.stdout}");
        return;
      }
    } catch (e) {
      MyLogger.getLogger().error("hsfOnline exception: $e");
    }
  }


  void hsfOffline() {
    MyLogger.getLogger().info("try to set hsfOffline......");
    try {
      ProcessResult downRes = Process.runSync("curl", ["http://localhost:12201/hsf/offline?k=HSF"]);
      if (downRes == null || downRes.exitCode != 0) {
        MyLogger.getLogger().error("hsfOffline failed. ${downRes?.exitCode} stdout:${downRes.stdout}");
        return;
      }
    } catch (e) {
      MyLogger.getLogger().error("hsfOffline exception: $e");
    }
  }

  close() async {
    await broadcastHttpIsolate(NoticeType.HttpShutdown, 0);

    _workerIsolateList = [];

    switch_destroy();
    hsf_core_stop();
    hsf_core_destroy();
//    TairBase.close();

    envBase.closeApp();
    spdlog_close();
//    trafficGuardEntry(2);
  }


  //广播给所有的HTTP Isolate
  broadcastHttpIsolate(int noticeType, var msg) async {
    try {
      for (int i = 0; i < _workerIsolateList.length; i++) {
        EventItem eventItem = EventPool.getEvent();
        _workerIsolateList[i].send(
            [
              EventType_Notice,
              [noticeType, eventItem.sendPort, eventItem.id, msg]
            ]);
        await eventItem.future.timeout(FUTURE_CLOSE_WAIT_TIME).catchError((e) =>
            MyLogger.getLogger().error('Broadcast http error:${e}'));
      };
    } catch (e, stack) {
      MyLogger.getLogger().error('Broadcast http error:${e}, ${stack}');
    }
  }

  //广播给所有的Biz Isolate
  broadcastBizIsolate(int noticeType, var msg) async {
    for (BizIsolateInfo bizInfo in _bizIsolateMap.values) {
      EventItem eventItem = EventPool.getEvent();
      bizInfo.sendPort.send([EventType_Notice, [noticeType, eventItem.sendPort, eventItem.id, msg]]);
      await eventItem.future.timeout(FUTURE_CLOSE_WAIT_TIME).catchError((e) =>
          MyLogger.getLogger().error('Broadcast biz error:${e}'));;
    };
  }

  Timer _checkIsolateTimer;
  void checkAllIsolateReady() {
    if (_checkIsolateTimer != null) {
      return;
    }
    _checkIsolateTimer = new Timer.periodic(new Duration(seconds: 2), (timer) {
      //如果各个isloate ready，则上线http服务和hsf服务
      isWorkerIsolatesReady().then((r) {
        if (r) {
          _checkIsolateTimer = null;
          timer.cancel();
          broadcastHttpIsolate(NoticeType.HttpOnline, 1);

          if (_hsfProviderEnable) {
            new Future.delayed(EnvFramework.hsfProviderDelay, () {
              hsfOnline();
            });
          }
        }
      });
    });
  }

  //判断所有Isolate是否初始化好
  Future<bool> isWorkerIsolatesReady() async {
    try {
      for (int i = 0; i < _workerIsolateList.length; i++) {
        EventItem eventItem = EventPool.getEvent();
        _workerIsolateList[i].send(
            [
              EventType_Notice,
              [NoticeType.WorkerIsolateReady, eventItem.sendPort, eventItem.id, 1]
            ]);
        dynamic r = await eventItem.future.timeout(FUTURE_CLOSE_WAIT_TIME).catchError((e) =>
            MyLogger.getLogger().error('Broadcast http error:${e}'));
        if (r == null || r == false) {
          return false;
        }
      };
    } catch (e, stack) {
      MyLogger.getLogger().error('getHttpIsolateStatus error:${e}, ${stack}');
      return false;
    }
    return true;
  }

}