import 'dart:async';
import 'dart:convert';

import 'package:saker/common/net/ip_address.dart';
import 'package:saker/framework/hsf/hsf_async.dart';
import 'package:saker/framework/jsonserializer/utils.dart';
import 'package:saker/framework/log/log_type.dart';
import 'package:saker/framework/services/service_var.dart';
import 'package:saker/framework/stat/service_monitor.dart';
import 'package:saker/framework/trace/trace_info.dart';
import 'package:saker/framework/log/log.dart';
import 'package:saker/framework/services/service_base.dart';
import 'package:saker/framework/services/service_result.dart';

String encodeResult(Map data, String type, bool isOK, {String errorCode, String errorMsg}) {
  if(type == 'mtop'){
    return json.encode({
      'success' : isOK,
      'errorCode' : errorCode,
      'errorMsg' : errorMsg,
      'payload' : data
    });
  }
  //对协议兼容，去掉无用的属性
  //FIXME: 消息limit_block返回处理
  else if(type == 'metaq'){
    data.remove('api');
    data.remove('v');
    Map _data = data['data'] as Map;
    _data.remove('serverTime');
    if(errorCode=='FAIL_SYS_FLOWLIMIT'){
      return json.encode( {
        'consumeStatus': 'CONSUME_SUCCESS',
        'delayLevelWhenNextCons': 1
      });
    }
    return json.encode(_data);
  }
  else{
    return json.encode(data);
  }
}

String getBizError(Map result_data, String errCode, String errMsg, TraceInfo traceInfo, String type) {
  result_data['ret'] = ['${errCode}::${errMsg}'];
  result_data['data'] = {};
  if (traceInfo.debugLogBuffer != null) {
    result_data['log'] = traceInfo.debugLogBuffer.toString();
  }

  //return json.encode(result_data);
  return encodeResult(result_data, type, false, errorCode: errCode, errorMsg: errMsg);
}

Future<String> processDispatch(String requestName,
    Map<dynamic, dynamic> reqMap,
    Map<String, String> parameters,
    Map<String, String> headers) async {
  MyLogger.getDebugLogger()?.debug(
      'ProcessDispatch:Path=${requestName}, headers:${headers}');

  MtopContext mtopContext = new MtopContext(parameters, headers);

  TraceInfo traceInfo = new TraceInfo();
  traceInfo.eagleeyeIndex = 0;
  traceInfo.eagleeyeTraceid = headers['eagleeye-traceid'] ?? "";
  traceInfo.eagleeyeRpcid = '${headers['eagleeye-rpcid'] ?? ""}.1';
  traceInfo.ttid = parameters.containsKey("ttid") ? parameters["ttid"] : "";
  traceInfo.mtopPath = requestName;
  if (reqMap.containsKey('x-isolate-redirect')) {
    traceInfo.debugLogBuffer = new StringBuffer();
  }
  if(headers.containsKey("eagleeye-userdata")){
    traceInfo.eagleeyeUserData = headers["eagleeye-userdata"];
    Map<String, String> userData = Uri.splitQueryString(headers["eagleeye-userdata"]);
    traceInfo.pressTestTag = userData["t"]??"";
    traceInfo.inReloadIsolate = reqMap.containsKey('x-isolate-redirect'); //userData.containsKey("dpath_env");
    traceInfo.eagleeyeContext = userData;
  }

  Map<String, String> eagle_ctx = {};
  bool eagleEyeFlag = false;
  eagle_ctx["svc_sig"] =  "${parameters["api"] ?? ""}:${parameters["v"] ?? ""}";
  eagle_ctx["traceid"] = traceInfo.eagleeyeTraceid;
  eagle_ctx["rpcid"] = traceInfo.eagleeyeRpcid;
  eagle_ctx["method_name"] = requestName;
  eagle_ctx["x-remote-ip"] = parameters["caller-remote-ip"] ?? "";
  eagle_ctx["userdata"] = headers["eagleeye-userdata"] ?? "";
  if(traceInfo.eagleeyeTraceid != null && traceInfo.eagleeyeTraceid != "") {
    http_eagleeye_start(json.encode(eagle_ctx).toString());
    eagleEyeFlag = true;
  }

  traceInfo.serviceMonitor = new ServiceMonitor()
    ..traceId = traceInfo.eagleeyeTraceid
    ..rpcId = traceInfo.eagleeyeRpcid
    ..ttid = parameters.containsKey("ttid") ? parameters["ttid"] : ""
    ..req = jsonEncode(reqMap) //+ " userid: "+ (parameters.containsKey('userId') ? parameters['userId'] : "")
    ..reqMethod = parameters["api"]
    ..mtopPath = requestName
    ..localIP = IPAddress.getHostIp()
    ..remoteIP = parameters["caller-remote-ip"]
    ..rt = new DateTime.now().millisecondsSinceEpoch
    ..userId = parameters['userId']
    ..userNick = parameters['nick']
    ..logType = LogType.MtopSvc;

  Map result_data = {
    'api': parameters['api'],
    'v': parameters['v']
  };

  if (!serviceEntries.containsKey(requestName)) {
    traceInfo.serviceMonitor.success = false;
    traceInfo.serviceMonitor.respSize = 0;
    traceInfo.serviceMonitor.errCode = "FAIL_SYS_SERVICE_MISS";
    traceInfo.serviceMonitor.flush();
    if(eagleEyeFlag)
      http_eagleeye_end(traceInfo.eagleeyeTraceid ?? "", 1, 0);
    return getBizError(result_data, 'FAIL_SYS_SERVICE_MISS', '服务不存在', traceInfo, parameters['type']);
  }

  ServiceBase serviceBase = serviceEntries[requestName];
  if(serviceBase.flowLimitBlock()) {
    traceInfo.serviceMonitor.success = false;
    traceInfo.serviceMonitor.respSize = 0;
    traceInfo.serviceMonitor.errCode = "FAIL_SERVICE_FLOW_LIMIT";
    traceInfo.serviceMonitor.flush();
    if(eagleEyeFlag)
      http_eagleeye_end(traceInfo.eagleeyeTraceid ?? "", 1, 0);
    //单独记录一条限流日志,便于统计和告警
    traceInfo.serviceMonitor.reqMethod = "SERVICE_FLOW_LIMIT";
    MyLogger.getLogger(name:"flowLimit")?.error("${traceInfo.serviceMonitor.toString()}");
    MyLogger.getLogger().error("flow limit actived, just return: ${traceInfo.eagleeyeTraceid} ${traceInfo.eagleeyeRpcid}.");
    return getBizError(result_data, 'FAIL_SYS_FLOWLIMIT', '请稍候再试', traceInfo, parameters['type']);
  }

  bool successFlag = true;
  String errCode = "";
  String a = await runZoned(() async {
    ServiceResult data = null;
    try {
      data = await serviceBase.processMtop(reqMap, mtopContext)
          .catchError((e, stackTrace) =>
          MyLogger.getLogger().error(
              'Process dispatch error biz code:$e\n${stackTrace}'));
    } catch (e,stackTrace) {
      MyLogger.getLogger().error(
          'Process dispatch error biz error catch:$e \n${stackTrace}');
    }

    if (data == null) {
      errCode = "FAIL_SERVICE_ERROR";
      successFlag = false;
      return getBizError(result_data, 'FAIL_SERVICE_ERROR', '服务错误', traceInfo, parameters['type']);
    }

    result_data['ret'] = ['${data.errcode}::${data.errmsg}'];
    result_data['data'] = data.data ?? {};
    if (traceInfo.debugLogBuffer != null) {
      result_data['log'] = traceInfo.debugLogBuffer.toString();
    }

    String result = encodeResult(result_data, parameters['type'], true);
    if(data.errcode != "SUCCESS"){
      successFlag = false;
      errCode = data.errcode;
    }
    return result;
  }, zoneValues: {#MtopContext: mtopContext, #TraceInfo: traceInfo},
      onError: (e, stackTrace) =>
          MyLogger.getLogger().error(
              'Process dispatch zone error:$e\n${stackTrace}')
  );

  //对外服务eagleeye打点日志，Monitor监控日志，限流统计计数减1
  if(eagleEyeFlag)
    http_eagleeye_end(traceInfo.eagleeyeTraceid ?? "", successFlag?0:1, a.length);
  traceInfo.serviceMonitor.success = successFlag;
  traceInfo.serviceMonitor.errCode = errCode;
  if(traceInfo.serviceMonitor.respSize == null){//如果业务侧自己定义了result长度，框架不再修改
    traceInfo.serviceMonitor.respSize = a?.length??0;
  }
  if(a.length < 10000) { /** 结果太大则忽略*/
    traceInfo.serviceMonitor.res = a;
  }
  traceInfo.serviceMonitor.flush();
  return a;
}