import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:saker/common/utils/type_cast.dart';
import 'package:saker/common/utils/http_util.dart';
import 'package:saker/framework/config/frame_work_config_manager.dart';
import 'package:saker/framework/environment/env_base.dart';
import 'package:saker/framework/event/event_pool.dart';
import 'package:saker/framework/event/event_pool_base.dart';
import 'package:saker/framework/event/notice_type.dart';
import 'package:saker/framework/log/log.dart';
import 'package:saker/framework/services/service_dispatch.dart';
import 'package:saker/framework/services/service_var.dart';
import 'package:mime/mime.dart';
import 'package:saker/framework/traffic/traffic_labour.dart';
import 'package:saker/framework/traffic/traffic_protocol.dart';

_writeBizError(HttpRequest request, Map result_data, String errCode, String errMsg) {
  result_data['ret'] = ['${errCode}::${errMsg}'];
  result_data['data'] = {};

  String a = json.encode(result_data);
  request.response
    ..write(a)
    ..close().catchError((e) => MyLogger.getLogger().error('_writeBizError:$e'));
}

_writeServerError(HttpRequest request, int errCode) {
  request.response
    ..statusCode = errCode
    ..write('Server error "${request.uri.path}"')
    ..close().catchError((e) => MyLogger.getLogger().error('_writeServerError:$e'));
}

//下线处理
_set_offline(HttpRequest request, int offline) {
  if (request?.connectionInfo?.remoteAddress?.address == '127.0.0.1' &&
      mainEventPort != null) {
    int nt = offline == 1 ? NoticeType.HttpOffline : NoticeType.HttpOnline;
    //通知主Isolate
    mainEventPort.send(
        [EventType_Notice, [nt, EventPool.sendPort, 0, offline]]);

    request.response
      ..write('OK')
      ..close().catchError((e) =>
          MyLogger.getLogger().error('ProcessHttp echo response:$e'));
    return;
  }
  request.response
    ..write('FAIL')
    ..close().catchError((e) =>
        MyLogger.getLogger().error('ProcessHttp echo response:$e'));
  return;
}

_health(HttpRequest request) {
  if (svcOfflineStatus == 0) {
    request.response
      ..write('ECHO')
      ..close().catchError((e) =>
          MyLogger.getLogger().error('Health response:$e'));
  } else {
    request.response
      ..statusCode = HttpStatus.notFound
      ..write('OFFLINE')
      ..close().catchError((e) =>
          MyLogger.getLogger().error('Health response:$e'));

    MyLogger.getLogger().error('HTTP Health Offline');
  }
}

_hotReload(HttpRequest request, Map<String, String> parameters) async {
  if (parameters.containsKey('redirect-hosts') && parameters.containsKey('tag')) {
    //转发到另一台机器
    try {
      String _path = request.uri.path;
      String _host = parameters['redirect-hosts'];

      Uri u = Uri.parse("http://${_host}:${EnvFramework.httpPort}${_path}?tag=${parameters['tag']}&type=${parameters['type']}");
      MyLogger.getLogger().info('HotReload redirect to ${u}');

      HttpClientRequest req = await httpClient.getUrl(u).timeout(new Duration(seconds: 5))
        ..headers.contentType = request.headers.contentType;

      HttpClientResponse response = await req.close()..timeout(new Duration(seconds: 11));
      String rtn = await response.transform(utf8.decoder).join();

      request.response
        ..statusCode = response.statusCode
        ..write(rtn)
        ..close().catchError((e) =>
            MyLogger.getLogger().error('HotReload redirect response error:$e'));

      return;
    } catch (e) {
      MyLogger.getLogger().error('HotReload redirect error:$e');
    }
  } else if (parameters.containsKey('tag')) {
    MyLogger.getLogger().info('HotReload request, tag=${parameters["tag"]}, type=${parameters["type"]}');
    EventItem eventItem = EventPool.getEvent();
    //通知主Isolate
    mainEventPort.send(
        [
          EventType_Notice,
          [
            NoticeType.BizHotReload,
            eventItem.sendPort,
            eventItem.id,
            parameters['tag'],
            parameters['type']??'compile'
          ]
        ]);

    var res = await eventItem.future.timeout(new Duration(seconds: 10))
        .catchError((e) => MyLogger.getLogger().error('HotReload wait error:$e'));
    if (res != null) {
      request.response
        ..write(res)
        ..close().catchError((e) =>
            MyLogger.getLogger().error('HotReload response error:$e'));
      return;
    }
  }

  request.response
    ..write('Hot reload fail!!!\nHot reload fail!!!\nHot reload fail!!!')
    ..close().catchError((e) =>
        MyLogger.getLogger().error('HotReload response error:$e'));
  return;
}

_fileReload(HttpRequest request) async {
  MyLogger.getLogger().warn('HTTP FileReload request');
  if (request.method == 'POST') {
    try {
      List<String> contentType = request.headers[HttpHeaders.contentTypeHeader];
      if (contentType.isNotEmpty) {
        var header = HeaderValue.parse(contentType[0]);
        print(header);
        if (header.parameters.containsKey('boundary')) {
          var content = await new MimeMultipartTransformer(header.parameters['boundary']).bind(request).toList()
//          var content = await request.transform(
//              new MimeMultipartTransformer(header.parameters['boundary']))
//              .toList()
              .catchError((e) =>
              MyLogger.getLogger().error(
                  'HTTP FileReload Error transform: ${e}'));
          for (var part in content) {
            if (part.headers.containsKey('content-disposition')) {
              header = HeaderValue.parse(part.headers['content-disposition']);
              String filename = header.parameters['filename'];

              final File file = new File('/tmp/' + filename);
              IOSink fileSink = file.openWrite();
              await part.pipe(fileSink).catchError((e) =>
                  MyLogger.getLogger().error(
                      'HTTP FileReload Error pipe: ${e}'));

              fileSink.close();
            }
          }
        }
      }
    } catch (e) {
      MyLogger.getLogger().error('HTTP FileReload Error: ${e}');
    }
  }
  request.response..write('OK')..close();
}

/**
 * 长连
 * */
final HttpClient httpClient = new HttpClient()..maxConnectionsPerHost=5;
final String REDIRECT_TAG = 'rd';

Future<bool> _redirect(String host, int port, HttpRequest request, Map parameters) async{
  String _path = request.uri.path;
  try {
    HttpClientRequest req = await httpClient.post(host, port, _path) /*1*/
      ..headers.contentType = request.headers.contentType; /*2*/

    request.headers.forEach((name, List<String> values) {
      if (name != 'host' && name != 'content-length') {
        req.headers.add(name, values);
      }
    });
    req.headers.add(REDIRECT_TAG, '1');

    req.write(makeQuery(parameters?.cast()));
    HttpClientResponse response = await req.close();
    String rtn = await response.transform(utf8.decoder).join();

    request.response
      ..statusCode = response.statusCode
      ..write(rtn)
      ..close().catchError((e) =>
          MyLogger.getLogger().error('ProcessHttp response:$e'));
    return true;
  }catch(e){
    MyLogger.getLogger().error('ProcessHttp exception :$e');
    return false;
  }
}

final Duration BIZ_REQ_TIMEOUT = new Duration(seconds: 6);

Future<String> _dispatch(String requestName,
    Map<dynamic, dynamic> reqMap,
    Map<String, String> parameters,
    Map<String, String> headers) async {
  String tag = null;
  if(headers.containsKey('eagleeye-userdata')){
    Map<String, String> userData = Uri.splitQueryString(headers['eagleeye-userdata']);
    tag = userData.containsKey('dpath_env') ? userData['dpath_env'] : null;
    if (tag != null && tag == 'DPathBaseEnv') {
      tag = null;
    }
  }
  if (tag == null || tag.isEmpty) {
    tag = reqMap['x-isolate-redirect'] as String;
  } else {
    reqMap['x-isolate-redirect'] = tag;
  }

  if (tag != null && tag.isNotEmpty) {
    if (bizIsolateMap.containsKey(tag)) {
      EventItem eventItem = EventPool.getEvent();
      bizIsolateMap[tag].send([EventType_Notice, [
        NoticeType.BizHttpReq,
        eventItem.sendPort,
        eventItem.id,
        [requestName, reqMap, parameters, headers]
      ]
      ]);
      dynamic res = await eventItem.future.timeout(BIZ_REQ_TIMEOUT)
          .catchError((e) => MyLogger.getLogger().error('Dispatch error:event error:$e'));
      return safeCast<String>(res);
    } else {
      MyLogger.getDebugLogger()?.debug(
          'Dispatch error:req=${requestName},tag=${tag}, tag not exist');
    }
  }
  return processDispatch(requestName, reqMap, parameters, headers);
}

Future processHttp(HttpRequest request) async {
  String requestName = request.uri.path;
  if (requestName != null && requestName == '/filereload') {
    await _fileReload(request);
    return;
  }

  request.response.headers.contentType = jsonContentType;

  Map<String, dynamic> __parameters  = {};
  Map<String, dynamic> _parameters = {};
  if (request.method == 'GET') {
    if (request?.requestedUri?.queryParameters != null) {
      __parameters = new Map.from(request?.requestedUri?.queryParameters);
    }
  } else if (request.method == 'POST') {
    String content = await utf8.decoder.bind(request).join()
        .catchError((e) =>
        MyLogger.getLogger().error('ProcessHttp POST receive:$e'));

    if(request.headers.contentType.mimeType  == 'application/json'){
      __parameters = json.decode(content) as Map<String, dynamic>;
    }else {
      __parameters = Uri.splitQueryString(content ?? '');
    }
    _parameters = new Map.from(__parameters);
  } else {
    MyLogger.getLogger().error(
        'ProcessHttp unsupport method: ${request.method}');

    _writeServerError(request, HttpStatus.methodNotAllowed);
    return;
  }

  Map<String, String> _headers = {};
  Map<String, String> parameters  = {};

  final String xEventType = (request.headers['x-event-type']!=null) ? request.headers['x-event-type'][0] : null;

  if(__parameters['type'] == 'mtop'){
    if(__parameters['payload'] == null){
      _writeServerError(request, HttpStatus.badRequest);
      return;
    }

    if(__parameters['payload'] is Map) {
      Map<String, dynamic> payload = __parameters['payload'] as Map<String, dynamic>;
      if(payload['headers'] != null && payload['headers'] is Map){
        _headers = Map.castFrom<String, dynamic, String, String>(payload['headers'] as Map<String, dynamic>);
      }
      payload.remove('headers');
      payload.remove('dataMap');
      parameters = Map.castFrom<String, dynamic, String, String>(payload);
    }else{
      parameters = Map.castFrom<String, dynamic, String, String>(json.decode(__parameters['payload'] as String) as Map<String, dynamic>);
    }
    parameters['type'] = 'mtop';
  }
  else if(xEventType == 'metaq'){
//    parameters = Map.castFrom<String, dynamic, String, String>(__parameters);
    parameters['type'] = xEventType;
  }
  else{
    parameters = new Map.from(__parameters);
  }

  Map result_data = {
    'api': parameters['api'],
    'v': parameters['v']
  };

  if (requestName == null || requestName == '') {
    _writeBizError(request, result_data, 'FAIL_SERVICE_MISS', '服务不存在');
    return;
  }

  if(preProcessing.containsKey(requestName)){
    preProcessing[requestName](request);
    return;
  }

  if (requestName == '/offline') {
    _set_offline(request, 1);
    return;
  }

  if (requestName == '/online') {
    _set_offline(request, 0);
    return;
  }

  if (requestName == '/hotreload') {
    _hotReload(request, parameters);
    return;
  }

  if (requestName == '/health' || requestName == '/' || requestName == '/nginx_status' || requestName == "/status.taobao") {
    _health(request);
    return;
  }

  //指定用户路由
  String userNick = parameters['nick'];
  if(userNick != null && request.headers[REDIRECT_TAG] == null
      && FrameWorkSwitchManager.betaHostIp != null
      && FrameWorkSwitchManager.betaUserNicks.contains(userNick)){
    if(await _redirect(FrameWorkSwitchManager.betaHostIp, EnvFramework.httpPort, request, parameters)) {
      return;
    }
  }

  Map<String, String> headers = {};
  request.headers.forEach((name, List<String> values) {
    if (values != null && values.length > 0) {
      headers[name] = values[0];
    }
  });
  _headers.forEach((String k, String v){
    headers[k.toLowerCase()] = v;
  });

  //在预发测试时转发
  if(headers.containsKey('redirect-hosts') && request.headers[REDIRECT_TAG] == null) {
    MyLogger.getDebugLogger()?.debug('redirect:${headers["redirect-hosts"]}');
    await _redirect(headers['redirect-hosts'], EnvFramework.httpPort, request, parameters);
    return;
  }

  if (headers.containsKey('x-mg-nativeproxyh5')) {
    parameters['nativeProxyH5'] = headers['x-mg-nativeproxyh5'];
  }

  Map<dynamic, dynamic> reqMap;

  String json_data = parameters['data'];

  if (json_data == null || json_data.isEmpty) {
    json_data = parameters['m_original_data']??'{}';
  }
  try {
    if(parameters['type'] == 'metaq'){
      reqMap = Map.castFrom<String, dynamic, String, dynamic>(__parameters);
    }else{
      reqMap = json.decode(json_data) as Map<dynamic, dynamic>;
    }
  } catch (e) {
    reqMap = {};
  }

  if (reqMap.containsKey('x-s2s-ext')) {
    Map<String, String> s2sExt = (json.decode(reqMap.remove('x-s2s-ext') as String) as Map)?.cast();
    if (s2sExt != null) {
      if (s2sExt.containsKey('sys_entrance')) {
        parameters['sysEntrance'] = s2sExt['sys_entrance'];
      }
      if (s2sExt.containsKey('nativeProxyH5')) {
        parameters['nativeProxyH5'] = s2sExt['nativeProxyH5'];
      }
      if (s2sExt.containsKey('platform')) {
        parameters['platform'] = s2sExt['platform'];
      }
      if (s2sExt.containsKey('appName')) {
        parameters['appName'] = s2sExt['appName'];
      }
      if (s2sExt.containsKey('deviceId')) {
        parameters['deviceId'] = s2sExt['deviceId'];
      }
      if (s2sExt.containsKey('lat')) {
        parameters['lat'] = s2sExt['lat'];
      }
      if (s2sExt.containsKey('lng')) {
        parameters['lng'] = s2sExt['lng'];
      }
      if (s2sExt.containsKey('myhost')) {
        headers['s2sclientip'] = s2sExt['myhost'];
      }
    }
  }

  headers['direct-ip'] = request.connectionInfo?.remoteAddress?.address;

  final CollectEntity traffic = new CollectEntity()
    ..request    = request
    ..params     = _parameters
    ..traceId    = rewriteTraceId(request,headers)
    ..invokeCat  = REQUEST_CAT.HTTP
    ..mainInvoke = true;

//  MyLogger.getDebugLogger()?.debug(
//      'ProcessHttp:Path=${requestName},Method=${request.method}\nParam=${parameters}\nHeaders= ${headers}');

  //传递remote ip address
  parameters['caller-remote-ip'] = request.connectionInfo.remoteAddress.address;

  TrafficLabour.inspect(traffic);

  String rtn = await _dispatch(requestName, reqMap, parameters, headers)
      .catchError((e, stackTrace) =>
      MyLogger.getLogger().error(
          'ProcessHttp error biz code:${e}\n${stackTrace}'));
  if (null == rtn) {
    _writeServerError(request, HttpStatus.internalServerError);
    TrafficLabour.collect(traffic..returnResult=rtn);
    return;
  }
  TrafficLabour.collect(traffic..returnResult=rtn);
  request.response
    ..write(rtn)
    ..close().catchError((e) =>
        MyLogger.getLogger().error('ProcessHttp response:$e'));

}
