// Copyright (c) 2019, the saker project authors. Please see the AUTHORS file
// for details. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.


import 'dart:async';
import 'dart:convert';
import 'dart:isolate';

import 'package:saker/framework/environment/env_base.dart';
import 'package:saker/framework/event/event_pool.dart';
import 'package:saker/framework/event/event_pool_base.dart';
import 'package:saker/framework/event/notice_type.dart';
import 'package:saker/framework/hsf/hsf_provider_base.dart';
import 'package:saker/framework/log/log.dart';
import 'package:saker/framework/services/service_http.dart';
import 'package:saker/framework/services/service_var.dart';
import 'package:saker/framework/jproxy_status/plugin_manager.dart';

import 'grpc/grpc_service.dart';

const String GRPC = 'grpc';
const String HTTP = 'http';

const String lc_iso_init_before="lc_iso_init_before";
const String lc_iso_init_start ="lc_iso_init_start";
const String lc_iso_init_ok ="lc_iso_init_ok";
const String lc_svr_init_start ="lc_svr_init_start";
const String lc_svr_init_started ="lc_svr_init_started";
const String lc_svr_init_failed ="lc_svr_init_failed";
const String lc_svr_close_failed ="lc_svr_close_failed";
const String lc_svr_close_success ="lc_svr_close_failed";

//A  muti-isolate service impl
class ServerService extends AbstractServer{

  AbstractServer _serviceServer;

  static ServerService _self;

  ServerService._(this._serviceServer);

  factory ServerService._forInstance(AbstractServer specKind) {
    if(_self == null) {
      _self = ServerService._(specKind);
    }
    return _self;
  }

  factory ServerService.http() {
   return ServerService._forInstance(HttpServerService());
  }

  factory ServerService.grpc() {
    return ServerService._forInstance(GrpcServerService());
  }

  @override
  String get serverId => _serviceServer._serverId;

  @override
  String get serverKind => _serviceServer.serverKind;

  @override
  Future init(List<dynamic> param){
    return _serviceServer.init(param);
  }

  @override
  Future start({List<dynamic>  param}) {
    //from var
    mainEventPort = param[1] as SendPort;
    int eventId = param[2] as int;
    int _serverIndex = param[3] as int;
    _serviceServer._serverId = serverKind+" ${_serverIndex}-${Isolate.current.hashCode}";
    try {
      _printServerLog(statusCode:lc_iso_init_before);
      if (param[4] != null && param[4] is Map) {
        Map m = param[4] as Map;
        bizIsolateMap = m.cast<String, SendPort>();
      }
      EventPool.notice_cb = server_notice_callback;
      _printServerLog(statusCode: lc_iso_init_start);
      EnvBase.getEnv().initIsolate();
      serviceEntries = EnvBase.getEnv().getServiceEntries();

      mainEventPort.send([
        EventType_Notice,
        [
          NoticeType.HsfProviderMap,
          _serverIndex,
          json.encode(HsfProviderBase.servicesRegister)
        ]
      ]);
      _printServerLog(statusCode: lc_iso_init_ok);
      _printServerLog(statusCode: lc_svr_init_start);
      //real start
      _startServer(param:[EventType_Any, eventId, EventStatus_OK, EventPool.sendPort]);
      _printServerLog(statusCode: lc_svr_init_started);

    } catch (e) {
      _printServerLog(statusCode: lc_svr_init_failed);
    }
  }

  _startServer({List<dynamic>  param}){
    MyLogger.getLogger().warn('${_serviceServer.serverId} starting @ ${EnvFramework.httpAddr}:${EnvFramework.httpPort}');
      _serviceServer.start(param: param).then((v){
        mainEventPort.send(param);
        MyLogger.getLogger().warn('${_serviceServer.serverId} start ok @ ${EnvFramework.httpAddr}:${EnvFramework.httpPort}');
      }).catchError((e) {
        MyLogger.getLogger().error('${_serviceServer.serverId} start error = $e');
      mainEventPort.send(param);
    });
  }

  @override
  Future close({String type, SendPort sendPort, int eventId}){
    if(_serviceServer!=null){
      _serviceServer.close();
    }
  }

  @override
  void shutdown({SendPort sendPort, int eventId}){
    MyLogger.getLogger().warn("Server ${this.serverId} is shutting down");
    if(_serviceServer!=null){
      _serviceServer.close(sendPort: sendPort, eventId: eventId).whenComplete((){
      _serviceServer=null;
      MyLogger.getLogger().warn("${this.serverId} has been closed");
      sendPort.send([EventType_Any, eventId, EventStatus_OK, 0]);
      EnvBase.getEnv().releaseIsolate();
      Isolate.current.kill();
      });
    }
  }

  @override
  void offline({SendPort sendPort, int eventId, int offline}){
    svcOfflineStatus = offline;
    MyLogger.getLogger().warn("svcOfflineStatus value change ${this.serverId} is ${svcOfflineStatus==0?'online':'offline'}");
    _serviceServer.offline();
    sendPort.send([EventType_Any, eventId, EventStatus_OK, svcOfflineStatus]);
  }

  void _printServerLog({String lifecycle, String statusCode}) {
    print('${new DateTime.now().toString()} ${_serviceServer._serverId} ${statusCode}');
  }
}

abstract class AbstractServer{

  String _serverId;

  String get serverKind;

  String get serverId => _serverId;

  Future init(List<dynamic> param);

  Future start({List<dynamic> param});

  Future close({String type, SendPort sendPort, int eventId});

  void shutdown({SendPort sendPort, int eventId});

  void offline({SendPort sendPort, int eventId, int offline});

}

void server_notice_callback(dynamic messages) {
  try {
    if (messages is! List) {
      return;
    }
    var nt = messages[0] as int;
    SendPort sendPort = messages[1] as SendPort;
    int eventId = messages[2] as int;

    if (nt == NoticeType.WorkerIsolateReady) {
      _http_ready(sendPort, eventId, EnvFramework.isPluginActived? PluginManager.getInstance().getStatus():true);
    }else if (nt == NoticeType.HttpOffline) {
      processor.offline(sendPort: sendPort,eventId: eventId,offline: 1);
    } else if (nt == NoticeType.HttpOnline) {
      processor.offline(sendPort: sendPort,eventId: eventId,offline: 0);
    } else if (nt == NoticeType.HttpShutdown) {
      processor.shutdown(sendPort: sendPort, eventId: eventId);
    } else if (nt == NoticeType.BizIsolateMapUpdate) {
      _update_biz_map(sendPort, eventId, (messages[3] as Map)?.cast<String, SendPort>());
    } else {
      MyLogger.getLogger().warn("${processor.serverId} notice discard: ${messages}");
    }
  } catch (e,stack) {
    MyLogger.getLogger().error("${processor.serverId} error: ${e}, ${stack}");
  }
}


void _update_biz_map(SendPort s, int eventId, Map<String, SendPort> m) {
  if (m != null) {
    bizIsolateMap = m;
  }
  MyLogger.getLogger().warn("${processor.serverId} BizMap update ${m}");
  s.send([EventType_Any, eventId, EventStatus_OK, EventPool.sendPort]);
}

void _http_ready(SendPort s, int eventId, bool val) {
  s.send([EventType_Any, eventId, EventStatus_OK, val]);
}