import 'dart:async';

import 'package:saker/common/net/ip_address.dart';
import 'package:saker/framework/environment/env_base.dart';
import 'package:saker/framework/event/event_pool_base.dart';
import 'package:saker/framework/event/tair_event_pool.dart';
import 'package:saker/framework/log/log.dart';
import 'package:saker/framework/log/log_type.dart';
import 'package:saker/framework/stat/service_monitor.dart';
import 'package:saker/framework/tair/dart_tair_extension_interface.dart';
import 'package:saker/framework/trace/trace_info.dart';

Map<String, int> tairMap = {};

int tair_close_all(){
  MyLogger.getDebugLogger()?.debug("tair close............");
  for (String x in tairMap.keys) {
    tair_close(x);
  }
  return 0;
}

class TairOpResult{
  int resCode;
  String result;

  String toString(){
    return "resCode:$resCode; result:$result";
  }
}

class TairService{
  static const Duration defaultTairTimeOut =  const Duration(seconds: 3);
  String userName ;
  int tcAddress;//must must must not modified by hand
  String userApp;
  TairService(String puserName, int apiTimeout,int localCacheSize,int localCacheTimeout){
    userName = puserName;
    tcAddress = tair_init(EnvFramework.appName, userName, apiTimeout, localCacheSize, localCacheTimeout, EnvFramework.tairLogPath, EnvFramework.tairLogLevel,io_thread_cnt: EnvFramework.tairIOThreadCnt);
    if(tcAddress <= 0){
      MyLogger.getLogger().error("init tair client failed: $puserName");
    }
    tairMap[puserName] = tcAddress;
  }

  //兼容FaaS无应用名的场景
  TairService.forUserApp(String puserName,String userApp,int apiTimeout,int localCacheSize,int localCacheTimeout){
    userName = puserName;
    userApp  = userApp;
    MyLogger.getLogger().error("CHANGED INIT TAIR: puserName = $puserName , username= $userName");
    tcAddress = tair_init(userApp, userName, apiTimeout, localCacheSize, localCacheTimeout, EnvFramework.tairLogPath, EnvFramework.tairLogLevel,io_thread_cnt: EnvFramework.tairIOThreadCnt);
    if(tcAddress <= 0){
      MyLogger.getLogger().error("init tair client failed: $puserName");
    }
    tairMap[puserName] = tcAddress;
  }

  int put(String key,String value,{int tairCacheTimeout :3600*1000,int version : 0} ){
    MyLogger.getDebugLogger()?.debug("TairService Sync Put: $key ");
    return tair_put(tcAddress, key, value, tairCacheTimeout, version);
  }
  String get(String key){
    MyLogger.getDebugLogger()?.debug("TairService Sync Get: $key ");
    return tair_get(tcAddress, key);
  }

  Future<Map<String,TairOpResult>> mput_async(Map<String,String> map,{int tairCacheTimeout: 3600*1000,int version : 0, Duration apiTimeout:defaultTairTimeOut})async {
    List<Future> futureList = [];
    Map<String,TairOpResult> results = {};
    map.forEach((String k, String v){
      futureList.add(_put_async(k,v,results,CacheTimeout: tairCacheTimeout, ver: version, apiTimeout: defaultTairTimeOut));
    });

    await Future.wait(futureList);
    return results;
  }

  Future _put_async(String key,String value,Map<String,TairOpResult> results,{int CacheTimeout: 3600*1000,int ver : 0, Duration apiTimeout:defaultTairTimeOut})async {
    TairOpResult tairOpResult = await put_async(key, value, tairCacheTimeout: CacheTimeout, version: ver, apiTimeout: defaultTairTimeOut);
    results[key] = tairOpResult;
    return;
  }

  Future<TairOpResult> put_async(String key,String value,{int tairCacheTimeout: 3600*1000,int version : 0, Duration apiTimeout:defaultTairTimeOut})async {
    MyLogger.getDebugLogger()?.debug("TairService async PUT REQ:$userName $key");
    String rpcId = TraceInfo.getEagleeyeRpcId();
    String traceId = TraceInfo.getEagleeyeTraceId();
    ServiceMonitor providerMonitor = new ServiceMonitor()
      ..traceId = traceId
      ..rpcId = rpcId
      ..reqMethod = "tair_put_async"
      ..req = userName
      ..localIP = IPAddress.getHostIp()
      ..ttid = TraceInfo.getttid()
      ..success = true
      ..remoteIP = ""
      ..mtopPath = TraceInfo.getMtopPath()
      ..rt = new DateTime.now().millisecondsSinceEpoch
      ..respSize = 0
      ..logType = LogType.TairPut;

    TairOpResult result = new TairOpResult()
      ..resCode = 0;

    EventItem eventItem = TairEventPool.getEvent();
    int code = tair_put_async(tcAddress, key, value, tairCacheTimeout, version, eventItem.sendPort, eventItem.id, traceId, rpcId, userName);
    if (code != 0) {
      TairEventPool.remove(eventItem.id);
      MyLogger.getLogger().error('TairService async PUT ERROR:put failed. RpcID:$rpcId key:$key errcode=${code}');
      result.resCode = -9999;
      return result;
    }

    dynamic d = await eventItem.future
        .timeout(apiTimeout, onTimeout: () {
      TairEventPool.remove(eventItem.id);
      MyLogger.getLogger().error('TairService async PUT ERROR:tair put failed. RpcID:$rpcId key:$key TIMEOUT');

      providerMonitor.success = false;
      providerMonitor.errCode = "TIMEOUT";
      result.resCode = -3989;
    }).catchError((e) {
      MyLogger.getLogger().error('TairService async PUT ERROR:tair put failed. RpcID:$rpcId key:$key errmsg=${e}');
      providerMonitor.success = false;
      providerMonitor.errCode = "$e";
      try {
        result.resCode = e.toString().contains(":")? int.tryParse(e.toString().split(":")[0])??-9999 : -9999;
      }catch(e){
        result.resCode = -9999;
      }
    });
    String r = d as String;
    providerMonitor.respSize = r?.length??0;
    providerMonitor.flush();

    result.result = rpcId;

    MyLogger.getDebugLogger()?.debug('TairService async PUT RES: RpcID:$rpcId key:$key result=${r}');
    return result;
  }

  Future<Map<String,TairOpResult>> mremove_async(List<String> keys,{ Duration apiTimeout:defaultTairTimeOut})async {
    List<Future> futureList = [];
    Map<String,TairOpResult> results = {};
    for(String key in keys){
      futureList.add(_remove_async(key,results,apiTimeout: defaultTairTimeOut));
    }
    await Future.wait(futureList);
    return results;
  }

  Future _remove_async(String key,Map<String,TairOpResult> results,{ Duration apiTimeout:defaultTairTimeOut})async {
    TairOpResult tairOpResult = await remove_async(key, apiTimeout: defaultTairTimeOut);
    results[key] = tairOpResult;
    return;
  }

  Future<TairOpResult> remove_async(String key, { Duration apiTimeout:defaultTairTimeOut})async {
    MyLogger.getDebugLogger()?.debug("TairService async REMOVE REQ:${EnvFramework.appName} $userName $key");
    String rpcId = TraceInfo.getEagleeyeRpcId();
    String traceId = TraceInfo.getEagleeyeTraceId();
    ServiceMonitor providerMonitor = new ServiceMonitor()
      ..traceId = traceId
      ..rpcId = rpcId
      ..reqMethod = "tair_remve_async"
      ..req = userName
      ..localIP = IPAddress.getHostIp()
      ..ttid = TraceInfo.getttid()
      ..success = true
      ..remoteIP = ""
      ..mtopPath = TraceInfo.getMtopPath()
      ..rt = new DateTime.now().millisecondsSinceEpoch
      ..respSize = 0
      ..logType = LogType.TairDel;

    TairOpResult result = new TairOpResult()
      ..resCode = 0;

    EventItem eventItem = TairEventPool.getEvent();
    int code = tair_rm_async(tcAddress, key, eventItem.sendPort, eventItem.id, traceId, rpcId,userName);
    if (code != 0) {
      TairEventPool.remove(eventItem.id);
      MyLogger.getLogger().error('TairService async REMOVE ERROR:remove failed. RpcID:$rpcId key:$key errcode=${code}');
      result.resCode = -9999;
      return result;
    }

    dynamic d = await eventItem.future
        .timeout(apiTimeout, onTimeout: () {
      TairEventPool.remove(eventItem.id);
      MyLogger.getLogger().error('TairService async REMOVE ERROR:tair remove failed. RpcID:$rpcId key:$key TIMEOUT');
      providerMonitor.success = false;
      providerMonitor.errCode = "TIMEOUT";
      result.resCode = -3989;
    }).catchError((e) {
      MyLogger.getLogger().error('TairService async REMOVE ERROR:tair remove failed. RpcID:$rpcId key:$key errmsg=${e}');
      providerMonitor.success = false;
      providerMonitor.errCode = "$e";
      try {
        result.resCode = e.toString().contains(":")? int.tryParse(e.toString().split(":")[0])??-9999 : -9999;
      }catch(e){
        result.resCode = -9999;
      }
    });
    String r = d as String;
    providerMonitor.respSize = r?.length??0;
    providerMonitor.flush();

    result.result = r;

    MyLogger.getDebugLogger()?.debug('TairService async REMOVE RES: RpcID:$rpcId key:$key result=${r}');
    return result;
  }

  Future<Map<String,TairOpResult>> mget_async(List<String> keys,{ Duration apiTimeout:defaultTairTimeOut})async {
    List<Future> futureList = [];
    Map<String,TairOpResult> results = {};
    for(String key in keys){
      futureList.add(_get_async(key,results,apiTimeout: defaultTairTimeOut));
    }
    await Future.wait(futureList);
    return results;
  }

  Future _get_async(String key,Map<String,TairOpResult> results,{ Duration apiTimeout:defaultTairTimeOut})async {
    TairOpResult tairOpResult = await get_async(key, apiTimeout: defaultTairTimeOut);
    results[key] = tairOpResult;
    return;
  }

  Future<TairOpResult> get_async(String key,{ Duration apiTimeout:defaultTairTimeOut, bool valueJavaFormat:false, TairValueType valueType:TairValueType.tair_value_string})async {
    MyLogger.getDebugLogger()?.debug("TairService async GET REQ:${EnvFramework.appName} $userName $key, valueJavaformat:$valueJavaFormat, valueFormat:$valueType");
    String rpcId = TraceInfo.getEagleeyeRpcId();
    String traceId = TraceInfo.getEagleeyeTraceId();
    ServiceMonitor providerMonitor = new ServiceMonitor()
      ..traceId = traceId
      ..rpcId = rpcId
      ..reqMethod = "tair_get_async"
      ..req = userName
      ..localIP = IPAddress.getHostIp()
      ..ttid = TraceInfo.getttid()
      ..success = true
      ..remoteIP = ""
      ..mtopPath = TraceInfo.getMtopPath()
      ..rt = new DateTime.now().millisecondsSinceEpoch
      ..respSize = 0
      ..logType = LogType.TairGet;

    TairOpResult result = new TairOpResult()
      ..resCode = 0;

    int valueFormat = 0;
    if(valueJavaFormat){
      valueFormat = valueFormat | tair_value_java_format;
    }
    if(valueType == TairValueType.tair_value_string){
      valueFormat = valueFormat | tair_value_string;
    }else if(valueType == TairValueType.tair_value_int64){
      valueFormat = valueFormat | tair_value_int64;
    }else if(valueType == TairValueType.tair_value_int32){
      valueFormat = valueFormat | tair_value_int32;
    }

    EventItem eventItem = TairEventPool.getEvent();
    int code = tair_get_async(tcAddress, key, eventItem.sendPort, eventItem.id, traceId, rpcId, userName, valueFormat);
    if (code != 0) {
      TairEventPool.remove(eventItem.id);
      MyLogger.getLogger().error('TairService async GET ERROR:get failed. RpcID:$rpcId key:$key errcode=${code}');
      result.resCode = -9999;
      return result;
    }

    dynamic d = await eventItem.future
        .timeout(apiTimeout, onTimeout: () {
      TairEventPool.remove(eventItem.id);
      MyLogger.getLogger().error('TairService async GET ERROR:tair get failed. RpcID:$rpcId key:$key TIMEOUT ');
      providerMonitor.success = false;
      providerMonitor.errCode = "TIMEOUT";
      result.resCode = -3989;
    }).catchError((e) {
      if(e.toString().contains("-3998")){
        MyLogger.getLogger().info('TairService async GET EMPTY. RpcID:$rpcId key:$key errmsg=${e}');
        providerMonitor.success = true;
        providerMonitor.errCode = "empty result.";
      }else {
        MyLogger.getLogger().error('TairService async GET ERROR:tair get failed. RpcID:$rpcId key:$key errmsg=${e}');
        providerMonitor.success = false;
        providerMonitor.errCode = "$e";
        try {
          result.resCode = e.toString().contains(":")? int.tryParse(e.toString().split(":")[0])??-9999 : -9999;
        }catch(e){
          result.resCode = -9999;
        }
      }
    });
    String r = d as String;
    providerMonitor.respSize = r?.length??0;
    providerMonitor.flush();

    result.result = r;

    MyLogger.getDebugLogger()?.debug('TairService async GET RES: RpcID:$rpcId key:$key result=${r}');
    return result;
  }

  void close(){
    tair_close(userName);
  }
}