import 'dart:async';
import 'dart:io';
import 'dart:isolate';
import 'package:http/io_client.dart';
import 'package:saker/framework/event/event_pool.dart';
import 'package:saker/framework/event/event_pool_base.dart';
import 'package:saker/framework/traffic/traffic_event.dart';
import 'package:saker/framework/log/log.dart';
import 'package:http/http.dart';
import 'dart:convert';
import 'package:saker/common/utils/type_cast.dart';


const GUARD_SERVER_PORT               = 8083;
const GUARD_TARGET_PORT               = 8084;
const GUARD_HB_TIMEOUT                = const Duration(seconds: 5);
const GUARD_HB_INTERVAL               = const Duration(seconds: 5);
const GUARD_BROADCAST_TIMEOUT         = const Duration(seconds: 10);
const GUARD_HB_MAX_FAILED_CNT_PER_MIN = 100;

_MainProbe _host ;

Logger log   = MyLogger.getLogger(name:"trafficGuard");

List<SendPort> _isolates;

trafficGuardEntry(int signal, {List<SendPort> children}) async {
  _isolates ??= children;
  _host     ??= new _MainProbe.Http();
  switch (signal) {
    case 1:
      _host.start().catchError((e) =>
          log?.error("TrafficGuard - Probe start error, see $e"));
      break;
    case 2:
      _host.close(shutdown: true).catchError((e) =>
          log?.error("TrafficGuard - Probe shutdown error, see $e"));
      break;
    case 3:
      _host.close().catchError((e) =>
          log?.error("TrafficGuard - Probe stop error, see $e"));
      break;
  }
}

_broadcastEvent(var event) async {
  try {
    for (int i = 0; i < _isolates.length; i++) {
      EventItem eventItem = EventPool.getEvent();
      _isolates[i].send(
          [
            EventType_Traffic_Guard,
            [event,eventItem.sendPort,eventItem.id,_isolates.length],
          ]);
      await eventItem.future.timeout(const Duration(seconds: 3)).catchError((e) =>
          log?.error('TrafficGuard - Broadcast error:${e}'));
    };
  } catch (e, stack) {
    log?.error("TrafficGuard - Broadcast to child isolate error:${e}, ${stack}");
  }
}

/**
 * Simple protocol irrelevant probe
 **/
abstract class _Probe{

  String   _hbTargetHost, _hbPath;
  Duration _hbInterval  , _hbTimeout;
  int      _hbFailedCntPerMin;
  int      _probeServerPort,_hbTargetPort;
  HandleEvt                              _handle;
  Timer                                  _refresherTimer;
  Timer                                  _heartbeatTimer;
  StreamController<TrafficGuardEvt>      _eventController;
  Stream                                 _probeServer;

  static const int successCount = 3;

  start() async{
    _initController();
    startProbeServer().catchError(
            (e) => log?.error("TrafficGuard - Probe server start failed, can be ignored if already started $e"));
    _heartbeat();
    await _eventController.stream.listen((event){
      handleEvent(event);
    },onDone: (){
      close();
    });
  }

  close({bool shutdown : false}) async{
    _close(shutdown:shutdown);
  }

  void handleEvent(TrafficGuardEvt event){
    _handle(event);
  }

  void _initController(){
    _eventController = new StreamController(
        onListen: (){
          log?.info("TrafficGuard - Probe start heartbeating");
        }
    );
  }

  _heartbeat() async {
    var initCheckCount = 0, failedCount    = 0;
    _refresherTimer = new Timer.periodic(const Duration(minutes: 10), (t){
      log?.info("TrafficGuard - Probe heartbeat refresh counter");
      failedCount = 0;
    });

    new Timer(new Duration(seconds: GUARD_HB_INTERVAL.inSeconds * (successCount+1)),(){
      initCheckCount = -1;
    });
    _heartbeatTimer = new Timer.periodic(this._hbInterval, (_) {
      if (!_eventController.isClosed) {
        ping().catchError((_){
          if (++failedCount >= _hbFailedCntPerMin) {
            _eventController.add(new TrafficGuardEvt(TrafficGuardEvt.GUARD_EVT_RUNTIME_FAILED));
          }
        }).then((_) {
          if (initCheckCount >= 0 && ++initCheckCount == successCount) {
            _eventController.add(new TrafficGuardEvt(TrafficGuardEvt.GUARD_EVT_HEALTH));
          }
        }).timeout(_hbTimeout, onTimeout: () {
          if (++failedCount >= _hbFailedCntPerMin) {
            _eventController.add(new TrafficGuardEvt(TrafficGuardEvt.GUARD_EVT_RUNTIME_FAILED));
          }
        });
      }
    });
  }

  _close({bool shutdown}) async{
    try{
      _heartbeatTimer?.cancel();
      _refresherTimer?.cancel();
      _eventController.close();
    }catch(e){
      log?.error("TrafficGuard - close probe server error see $e");
    }
  }

  Future ping();

  Future startProbeServer();

}

/**
 * A simple http probe
 */
class _HttpProbeDelegate extends _Probe{

  Client     _client = new IOClient();
  static  var upStreamReg = new RegExp('^\/traffic\/(updateTaskInfo|doReplay)\?\$');
  static  var cmdReg      = new RegExp('^\/traffic\/(stop|start|upd|shutdown)\?\$');
  var    _url;

  @override
  ping() async{
    await  _client.post(_url ??= 'http://$_hbTargetHost:$_hbTargetPort$_hbPath');
  }

  @override
  startProbeServer() async{
    this._probeServer = await HttpServer.bind(
      InternetAddress.anyIPv4,
      _probeServerPort,
      shared: true
    );
    log?.info("TrafficGuard - Probe server started at http://localhost:$_probeServerPort");
    _probeServer.transform(
        new StreamTransformer<HttpRequest, TrafficGuardEvt>.fromHandlers(handleData: _sinkHttpData))
        .listen((evt) => handleEvent(evt), onError: (e)  =>  log?.error("TrafficGuard - Probe httpListener error see $e"),
      onDone: () {
        (_probeServer as HttpServer).close();
      },
    );
  }

  void _sinkHttpData(HttpRequest httpRequest, EventSink sink) async {
    String path  = httpRequest.uri.path;
    TrafficGuardEvt event;
    try {
      if (path.startsWith('/traffic')) {
        String action = path.substring(path.lastIndexOf('/') + 1);
        if (path.startsWith(upStreamReg)) {
          event = new TrafficGuardEvt(TrafficGuardEvt.GUARD_EVT_UPSTREAM);
          if (action == 'doReplay') {
            event.msg = [
              action,
              httpRequest.requestedUri.queryParameters,
            ];
          } else {
            await utf8.decoder.bind(httpRequest).join().then((content) {
              event.msg = [action,content];
              log?.info("TrafficGuard - Probe Server event ${event.msg}");
            }).catchError((e) {
              log?.error("TrafficGuard - Probe Server parse http request failed ...");
            }
            );
          }
        }
        if (path.startsWith(cmdReg)) {
          switch (action) {
            case "start" :
              event = new TrafficGuardEvt(TrafficGuardEvt.GUARD_EVT_CMD_START);
              break;
            case "stop" :
              event = new TrafficGuardEvt(TrafficGuardEvt.GUARD_EVT_CMD_STOP);
              break;
            case "shutdown" :
              event = new TrafficGuardEvt(TrafficGuardEvt.GUARD_EVT_CMD_SHUTDOWN);
              break;
          }
          if (event != null) {
            event.msg = [action, ""];
          }
        }
      }
      if (event != null) {
        httpRequest.response.statusCode = HttpStatus.ok;
        httpRequest.response.writeln("{\"success\" :true}");
        sink.add(event);
      }else{
        httpRequest.response.statusCode = HttpStatus.forbidden;
      }
    }catch(e){
    }
    finally{
      httpRequest.response.close().catchError((e) => log?.error("TrafficGuard - $e"));
    }
  }

  @override
  close({bool shutdown: false}) async{
    if (shutdown) {
       await safeCast(_probeServer).close(force: true).then(
              (_) => _eventController.add(new TrafficGuardEvt(TrafficGuardEvt.GUARD_EVT_CMD_STOP)));
    }
    super.close(shutdown: shutdown);
  }
}


class _MainProbe {

  static final _MainProbe _mainProbe = new _MainProbe._internal();
  bool   _started = false;
  _Probe _delegate;

  _registerHandler() async {
    _delegate._handle = (TrafficGuardEvt evt) {
        log?.info("TrafficGuard - Probe server @ ${Isolate.current.hashCode} received incomming event ");
        print(evt);
        if(evt.eventKey == TrafficGuardEvt.GUARD_EVT_CMD_START || evt.eventKey == TrafficGuardEvt.GUARD_EVT_HEALTH){
          _broadcastEvent(new TrafficGuardEvt(TrafficGuardEvt.LABOUR_EVT_START))
              .timeout(GUARD_BROADCAST_TIMEOUT)
              .catchError((e) => log?.error("TrafficGuard - Broadcast to child isolate error:${e}"))
              .then((_)=>start());
        }
        else if(evt.eventKey == TrafficGuardEvt.GUARD_EVT_UPSTREAM){
          _broadcastEvent(new TrafficGuardEvt(TrafficGuardEvt.LABOUR_EVT_UPSTREAM, message: evt.msg))
              .timeout(GUARD_BROADCAST_TIMEOUT)
              .catchError((e) => log?.error("TrafficGuard - Broadcast to child isolate error:${e}"));
        }
        else if(evt.eventKey  == TrafficGuardEvt.GUARD_EVT_CMD_STOP
              ||evt.eventKey  == TrafficGuardEvt.GUARD_EVT_RUNTIME_FAILED
              ||evt.eventKey  == TrafficGuardEvt.GUARD_EVT_CMD_SHUTDOWN){
          _broadcastEvent(new TrafficGuardEvt(TrafficGuardEvt.LABOUR_EVT_CLOSE))
              .timeout(GUARD_BROADCAST_TIMEOUT)
              .catchError((e) => log?.error("TrafficGuard - Broadcast to child isolate error:${e}"))
              .then((_)=>close(shutdown: evt.eventKey == TrafficGuardEvt.GUARD_EVT_CMD_SHUTDOWN));
        }
    };
  }

  start() async {
    if (!_started) {
      _registerHandler();
      _started = true;
      _delegate.start();
    }
  }

  close({bool shutdown: false}) async {
    if (_delegate != null && _started) {
      _delegate.close(shutdown: shutdown).timeout(const Duration(seconds: 5))
          .then((v) {_started = false;
          log?.warn("TrafficGuard - Probe server closed successfully");
      });
    }
  }

  factory _MainProbe.Http({
        int serverPort        : GUARD_SERVER_PORT,
        int probeServerPort   : GUARD_SERVER_PORT,
        Duration hbTimeout    : GUARD_HB_TIMEOUT,
        int hbFailedCntPerMin : GUARD_HB_MAX_FAILED_CNT_PER_MIN,
        Duration hbInterval   : GUARD_HB_INTERVAL,
        String hbTargetHost   : "localhost",
        int hbTargetPort      : GUARD_TARGET_PORT})
        => _mainProbe
        .. _delegate = (new _HttpProbeDelegate()
          .._hbPath            = "/hb"
          .._hbTargetPort      = hbTargetPort
          .._hbTargetHost      = hbTargetHost
          .._probeServerPort   = probeServerPort
          .._hbInterval        = hbInterval
          .._hbFailedCntPerMin = hbFailedCntPerMin
          .._hbTimeout         = hbTimeout
        );

  _MainProbe._internal();
}