import 'dart:async';
import 'dart:collection';
import 'dart:convert';
import 'dart:io';
import 'dart:isolate';

import 'package:http/http.dart';
import 'package:http/io_client.dart';
import 'package:saker/framework/environment/env_base.dart';
import 'package:saker/framework/event/event_pool_base.dart';
import 'package:saker/framework/trace/trace_info.dart';
import 'package:saker/framework/traffic/traffic_event.dart';
import 'package:saker/framework/traffic/traffic_guard.dart';
import 'package:saker/framework/traffic/traffic_protocol.dart';
import 'package:saker/common/utils/type_cast.dart';

const GW_TIMEOUT = const Duration(seconds: 5);
const DEFAULT_TASK_ID = "1";

SendPort _mainSendPort;

void handleTrafficEvent(var msg){
  try{
    if(msg!=null && msg[1] is SendPort){
      _mainSendPort ??= (safeCast(msg[1]));
    }
    final TrafficGuardEvt evt = safeCast(msg[0]);
    final String action = (evt.msg!=null)? safeCast(evt.msg[0]):"";
    switch(evt.eventKey) {
      case TrafficGuardEvt.LABOUR_EVT_UPSTREAM:
        if (action == "updateTaskInfo") {
          TrafficLabour._refreshTrafficTask(safeCast(evt.msg[1]));
        }
        break;
      case TrafficGuardEvt.LABOUR_EVT_START:
        TrafficLabour._start();
        break;
      case TrafficGuardEvt.LABOUR_EVT_CLOSE:
        TrafficLabour._close();
        break;
    }
  }catch(e){
    log?.error("TrafficLabour - event error see $e @${Isolate.current.hashCode}");
  }finally{
    _sendReply(msg).catchError((e) =>
        log?.error("TrafficLabour - send response to main isolate error see $e"));
  }
}

_sendReply(msg) async{
  if(_mainSendPort!=null){
    _mainSendPort.send([EventType_Any, msg[2], EventStatus_OK, 0]);
  }
}

/**
 *  TrafficLabour provides static access to collect application's traffic
 *  and replay functionality.
 *
 */
class TrafficLabour{

  static final TrafficLabour _traffic =  new TrafficLabour._internal();
  bool _isStarted = false;
  _TrafficInvokeManager _invokeManager;
  _Collector            _collector;

  static FutureOr inspect(CollectEntity entity){
    if(! _traffic._isStarted || entity == null){
      return null;
    }
    return _traffic._inspect(entity);
  }

  static collect(CollectEntity entity) async {
    if(_traffic._isStarted){
      _traffic._collect(entity);
    }
  }

  static void  _refreshTrafficTask(String content) {
    if(_traffic._isStarted){
      _traffic._refreshTrafficTaskInternally(content);
    }
  }

  static _close() async => _traffic._closeInternally();

  static _start() async => _traffic._startInternally();

  _closeInternally() async{
    _collector.cancel().then((_)=>_isStarted=false);
  }

  void _refreshTrafficTaskInternally(String content){
    _invokeManager.refreshTaskRule(
        DEFAULT_TASK_ID, new TrafficTaskInfo.fromJson(jsonDecode(content)));
  }

  void _startInternally() {
    if(!_isStarted){
      _collector.init();
      _isStarted = true;
    }
  }

  FutureOr _collect(CollectEntity pak) async => _collector.collect(pak);

  FutureOr _inspect(CollectEntity req) => _invokeManager.inspectInvoke(req);

  factory TrafficLabour() => _traffic;

  TrafficLabour._internal(){
    _invokeManager = new _TrafficInvokeManager();
    _collector     = new _Collector().._invokeManager=_invokeManager;
  }
}


/**
 * Collect call protocol msg in async way including msg filtering, bulk-assembling &
 * sending msg to gateway.
 */
class _Collector{

  final   Queue  _buffer       = new DoubleLinkedQueue();
  static  int  SIZE_OF_BUFFER  = 100;
  static  int  BULK_SIZE       = 1;
  StreamController<CollectEntity>  _collectStream;
  Timer                            _autoFlushTimer;
  _TrafficInvokeManager            _invokeManager;

  void init() {
    _collectStream ??= new StreamController(
        onListen: () {
          log?.warn("TrafficLabour - collector stream is listening...");
          _autoFlushTimer =
          new Timer.periodic(const Duration(minutes: 5), (t) {
            _statInfo();
            _flush();
          });
        },
        onCancel: () {
          _autoFlushTimer.cancel();
          _flush();
          log?.warn("TrafficLabour - collector stream closed...");
        }
    )
      ..stream.transform(
          new StreamTransformer.fromHandlers(handleData: _assemble))
          .listen((bulk) {
        if (!_collectStream.isClosed) {
          //TODO: bulked group sending
          bulk.forEach((req) {
            _Gateway.sendRequest(_Gateway.ACTION_COLLECT_INVOKE, req.toString()
                , queryString: {"invokeType": req.requestCatString()});
          });
        }
      }, onError: (e,stack) {});
  }

  collect(CollectEntity each) async {
    try {
      if (!_collectStream.isClosed && each!=null) {
        _collectStream.add(each);
      }
    } catch (e) {}
  }

  cancel() async{
    if(!_collectStream.isClosed){
      _collectStream.close().catchError((e,stack)=>log?.error("TrafficLabour - collector close error , see $e $stack"));
    }
  }

  void _assemble(CollectEntity request, EventSink<List<BaseJsonRequest>> sink){
    String traceId;
    try{
      final InvokeParseResult parseResult = _invokeManager.digest(request);
      final String invokeId = parseResult.invokeId;
      traceId  = parseResult.traceId;
      if (parseResult.isDummy || traceId == null) {
        return;
      }
      if(_invokeManager.head(traceId) == null || _invokeManager.head(traceId).isReplay){
        return;
      }
      //if it is Main-invoke means that it is end of the call
      if (request.mainInvoke) {
        final InvokeCollectRequest collectRequest = new InvokeCollectRequest();
        if (_invokeManager.hasSubInvoke(traceId)) {
          List<SubInvokeCollectRequest> subInvokes = new List();
          int depth =  _invokeManager.invokeStack(traceId).length;
          for (int i = 0; i < depth-1; i++) {
            subInvokes.add(_invokeManager.drainInvoke(traceId));
            subInvokes[i].index = (depth-1)-i;
          }
          collectRequest.subInvokeList = subInvokes;
        }
        collectRequest.traceId  = traceId;
        collectRequest.taskInfo = _invokeManager.taskInfo();
        switch (request.invokeCat) {
          case REQUEST_CAT.HTTP:
            HttpMainInvoke httpInvoke = safeCast(parseResult.currReq);
            collectRequest.mainInvoke = httpInvoke..responseBody = request.returnResult;
            if (_buffer.length <= SIZE_OF_BUFFER) {
              _buffer.add(collectRequest);
              if (_buffer.length > 0 && _buffer.length % BULK_SIZE == 0) {
                //TODO: SORTING AND DO ASSEMBLY
                List<BaseJsonRequest> bulk = new List<BaseJsonRequest>();
                for (int i = 0; i < BULK_SIZE; i++) {
                  if(!_buffer.isEmpty){
                    bulk.add(safeCast(_buffer.removeFirst()));
                  }
                }
                if (!_collectStream.isClosed) {
                  sink.add(bulk);
                }
              }
            } //drop if reached limits
            break;
          default:
            break;
        }
      } else {
        List<String> _params = new List<String>();
        parseResult.invokeParams.forEach((e){
          _params.add(jsonEncode(safeCast(e)));
        });
        _invokeManager.attachInvoke(traceId, data: new SubInvokeCollectRequest()
          ..invokeIdentity = invokeId
          ..returnObject   = request.returnResult
          ..params         = _params
        );
      }
    }catch(e,stack){
      print("$e $stack");
    }finally{
      if (traceId!=null && request.mainInvoke){
        _invokeManager.finishInvoke(traceId);
      }
    }
  }

  void _statInfo()  =>
      log?.error('''TrafficLabour - Collector info Queue Status : ${_buffer.length} of $SIZE_OF_BUFFER Bulk  Size : $BULK_SIZE TaskInfo : ${_invokeManager.taskInfo()} Isolates Cnt : ${_invokeManager.invokeRateFac} Invoke Status : ${_invokeManager._invokeCounters}''');

  /**
   *Flush fixed length bulks
   **/
  _flush() async{
    //TODO: AFTER bulk sending
  }
}


class _TrafficInvokeManager{

  final Map<String,int>                               _invokeCounters = {};
  final Map<String, Queue>                            _invokeStack    = {};
  static const int _defaultInvokeChainLength        = 128;
  static const int _defaultMainInvokeQty            = 128;
  TrafficTaskInfo _taskInfo;
  final int invokeRateFac = EnvFramework.httpConcurrence;

  void refreshTaskRule(String taskId,TrafficTaskInfo content){
    log?.warn("TrafficLabour - taskInfo changed to $content");
    if(content!=null){
      _taskInfo = content;
      _invokeCounters.clear();
    }
  }

  FutureOr inspectInvoke(CollectEntity entity){
    try {
      final InvokeParseResult curr = digest(entity);
      if(entity.mainInvoke){
        if(_whiteListed(curr.invokeId)){
          beginInvoke(curr);
        }
        return null;
      }
      else{
        InvokeParseResult mainInvoke = head(entity.traceId);
        if (!InvokeParseResult.isValidReplayRequest(mainInvoke)) {
          return null;
        }
        SubInvokeRequest query = new SubInvokeRequest()
          ..invokeIdentity = curr.invokeId
          ..traceId        = curr.traceId
          ..stacktrace     = ""
          ..extInfo        = ""
          ..params         = curr.invokeParams;
        return _Gateway.sendRequest(
            _Gateway.ACTION_QUERY_INVOKE, query.toString(),
            callback: (rspJson) {
              SubInvokeResponse json = new SubInvokeResponse.fromJson(rspJson);
              return json.success && !json.invokeThrough
                  ? json.returnObject
                  : null;
            }
        );
      }
    } catch (e) {
      return null;
    }
  }

  InvokeParseResult digest(CollectEntity req) {
    var curr = Zone.current[#TraceInfo];
    if(curr!=null){
      TraceInfo traceInfo = safeCast(curr);
      req.traceId =  req.traceId ?? traceInfo.eagleeyeTraceid;
    }
    return req.mainInvoke ? (head(req.traceId) ?? _digest(req)) : _digest(req);
  }

  InvokeParseResult _digest (CollectEntity request){
    InvokeParseResult rt = InvokeParseResult.dummyResult;
    try {
      if (request.invokeCat == REQUEST_CAT.HTTP) {
        final HttpRequest httpRequest = safeCast(request.request);
        final Uri uri = httpRequest.requestedUri;
        Map<String, String> getQuery  = httpRequest.uri.queryParameters;
        rt = new InvokeParseResult()
          ..isReplay   = ("true" == getQuery["isFromDoomReplay"])
          ..invokeType = REQUEST_CAT.HTTP
          ..traceId    = request.traceId
          ..invokeId   = "${uri.path.substring(uri.path.lastIndexOf('?') + 1)}";
        Map<String,List> formData = new Map();
        request.params.forEach((k,v) => formData.putIfAbsent(safeCast(k), ()=>safeCast(v)));
        final Map<String, List<String>> respHeaders = {};
        httpRequest.response.headers.forEach((name, list) =>
            respHeaders.putIfAbsent(name, () => list));
        final Map<String, List<String>> reqHeaders = {};
        httpRequest.headers.forEach((name, list) =>
            reqHeaders.putIfAbsent(name, () => list));
        rt.currReq        = new HttpMainInvoke()
          ..queryString    = httpRequest.uri.query
          ..url            = uri.hasQuery?
          uri.toString().substring(0,uri.toString().lastIndexOf("?")):uri.toString()
          ..responseHeadrs = respHeaders
          ..httpStatus     = httpRequest.response.statusCode
          ..requestHeaders = reqHeaders
          ..methodType     = httpRequest.method
          ..formData       = formData;
      }else{
        List _params = new List();
        jsonDecode(request.params == null ? "{}" : safeCast(request.params)).forEach((k,v){
          _params.add({safeCast(k):safeCast(v)});
        });
        rt = new InvokeParseResult()
          ..invokeParams = _params
          ..traceId      = request.traceId
          ..currReq      = request.returnResult
          ..invokeId     = request.invokeId;
      }
    } catch (e,stack) {print("$e $stack");}
    finally {
      return rt;
    }
  }

  TrafficTaskInfo taskInfo({idx:DEFAULT_TASK_ID})  =>  _taskInfo;

  Queue invokeStack(String traceId)  => _invokeStack[traceId];

  bool hasSubInvoke(String traceId)  =>  invokeStack(traceId)!=null &&  invokeStack(traceId).length>1;

  InvokeParseResult head(String traceId){
    final Queue stack = invokeStack(traceId);
    return (stack!=null &&  stack.length >=1) ? (safeCast(stack.first)): null;
  }

  void beginInvoke(InvokeParseResult mainInvoke){
    //filter non-replay traffic
    if(!mainInvoke.isReplay && pass(mainInvoke.invokeId)){
      return;
    }
    attachInvoke(mainInvoke.traceId, invokeCxt:mainInvoke);
  }

  void attachInvoke(String traceId, {InvokeParseResult invokeCxt, SubInvokeCollectRequest data}) {
    final Queue stack = invokeStack(traceId);
    if (stack != null && (_invokeStack.length >= _defaultMainInvokeQty ||
        stack.length >= _defaultInvokeChainLength)) {
      return;
    }
    else if(stack!=null && data!=null){
      stack.add(data);
    }
    else if(invokeCxt!=null){
      _invokeStack.putIfAbsent(traceId, ()=>new DoubleLinkedQueue()..add(invokeCxt));
    }
  }

  SubInvokeCollectRequest drainInvoke(String traceId)  => (_invokeStack[traceId].removeLast() as SubInvokeCollectRequest);

  void finishInvoke(String traceId) {
    _invokeStack.remove(traceId);
  }
   //if not exit create one
  int _addInvokeCount(String invokeId) =>
        _invokeCounters.update(invokeId, (cnt) => ++cnt, ifAbsent: () => 1);

  bool _whiteListed(String invokeId) => taskInfo()!=null && taskInfo().invokeCollectRate[invokeId]!=null;

  bool pass(String invokeId) =>
        !(_addInvokeCount(invokeId) %
            (1000 * invokeRateFac / taskInfo().invokeCollectRate[invokeId]) <= 1);

}


class CollectEntity{

  String             _traceId;
  REQUEST_CAT        _invokeCat;
  dynamic            _params;
  dynamic            _request;
  String             _returnResult;
  bool               _mainInvoke;
  String             _invokeId;

  String      get traceId      => _traceId;
  REQUEST_CAT get invokeCat    => _invokeCat;
  dynamic     get params       => _params;
  dynamic     get request      => _request;
  bool        get mainInvoke   => _mainInvoke;
  String      get invokeId     => _invokeId;
  String      get returnResult     => _returnResult;
  set mainInvoke(bool value)       => _mainInvoke   = value;
  set traceId(String value)        => _traceId      = value;
  set invokeCat(REQUEST_CAT value) => _invokeCat    = value;
  set params(dynamic value)        => _params       = value;
  set request(dynamic value)       => _request      = value;
  set invokeId(String value)       => _invokeId     = value;
  set returnResult(String value)   => _returnResult = value;

  @override
  toString() =>
      "{\"traceId:\"$_traceId,\"invokeId:\"$_invokeId,\"params\":$_params,\"mainInvoke\":$_mainInvoke}";
}

class _Gateway {

  static const String ACTION_COLLECT_INVOKE = "collectInvoke";
  static const String ACTION_QUERY_INVOKE   = "querySubInvoke";
  static Client _client;

  static Future sendRequest(String action, String parameters,
      {String schema: "http", String host: "localhost", int port: 8084, Map<String, String> queryString, callback(String body)}) {
    final Uri uri = new Uri(scheme: schema,
        queryParameters: queryString,
        port: port,
        host: host,
        path: action);
    _client ??= new IOClient();
//    MyLogger.getDebugLogger(name:"trafficGuard")?.debug("TrafficLabour - post request : $uri $parameters");
    return  _client.post(uri.toString(), body: parameters)
        .timeout(GW_TIMEOUT)
        .then((resp) {
//      MyLogger?.getDebugLogger(name:"trafficGuard")?.debug("TrafficLabour - response ${resp.body}");
      var rt         = (callback!=null) ? callback(resp.body) : resp.body;
      bool isSuccess = safeCast(jsonDecode(resp.body)["success"]);
      return resp != null && isSuccess
          ? rt
          : null;
    }).catchError((e, stack) => null);
  }
}

String rewriteTraceId(HttpRequest request, Map<String,String> headers){
  Map<String,String> getQuery = request.requestedUri.queryParameters;
  String doomTraceId = getQuery["doomReplayTraceId"];
  if(doomTraceId !=null &&  doomTraceId!=""){
    headers["eagleeye-traceid"] = doomTraceId;
  }
  return headers["eagleeye-traceid"];
}
