import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:grpc/grpc.dart';
import 'package:grpc/service_api.dart';
import 'package:protobuf/protobuf.dart';
import 'package:saker/common/net/ip_address.dart';
import 'package:saker/common/utils/type_cast.dart';
import 'package:saker/framework/hsf/hsf_async.dart';
import 'package:saker/framework/log/log.dart';
import 'package:saker/framework/log/log_type.dart';
import 'package:saker/framework/services/grpc/src/generated/df_gaia_message.pb.dart'
    as msg;
import 'package:saker/framework/services/grpc/src/generated/df_gaia_message.pb.dart';
import 'package:saker/framework/services/grpc/src/generated/df_gaia_service.pbgrpc.dart';
import 'package:saker/framework/services/grpc/src/generated/df_gaia_trigger_metaq.pb.dart';
import 'package:saker/framework/services/grpc/src/generated/df_gaia_trigger_mtop.pb.dart';
import 'package:saker/framework/services/grpc/src/generated/google/protobuf/any.pb.dart';
import 'package:saker/framework/services/service_result.dart';
import 'package:saker/framework/services/service_var.dart';
import 'package:saker/framework/stat/service_monitor.dart';
import 'package:saker/framework/trace/trace_info.dart';
import 'package:saker/plugin/faas/common/instances.dart';
import 'package:path/path.dart';
import 'package:saker/plugin/faas/container/conf/gaia_config.dart';
import 'package:saker/plugin/faas/container/function.dart';

import '../function.dart' as func;
import 'df_trigger_metaq.dart';
import 'df_trigger_mtop.dart';

//Event processor function
typedef EventProccessor<Req, Resp, Rlt> = Future<Rlt> Function(
    Req request, FunctionService handler, func.Context contex);

// function result builder
typedef EvtResultBuilder<Resp, Rlt> = Rlt Function(
    Resp response, Map<String, dynamic> functionHandlerCtx, bool isSucceed,
    {String errorCode, String errorMsg, Map data, String buf});

//An abstract function service protocol
abstract class FunctionServiceProtocol {}

//Intercept function for flow control
Interceptor beforeCall = (ServiceCall call, ServiceMethod method) {

  //invoke method will be
  if (serviceEntries[func_virtual_path].flowLimitBlock()) {
    //tagging
    call.headers.putIfAbsent(func_cxt_flowlimit, () => 'true');
    return null;
  }
  return null;
};

class GaiaGrpcServiceProtocol extends MessengerServiceBase {
  @override
  Future<msg.FunctionConfig> getConfig(
      ServiceCall call, msg.GetConfigRequest request) async {

    final Map<String, List<int>> triggerMap = {};
    msg.FunctionConfig config = msg.FunctionConfig()
      ..functionName = functionName
      ..functionVersion = functionVersion;
    Directory parentPath = Directory(mtopConfigpath).parent;
    if (Directory(mtopConfigpath).existsSync()) {
      List<FileSystemEntity> files = Directory(mtopConfigpath).listSync();
      if (files.isNotEmpty) {
        Directory(mtopConfigpath).listSync().forEach((triggerFile) {
          triggerMap[basename(triggerFile.path)] =
              utf8.encode(File(triggerFile.path).readAsStringSync());
        });
      } else if (File(parentPath.path + "/mtop.yaml").existsSync()) {
        triggerMap['mtop.yaml'] = utf8.encode(
             File(parentPath.path + "/mtop.yaml").readAsStringSync());
      } else {
        throw Exception('Trigger not found  ${mtopConfigpath}');
      }
    }
    config.configs
      ..clear()
      ..addAll(triggerMap);
    return config;
  }

  @override
  Future<msg.Result> checkHealth(
      ServiceCall call, msg.CheckHealthRequest request) async {
    if (svcOfflineStatus == 0) {
      return msg.Result()..success = true;
    } else {
      MyLogger.getLogger().error('HTTP Health check Offline');
      return msg.Result()
        ..success = false
        ..errorCode = 'FAILED'
        ..errorMsg = 'Health check failed';
    }
  }

  @override
  Future<msg.EventResult> invokeFunction(
      ServiceCall call, msg.Event request) async {
    func.FunctionService handler =
        serviceEntries[func_virtual_path] as FunctionService;
    func.Context context = func.Context.request(request.id);

    //build function running context
    _initFunctionContext(call, handler, context, request);

    switch (request.type) {
      case func.func_evt_mtop:
        return handleMtop(request.payload.unpackInto(MtopFunctionRequest()),
            handler, context);
        break;
      case func.func_evt_metaq:
        return handleMetaq(request.payload.unpackInto(MetaqFunctionRequest()),
            handler, context);
        break;
      default:
        MyLogger.getLogger().error(
            'Function payload is empty, unknown event type ${request.type}');
        return $_buildEventResult(false,
            errorCode: func_msg_inter_map[func_msg_provider_error].toString(),
            errorMsg: func_msg_provider_error);
        break;
    }
  }

  void _initFunctionContext(ServiceCall call, func.FunctionService handler,
      func.Context context, msg.Event request) {
    func.FunctionInfo funcInfo = func.FunctionInfo()
      ..functionName = functionName
      ..functionVersion = functionVersion;

    func.FunctionContainer funcContainer = func.FunctionContainer()
      ..runtimeEnv = GaiaConfig.env
      ..runtimeUnit = GaiaConfig.unit
      ..containerVersion = GaiaConfig.containerVersion;

    context
      ..functionInfo = funcInfo
      ..functionContainerInfo = funcContainer;

    //init custom runtime args
    context.functionHandlerContext = Map.from(request.traceContext)
      ..[func_cxt_flowlimit] = _isBlocked(call);

    context.functionHandlerContext[func_evt_type] = eventTypeMap[request.type];
  }

  bool _isBlocked(ServiceCall call) =>
      call.headers[func_cxt_flowlimit] == 'true';
}

//build dummy event
msg.EventResult $_buildEventResult<R extends GeneratedMessage>(bool isSucceed,
    {R response,
    String errorCode = func_msg_success,
    String errorMsg = func_msg_success,
    msg.EventResult evt_result}) {
  evt_result ?? msg.EventResult()
    ..success = isSucceed
    ..resultMsg = isSucceed ? errorMsg : "请稍候再试";

  if (response != null) {
    evt_result..payload = Any.pack(response);
  }
  return evt_result;
}

//A event handler template for 'Gaia', trigger irrelevant.
//Should not bind any specific trigger
Future<EventResult> $_handle<Req, Resp, Rlt>(
    Req request,
    Resp response,
    final FunctionService handler,
    final func.Context context,
    final func.Event event,
    EvtResultBuilder<Resp, EventResult> buildResult) async {
  //trace begin:
  final TraceInfo traceInfo = TraceInfo();
  traceInfo.eagleeyeIndex = 0;
  traceInfo.eagleeyeTraceid =
      safeCast(context.functionHandlerContext[func_cxt_trace_id]) ?? "";
  traceInfo.eagleeyeRpcid =
      safeCast(context.functionHandlerContext['EagleEye-RpcId']) ?? "";
  traceInfo.eagleeyeUserData =
      safeCast(context.functionHandlerContext['EagleEye-UserData']) ?? "";
  traceInfo.ttid = safeCast((context.functionHandlerContext['ttid'])) ?? "";
  traceInfo.mtopPath = func_virtual_path;

  if (traceInfo.eagleeyeUserData.isNotEmpty) {
    Map<String, String> userData =
        Uri.splitQueryString(traceInfo.eagleeyeUserData);
    traceInfo.pressTestTag = userData["t"] ?? "";
//      traceInfo.inReloadIsolate = reqMap.containsKey('x-isolate-redirect');
// userData.containsKey("dpath_env");
    traceInfo.eagleeyeContext = userData;
  }

  final Map<String, String> eagle_ctx = {};
  bool eagleEyeFlag = false;
  eagle_ctx["svc_sig"] =
      "${context.functionHandlerContext['api']}:${context.functionHandlerContext['v']}";
  eagle_ctx["traceid"] = traceInfo.eagleeyeTraceid;
  eagle_ctx["rpcid"] = traceInfo.eagleeyeRpcid;
  eagle_ctx["method_name"] = func_virtual_path;
  eagle_ctx["x-remote-ip"] =
      safeCast(context.functionHandlerContext['x-remote-ip']);
  eagle_ctx["userdata"] = traceInfo.eagleeyeUserData;

  if (traceInfo.eagleeyeTraceid != null && traceInfo.eagleeyeTraceid != "") {
    http_eagleeye_start(json.encode(eagle_ctx).toString());
    eagleEyeFlag = true;
  }

  traceInfo.serviceMonitor = ServiceMonitor()
    ..traceId = traceInfo.eagleeyeTraceid
    ..rpcId = traceInfo.eagleeyeRpcid
    ..ttid = traceInfo.ttid
    ..req = jsonEncode(safeCast(context.functionHandlerContext[
        'data'])) //+ " userid: "+ (parameters.containsKey('userId') ? parameters['userId'] : "")
    ..reqMethod = safeCast(context.functionHandlerContext['api'])
    ..mtopPath = func_virtual_path
    ..localIP = IPAddress.getHostIp()
    ..remoteIP = safeCast(context.functionHandlerContext['x-remote-ip'])
    ..rt = DateTime.now().millisecondsSinceEpoch
    ..userId = safeCast(context.functionHandlerContext['userId'])
    ..userNick = safeCast(context.functionHandlerContext['nick'])
    ..logType = LogType.MtopSvc;

  String str_rlt;
  EventResult rlt;
  rlt = await runZoned(() async {
    try {
      //service.init();
      if (safeCast(context.functionHandlerContext[func_cxt_flowlimit])) {
        ServiceResult serviceResult = handler
                    .callbacks[HandlerLifecycleEvt.flowBlocked] !=
                null
            ? handler.callbacks[HandlerLifecycleEvt.flowBlocked](event, context)
            : null;
        traceInfo.serviceMonitor.success = false;
        traceInfo.serviceMonitor.respSize = 0;
        traceInfo.serviceMonitor.errCode = "FAIL_SERVICE_FLOW_LIMIT";
        traceInfo.serviceMonitor.flush();
        if (eagleEyeFlag) {
          http_eagleeye_end(traceInfo.eagleeyeTraceid ?? "", 1, 0);
        }
        traceInfo.serviceMonitor.reqMethod = "SERVICE_FLOW_LIMIT";
        MyLogger.getLogger(name: "flowLimit")
            ?.error("${traceInfo.serviceMonitor.toString()}");
        MyLogger.getLogger().error(
            "Flow limit actived, just return: ${traceInfo.eagleeyeTraceid} ${traceInfo.eagleeyeRpcid}.");
        //if callback is null or serviceResult is null, return false result
        if (serviceResult == null) {
          return buildResult(response, context.functionHandlerContext, false,
              errorCode: func_msg_flowlimit_blocked);
        }
        return buildResult(response, context.functionHandlerContext, true,
            data: safeCast(serviceResult.data),
            errorCode: serviceResult.errcode,
            errorMsg: serviceResult.errmsg);
      }

      //before handler Call
      if (handler.callbacks[HandlerLifecycleEvt.beforeCall] != null) {
        handler.callbacks[HandlerLifecycleEvt.beforeCall](event, context);
      }

      handler.callbacks[HandlerLifecycleEvt.beforeCall] ??
          handler.callbacks[HandlerLifecycleEvt.beforeCall];

      ServiceResult serviceResult = await handler.handle(event, context);
      if (serviceResult == null) {
        return buildResult(response, context.functionHandlerContext, true,
            errorCode: func_msg_service_fail);
      }

      str_rlt = jsonEncode(serviceResult.data);

      //after  handler call
      if (handler.callbacks[HandlerLifecycleEvt.afterCall] != null) {
        handler.callbacks[HandlerLifecycleEvt.afterCall](event, context);
      }

      return buildResult(response, context.functionHandlerContext, true,
          buf: str_rlt,
          data: safeCast(serviceResult.data),
          errorCode: serviceResult.errcode,
          errorMsg: serviceResult.errmsg);
    } catch (e, stackTrace) {
      MyLogger.getLogger().error(
          'Process function handler error biz error catch:$e \n${stackTrace}');
      return buildResult(response, context.functionHandlerContext, false,
          errorCode: func_msg_failed);
    }
  },
      zoneValues: {#FunctionContext: context, #TraceInfo: traceInfo},
      onError: (e, stackTrace) =>
          MyLogger.getLogger().error('Process zone error:$e\n${stackTrace}'));

  if (eagleEyeFlag) {
    http_eagleeye_end(
        traceInfo.eagleeyeTraceid ?? "", rlt.success ? 0 : 1, str_rlt.length);
  }
  traceInfo.serviceMonitor.success = rlt.success;
  traceInfo.serviceMonitor.errCode = rlt.resultCode;
  if (traceInfo.serviceMonitor.respSize == null) {
    //如果业务侧自己定义了result长度，框架不再修改
    traceInfo.serviceMonitor.respSize = rlt.payload.value?.length ?? 0;
  }
  if (str_rlt.length < 10000) {
    /** 结果太大则忽略*/
    traceInfo.serviceMonitor.res = str_rlt;
  }
  traceInfo.serviceMonitor.flush();
  return rlt;
}
