@TestOn("linux || mac-os")

import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:isolate';

import 'package:saker/common/utils/type_cast.dart';
import 'package:saker/framework/annotation/router_annotation.dart';
import 'package:saker/framework/environment/env_base.dart';
import 'package:saker/framework/event/event_pool.dart';
import 'package:saker/framework/event/event_pool_base.dart';
import 'package:saker/framework/event/notice_type.dart';
import 'package:saker/framework/jsonserializer/proto.dart';
import 'package:saker/framework/jsonserializer/utils.dart';
import 'package:saker/framework/services/service_base.dart';
import 'package:saker/framework/services/service_isolate_entry.dart';
import 'package:saker/framework/services/service_result.dart';
import 'package:test/test.dart';

import 'test_base.dart';

EnvBase _envBase;

void main() {
  SendPort httpIsolatePort;

  setUp(() async {
    TestBase.initConfFile();

    _envBase = new EnvTest(serviceEntries);
    _envBase.init(DART_TEST_ROOT, 'TEST');

    EventItem eventItem = EventPool.getEvent();
    var messages = [
      EnvFramework.rootPath,
      eventItem.sendPort,
      eventItem.id,
      0,
      new Map<String, SendPort>(),
      _envBase
    ];

    await Isolate.spawn(worker_isolate_entry, messages).catchError((e) =>
        print('Http isolate spawn error:$e'));

    httpIsolatePort = safeCast(await eventItem.future.timeout(new Duration(seconds: 1))
        .catchError((e) =>
        print('Http isolate wait error:$e')));
  });

  tearDown(() {
    EventItem eventItem = EventPool.getEvent();
    httpIsolatePort.send([
      EventType_Notice,
      [NoticeType.HttpShutdown, eventItem.sendPort, eventItem.id, 0]
    ]);
  });

  void setHttpOnline() async {
    EventItem eventItem = EventPool.getEvent();
    httpIsolatePort.send([
      EventType_Notice,
      [NoticeType.HttpOnline, eventItem.sendPort, eventItem.id, 1]
    ]);
    await eventItem.future;
  }

  void setHttpOffline() async {
    EventItem eventItem = EventPool.getEvent();
    httpIsolatePort.send([
      EventType_Notice,
      [NoticeType.HttpOffline, eventItem.sendPort, eventItem.id, 1]
    ]);
    await eventItem.future;
  }

  group("HTTP框架测试", () {
    test("心跳检测", () async {
      HttpClientRequest request = await new HttpClient().post('127.0.0.1', TEST_HTTP_PORT, '/health');
      HttpClientResponse response = await request.close();
      String result = await response.transform(utf8.decoder).join();

      expect(response.statusCode, 404);
      expect(result, 'ECHO');

      await setHttpOnline();
      request = await new HttpClient().post('127.0.0.1', TEST_HTTP_PORT, '/health');
      response = await request.close();
      result = await response.transform(utf8.decoder).join();
      expect(response.statusCode, 200);
      expect(result, 'ECHO');

      await setHttpOffline();
      request = await new HttpClient().post('127.0.0.1', TEST_HTTP_PORT, '/health');
      response = await request.close();
      result = await response.transform(utf8.decoder).join();
      expect(response.statusCode, 404);
      expect(result, 'ECHO');
    });

    test("路由错误-不存在的路由", () async {
      HttpClientRequest request = await new HttpClient().post('127.0.0.1', TEST_HTTP_PORT, '/Test/NotFound');
      HttpClientResponse response = await request.close();
      String result = await response.transform(utf8.decoder).join();

      expect(response.statusCode, 200);
      expect(result.contains('"FAIL_SYS_SERVICE_MISS::服务不存在"'), true);
    });

    test("路由正确-普通请求", () async {
      HttpClientRequest request = await new HttpClient().post('127.0.0.1', TEST_HTTP_PORT, '/Hello');
      HttpClientResponse response = await request.close();
      String result = await response.transform(utf8.decoder).join();

      expect(response.statusCode, 200);
      expect(result.contains('"SUCCESS::接口调用成功"'), true);
    });

    test("路由正确-入参和出参", () async {
      PHelloTestReq req = new PHelloTestReq()..params = 'TEST-STRING';

      Map<String, String> context = {
        'data': json.encode(req.toJson())
      };

      HttpClientRequest request = await new HttpClient().post('127.0.0.1', TEST_HTTP_PORT, '/Hello');
      request.write(TestBase.mapToQuery(context, encoding: utf8));
      HttpClientResponse response = await request.close();
      String result = await response.transform(utf8.decoder).join();

      expect(response.statusCode, 200);
      expect(result.contains(req.params), true);
      expect(result.contains('"SUCCESS::接口调用成功"'), true);
    });
  });
}

class PHelloTestReq extends ApiRequestBase with PHelloTestReqMixin  {
  final String apiName = 'mtop.taobao.idle.dart.hello';
  final String apiVersion = '1.0';

  String params;
}

class PHelloTestRes extends ApiResponseBase with PHelloTestResMixin {
  String data;
}

class PHelloTestReqMixin {
  final bool apiMsgType = false;
  final bool needAuth = false;
  Map<String, dynamic> get apiParams => toJson();

  PHelloTestRes parseResponse(Map<dynamic, dynamic> json) {
    return new PHelloTestRes()..fromJson(json);
  }

  void fromJson(Map<dynamic, dynamic> json) {
    _fromJsonPHelloTestReq(this as PHelloTestReq, json);
  }

  void _fromJsonPHelloTestReq(PHelloTestReq c, Map<dynamic, dynamic> json) {
    //从Map转成对象的实现
    c.params = JsonUtil.parseString(json['params']);

  }

  Map<String, dynamic> _toJsonPHelloTestReq(PHelloTestReq c) {
    Map<String, dynamic> val = <String, dynamic>{};

    void writeNotNull(String key, dynamic value) {
      if (value != null) {
        val[key] = value;
      }
    }
    writeNotNull('params', c.params);

    return val;
  }

  Map<String, dynamic> toJson() {
    return _toJsonPHelloTestReq(this as PHelloTestReq);
  }
}

class PHelloTestResMixin {
  //服务器当前时间
  String serverTime;

  String debugInfo;

  //被加密字段，需要解密
  Set<String> needDecryptKeys;

  void fromJson(Map<dynamic, dynamic> json) {
    _fromJsonPHelloTestRes(this as PHelloTestRes, json);
  }

  void _fromJsonPHelloTestRes(PHelloTestRes c, Map<dynamic, dynamic> json) {
    //从Map转成对象的实现
    c.data = JsonUtil.parseString(json['data']);

    c.serverTime = JsonUtil.parseString(json['serverTime']);
    c.debugInfo = JsonUtil.parseString(json['debugInfo']);
    c.needDecryptKeys = (json['needDecryptKeys'] as List)?.map<String>((dynamic e) => JsonUtil.parseString(e))?.toSet();

  }

  Map<String, dynamic> _toJsonPHelloTestRes(PHelloTestRes c) {
    Map<String, dynamic> val = <String, dynamic>{};

    void writeNotNull(String key, dynamic value) {
      if (value != null) {
        val[key] = value;
      }
    }
    writeNotNull('data', c.data);

    writeNotNull('serverTime', c.serverTime);
    writeNotNull('debugInfo', c.debugInfo);
    writeNotNull('needDecryptKeys', c.needDecryptKeys?.toList());

    return val;
  }

  Map<String, dynamic> toJson() {
    return _toJsonPHelloTestRes(this as PHelloTestRes);
  }
}

@RouterAnnotation("/Hello")
class HelloTestService extends ServiceBase<PHelloTestReq, PHelloTestRes> {
  @override
  Future<ServiceResult> process(final PHelloTestReq req, final MtopContext context) async {
    var res = new PHelloTestRes();

    if (req.params != null) {
      res.data = req.params;
    } else {
      res.data = "Hello World!";
    }

    return buildSuccResult(res);
  }
}

PHelloTestReq parsePHelloTestReq(Map data) {
  return new PHelloTestReq()..fromJson(data??new Map());
}

Map<String, ServiceBase> serviceEntries = {
  '/Hello': new HelloTestService()
    ..servicePath = '/Hello'
    ..serviceRequestParser = parsePHelloTestReq,
};
